/*************************************************************************************\


 #     #   #####   #     #   #####    #####   ######      #  #######  #####   ######
 #     #  #     #  ##    #  #     #  #     #  #     #    #      #    #     #  #     #
 #  #  #  #     #  # #   #  #              #  #     #   #       #          #  #     #
 #  #  #  #######  #  #  #  # #####   #####   #     #  #   #    #     #####   #     #
 #  #  #  #     #  #   # #  #     #  #        #     #  #######  #          #  #     #
 #  #  #  #     #  #    ##  #     #  #        #     #      #    #    #     #  #     #      
  ## ##   #     #  #     #   #####   #######  ######       #    #     #####   ######


                                  Preprocessor of
                 2D Wang Tiles with Elliptical Inclusions for T3D
                           Copyright: Daniel Rypl 2019

\*************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <stdarg.h>
#include <limits.h>
//#include <values.h>

/* TODO
	 implement isotropic scale of inclusions
	 implement anisotropic scale of inclusions
	 implement directives as options !!!
	 separate option for absolute epsilon and relative epsilon */

/* tentative command line options:
	 -mrr max radius rate
	 -mt max tiles
	 -re relative epsilon
	 -ae absolute epsilon
	 -es epsilon subdivisions
	 -me multiple epsilon
	 -ns no shift
	 -if input file
   -of output file
	 -ati arbitrary tile id
	 -aco allow center outside
	 -chc check compatibility
	 -spr separate periodic repetition
	 -rmr report missing redundand
	 -fe full estimate (keep rather directive REDUCED_ESTIMATE)
	 -ofs offset
	 -det detouch
	 -ovd overdetouch
	 -dro disable reference overlap
	 -cho check overlap
	 -chi check isolated
	 -us use side
	 -uc use code
	 -gs grid setup
	 -ds diagonal setup
	 -h help
   -X graphics 

	 DEF_REL_EPSILON ==> not needed any more */

/* for testing use:

	 >> for i in Testing/is_x/is_x_*.geom
	 >> do
	 >> wang3d4t3d $i /dev/null 0.001
	 >> wang2d4t3d 2d_geom_3d.geom 2d_geom_3d.t3d
	 >> ../T3d-qhm/src/t3d_run.scr 2d_geom_3d.t3d -X -$
	 >> done

	 ensure that wang3d4t3d is compiled with GENERATE_2D_INPUT and that SIDE_2D is set to XN;
	 it is recommended to compile wang3d4t3d with IS_TESTING, with MIRROR_X|Y|Z = 0 and with PER_REP_ALL;
	 if wang3d4t3d is compiled with IS_TESTING, ensure that CYCLE_COORD = 0;
	 if wang3d4t3d is compiled with IS_TESTING, it is recommended to try both with and without INTERFACE
	 if wang3d4t3d is compiled with IS_TESTING and PER_REP_ALL, it is recommended to try both with and without ORIGINAL */


#define VERSION_MAJOR         2
#define VERSION_MINOR         1

//#define DEBUG

/* basic user customisable directives */

#define MAX_RADIUS_RATE       2.5           /* 2.0 <= MAX_RADIUS_RATE */

/* warning: the larger MAX_TILES the less number of inclusions is allowed to intersect particular tile side */

#define MAX_TILES             100           /* should be power of 10 (real maximal number of tiles is max_tiles - 1) */

#define DEF_REL_EPSILON       0.001         /* 0 < DEF_REL_EPSILON <= MAX_REL_EPSILON */
#define MAX_REL_EPSILON       0.1           /* 0 < MAX_REL_EPSILON << 1.0 */

#define ARBITRARY_TILE_ID                   /* allow arbitrary non-negative numbers for tiles */

//#define SEPARATE_PER_REP                    /* periodic repetition specified separately for each direction ==> increases size of input file */

/* detouching reduces violation of proximity criteria but increseases risk of mutual intersection of inclusions */

//#define DETOUCH                             /* shift not intersected inclusion close to tile side inside tile */  

#ifdef DETOUCH
//#define OVERDETOUCH                         /* shift intersected inclusion close to tile side inside tile */
#endif

#define ALLOW_OVERLAP                       /* defined reference inclusions are allowed to mutually overlap */

#define CHECK_OVERLAP                       /* check for overlap */

#define ALLOW_OUTSIDE                       /* allow for inclusion center of which falls out of rererence cube */

/* check of compatibility is not recommended if REDUCED_ESTIMATE is not defined, because it potentionally leads 
	 (in case of huge sets of overlapping reference inclusions) to significant memory consumption */

#define CHECK_COMPATIBILITY                 /* check compatibility of sides with the same nonzero code */

#ifdef CHECK_COMPATIBILITY
//#define REPORT_MISSING_REDUNDANT            /* report missing and redundant inclusions */
#endif

/* note that isolated nodes may cause problems only on boundary of mesh assembly which is however likely to be subjected to periodic conditions,
	 it is thus probably desirable to keep them in the mesh and thus check for isolated nodes is not necessary */

//#define CHECK_ISOLATED                      /* check for isolated nodes which may cause problems in mesh assembly */

/* there are 3 approaches how to specify keyword tile:
	 1) it is specified only for patches (matrix and inclusions) and then it is propagated in t3d to boundary and fixed entities;
	 note that in this approach there is no need to specify keyword tile for interfaces because they do not have interior nodes;
	 2) it is specified only for matrix patch, during t3d input of the pathes it is propagated to its subpatches 
	 and then it is propagated to boundary and fixed entities;
	 this approach requires that all inclusions are specified as subpatches;
	 IMPORTANT: this approach is not applicable when using interfaces !!!
	 tile is propagated from matrix to outer hole subpatch, but it is not propagated to inner solid patch;
	 CAUTION: this approach is not applicable when subpatches are used recursively because patch does not store its subpatches !!!
	 3) it is specified for all model entitities (except virtual ones);
	 
	 in order to enable simple and safe output of renumbered meshes, it is desirable to ensure that all physical model entities 
	 on t3d input refer to a particular tile;
	 when approach 1 or 2 is adopted, this must be checked after tile propagation (to boundary and fixed entities);
	 in order to avoid writing special check routine, the check can be performed during discretization;
	 the check during input can be used only to trigger the tile propagation to boundary and fixed entities (if tile specification 
	 is missing on any physical entity), but it is not guaranteed that all entities will be visited during propagation;
	 in this approach, missing tile specification for region (in 3D problem) or patch (in 2D problem) is considered an error;
	 it implies that propagation to subpatch makes no sense in such a case;
	 when approach 3 is adopted, t3d should check whether keyword tile is present on input for each physical model entity;
	 if this check is performed also for interfaces, then tile keyword must be specified also for interfaces;

	 currently approach 3 is implemented;
	 the major disadvantage of approach 3 is that it unneccessarily increases size of t3d input file (by 2-4%);

	 note that combination of approaches 1 and 2 can be implemented;
	 it means that tile is specified for matrix patch and patches of inclusions with interface */

/* keyword TILE must be used for t3d4wt */

#define USE_TILE                            /* keep defined !!! */

/* there are also 2 approaches how to specify keyword side:
	 1) it is specified only for tile side curves and then it is cumulatively propagated in t3d to boundary and fixed entities;
	 note that only nonzero side is propagated;
	 2) it is specified for all relevant model entities

	 approaches 1 is easier in the sense that only simple side (edge) specification is required;
	 approach 3 is more complicated in the sense that simple (edge) as well as combined (corner) side specification is required;

	 since there is no easy apriori check, that nonzero side specification is missing at particular model entity,
	 approach 1 is considered better and safer;
	 note that coordinate based check may be nontrivial if more tiles are processed together which is however a typical case;

	 currently approach 1 is implemented;

	 since side can be easily retrieved from curve id, keyword side may not be used !!!
	 if USE_SIDE is not defined, it must be ensured that id of curves not lying on tile sides does not end by 1, 2, 3, 4, 5, 6, 7, 8 */
 
#define USE_SIDE                            /* specify tile sides explicitly */

/* note: codes need not be used;
	 they are currently used by t3d for coloring of matching sides only;
	 the original motivation of using codes was to enable to set up mirroring by t3d;
	 but this requires unique global patch numbering which may no be available if tiles are prepared by other preprocessor;
	 therefore the mirroring is handled here */

#define USE_CODE                            /* specify tile codes */

/* simple mirroring can be used to override incompatibility of matching entities due to round-off error
	 or numerical instability (see t3d_split_curve_start|end in curve.c);
	 the problem of numerical instability is temporarily (but not generally) solved in t3d4WT */

//#define USE_SIMPLE                          /* use explicit simple mirroring */

/* max shift (possibly simultaneously in each direction) when resolving (a) touching of ellipse with tile sides and (b) intersection of ellipse with 
	 tile edges is defined as MULTIPLE_EPS * epsilon;
	 note that this maximal shift is not considered when performing shift due to side touching but is considered when performing overdetouch;
	 this implies that MULTIPLE_EPS must be at least 1 !!!
	 value smaller than 2 may prevent overdetouch !!!
	 note that it is not likely that ellipse shifted when resolving side touching will be shifted in the same direction when resolving touching of intersection 
	 ellipse with tile edges and when resolving intersection of ellipse with tile edges;
	 but this cannot be guaranteed !!! */

#define MULTIPLE_EPS          2             /* MULTIPLE_EPS >= 1 (recommended value 2) */

/* warning: meshes obtained for GRID_3D, GRID_2D and DIAGONAL setup are generally different */

/* note: the rules below concerning the size of the octree (and consequently min_epsilon) for GRID_3D and GRID_2D are valid only 
	 if the size of the octree does not account for the elimination of mesh size octree dispersion (in other words only when codes = 0);
	 if octree dispersion is handled the size of octree for GRID_2D may be the same as the size of octree for GRID_3D if
	 k_GRID_3D = k_GRID_2D = k, where 2^k >= num_tiles_1D;
	 k_GRID_3D = (int)(log((int)(cbrt(tiles - 0.5)) + 1) / log(2) + 0.99999999) = (int)(log((int)(sqrt(tiles - 0.5)) + 1) / log(2) + 0.99999999) = k_GRID_2D;
	 for example: tiles = 4: 1 == 1 ===> GRID_3D = GRID_2D
	              tiles = 5: 1 != 2 ===> GRID_3D > GRID_2D
								...
								tiles = 8: 1 != 2 ===> GRID_3D > GRID_2D
								tiles = 9: 2 == 2 ===> GRID_3D = GRID_2D
								...
								tiles = 16: 2 == 2 ==> GRID_3D = GRID_2D
                tiles = 17: 2 != 3 ==> GRID_3D > GRID_2D */

#define GRID                  1             /* tiles are setuped into reqular 2D grid ==> smaller octree ==> smaller min_epsion */
#define DIAGONAL              2             /* tiles are setuped along spacial diagonal of the 1st quadrant of GCS ==> larger octree ==> larger min_epsilon */

#define SETUP                 GRID          /* GRID or DIAGONAL */

/* advanced user customisable directives */

//#define MIN_MAX_HALVING                   /* halving of intersection ellipse crossing tile edges (also) with respect to min and max diameter (NOT FULLY IMPLEMENTED !!!) */

/* warning: different meshes are obtained for MAX_ALLOWED_SIZE and MAX_PRESENT_SIZE excced size
	 because of different size of root octant and consequently also different min_epsilon */

#define MAX_ALLOWED_SIZE      1             /* maximal allowable size of inclusions is considered ==> lager offset, constant root_size */
#define MAX_PRESENT_SIZE      2             /* maximal present size of inclusion (although possibly not used) is considered ==> smaller offset, different root size */
//#define MAX_REAL_EXCESS       3             /* hard to implement ==> not implemented */

#define EXCESS_SIZE           MAX_ALLOWED_SIZE      /* MAX_ALLOWED_SIZE or MAX_PRESENT_SIZE */

/* warning: recommended values of offset are based on the assumption that virtual entities do not exceed out of the tile;
	 but this is generally not true and computed offset must be appropriately increased */

#define OFFSET                0.5           /* relative (with respect to tile size) distance between tiles (0.5 <= OFFSET <= 1.0) */

#if SETUP == GRID

/* at least tile offset will ensure the cubic tiles will not overlap in hydrostatic view (vi nor 1 1 1; vi vup 0 0 1);
	 (the best light direction is then render ldir 1 2 8);
	 but this increases octree and thus also min_epsilon;
	 it makes sense to use it only if tiles are cubes */

//#define TILE_OFFSET                         /* if the computed offset is smaller than tile size, use tile size as offset */
#endif

#define REDUCED_ESTIMATE                    /* prevent counting all inclusions into estimate */

#ifdef REDUCED_ESTIMATE

#define INTERVALS_2D          10            /* number of intervals for histogram of segments (10 <= INTERVALS_2D) */
#define INTERVALS_1D          10            /* number of intervals for histogram of elipses (10 <= INTERVALS_1D) */

#define INTERVAL_RATE_2D      2.0           /* last to first interval size ratio (1.0 <= INTERVAL_RATE_2D) */
#define INTERVAL_RATE_1D      2.0           /* last to first interval size ratio (1.0 <= INTERVAL_RATE_1D) */

#define CONSTANT_SERIES       0             /* redundant - constant serie can be achieved as arithmetic or geometric serie with INTERVAL_RATE set to 1 */ 
#define ARITHMETIC_SERIES     1
#define GEOMETRIC_SERIES      2

#define SERIES_2D             GEOMETRIC_SERIES
#define SERIES_1D             GEOMETRIC_SERIES

#endif

/* do not modify setting below unless necessary */

/* note: if zero CURVE_RATE or CURVE_RATE >> 1 is used, there is a risk that elliptical arcs might be discretized just
	 by a single segment coinciding with tile edge;
	 therefore density 2.0 specification is used for elliptical arcs starting and ending at the same tile edge;
	 it may be desirable to specify appropriate density also for some arcs starting and ending on different tile edges;
	 in such a case, however, the orientation of the intersection ellipse with respect to tile edges is relevant for approriate setting */

#define CURVE_RATE            1.0           /* default rate for t3d curves */

#define RIGHT_ANGLE           0.001         /* cos(89.94 degs) */
#define ZERO_ANGLE            0.9999995     /* cos(0.057 degs) */

#define MAX_ITER              100           /* maximum number of iterations for projection to ellipse */

#define CHECK_CORNER_HOLES_APRIORI          /* check for corner holes before the tile is processed */  /* not implemented yet - dodelat */

//#define PARAMETRIC_ROUTINES                 /* not really needed */

#define ITEMS_PER_LINE        10            /* number of ids on t3d input line */

#define BUFFER_SIZE           1024

#define INDENT                "       "

#define REMARK                '#'
#define NEWLINE               '\n'
#define CR                    '\r'
#define SPACE                 ' '

#define SEP_STRING            "################################################################################"

/* there may be realized some (not negligible) savings on the size of produced t3d input file
	 by replacing format %.13e by %.14g (both output 14 significant digits);
	 precision of g format should not exceed 17 (it seems that double is stored using 17 significant digits);
	 when using larger precision, a random rubish is printed which prevents removal of zeroes from fractional part
	 resulting in undesirable increase of file size;
	 for example:
	 printf("%.17g\n", 1.8); ==> 1.8
	 printf("%.18g\n", 1.8); ==> 1.80000000000000007
	 note: when using format %A.Be, A=B+7 assuming that only two-digit exponent is used */

//#define FORMAT                "%20.13e"
//#define FORMAT                "%.13e"
#define FORMAT                "%.15g"

#define FMT1                  FORMAT
#define FMT2                  FORMAT" "FORMAT
#define FMT3                  FORMAT" "FORMAT" "FORMAT



#define NOTIFICATION          0
#define WARNING               1
#define ALARM                 2
#define GENERAL_ERROR         3
#define FILE_OPEN_ERROR       4
#define FILE_READ_ERROR       5
#define FILE_WRITE_ERROR      6
#define MEMORY_ERROR          7
#define INPUT_ERROR           8

#define EPSILON               1.0e-10
#define EPSILON_T             1.0e-5
#define EPSILON_U             1.0e-5
#define EPSILON_V             1.0e-5

#define SQRT2                 1.41421356237309504880168872420969808          /* sqrt(2.0) */
#define SQRT1_2               0.707106781186547524400844362104849039         /* sqrt(1.0 / 2.0) */
#define PI_1                  3.14159265358979323846264338327950288          /* PI / 1.0 */
#define PI_2                  1.57079632679489661923132169163975144          /* PI / 2.0 */
#define PI_4                  0.78539816339744830961566084581987572          /* PI / 4.0 */
#define LN_2                  0.693147180559945309417232121458176568         /* log(2.0) */

#ifdef DEBUG
//#define DEBUG_PRINT

#ifdef ELIXIR
#define DEBUG_DRAW
#endif

#else /* DEBUG */

#ifdef ELIXIR
#undef ELIXIR                               /* suppress graphics if not in DEBUG mode */
#endif

#endif /* DEBUG */

#ifdef ELIXIR

#include "Esimple.h"
#undef NO
#undef YES

#endif /* ELIXIR */

/* IMPORTANT: it is generally not possible to use several geometrically identical inclusions occupying identical position (including periodic repetition);
	 since the inclusions are processed in order corresponding to their id, it is likely that sides with the same code using different inclusion 
	 would be topologically not identical and thus not mirrorable;

	 o---11111---222-----333333---o    222 - intersection of inclusion 2 with top side
   |                            |          inclusions 2 is processed as the 2nd inclusion (from 3 inclusions intersecting top side)

	 o---11111---444-----333333---o    444 - intersection of inclusion 4 with top side
   |                            |          inclusions 4 is processed as the 3rd inclusion (from 3 inclusions intersecting top side)

#ifdef DEBUG
	 if ALLOW_OVERLAP is not defined and CHECK_OVERLAP is defined, then identical inclusion occupying identical
	 position (including periodic repetition) should be detected in the check
#endif */

/* IMPORTANT: almost touching intersection must be resolved on the level of model;
	 it is not enough to only degenerate intersection into touching point;
	 this could be done only if inclusion exceeds tile side by less than min_epsilon;
	 for example if inclusion exceeds tile side by less than epsilon but its intersection with tile side is (possibly much) larger than epsilon
	 then it is not possible to pretend that the intersection takes place at the touching point only;
	 if the elements to be generated are smaller than the real intersection, the mesh generated over the inclusion would penetrate
	 through tile side, which is not permissible;
	 it is therefore necessary to perform some action, eg. shifting, isotropic scaling, anisotropic scaling, .... 


            ____-----------____             ----                              _o------o_     _o------o_
       __---                   ---__           | < eps    generated         _-          -_ _-          -_
   ---x-----------------------------x---    ----           ======>     -------------------x---------------------
     -                               -                      mesh         _-            touching            -_
                                                                        o               point                o
      |        >> epsilon           |
      |-----------------------------|          
                > msz

   note that there is currently no way how to resolve the case when inclusion exceeds tile by more than epsilon but its intersection with
	 tile side is less than minimum (problem driven) epsilon;

	 IMPORTANT: there is currently no good way how to ensure that boundary nodes on the same entity (matrix) will get closer to each other than min_epsilon
	 (which is likely to result in t3d error) if inclusion is touching or intersecting tile side;
	 for example

      ____                       ____
          o----____     ____----o           ---- < min_epsilon
	 -------o----------x----------o--------   ----

	 the only solution is to use OVERDETOUCH */

/* concept of halving solves for the problem when elliptical arc to be constructed has end points on (or very close to) the diameter of the ellipse;
	 such arc cannot be constructed by fixing a curve between the end points to a quadratic curve constructed using appropriate extreme points
	 of ellipse with respect to 2 perpendicular directions because the weight of the ctrl point (of the fixed curve) is approaching zero, while
	 the coordinates of the ctrl point are approaching infinity, which makes the geometry unstable;
	 such ellipse must be halved, which results in spliting of the elliptical arc;
	 the halving of the ellipse is performed either with respect to ellipse x|y|z-extreme points or min|max-extreme points;
	 note that I am not using cubic representation for half-ellipticxal physical curve (on ellipse intersecting tile edges)
	 because the decision whether the elliptic arc is half-ellipse is based on floating point operations;
	 moreover this case is not very likely */

#define SIMPLIFY                            /* obsolete - to be deleted */


#define B_0(T) (1.0-2.0*(T)+(T)*(T))
#define B_1(T) (2.0*(1.0-(T))*(T))
#define B_2(T) ((T)*(T))

#define dB_0(T) (-2.0+2.0*(T))
#define dB_1(T) (2.0-4.0*(T))
#define dB_2(T) (2.0*(T))


#define swap(VAL1, VAL2, TMP)              {(TMP) = (VAL1); (VAL1) = (VAL2); (VAL2) = (TMP);}

#define size_vec(VEC)                      ((VEC).x * (VEC).x + (VEC).y * (VEC).y)

#define mul_vec(VEC, VALUE)                {(VEC).x *= (VALUE); (VEC).y *= (VALUE);}
#define div_vec(VEC, VALUE)                {(VEC).x /= (VALUE); (VEC).y /= (VALUE);}

#define copy_vec(DEST, SRC)                {(DEST).x = (SRC).x; (DEST).y = (SRC).y;}
#define copy_vec_ELIXIR(DEST, SRC)         {(DEST).x = (SRC).x; (DEST).y = (SRC).y; DEST.z = 0.0;}

#define copy_mul_vec(DEST, SRC, VALUE)     {(DEST).x = (SRC).x * (VALUE); (DEST).y = (SRC).y * (VALUE);}
#define copy_div_vec(DEST, SRC, VALUE)     {(DEST).x = (SRC).x / (VALUE); (DEST).y = (SRC).y / (VALUE);}

#define add_vec(RES, VEC1, VEC2)           {(RES).x = (VEC1).x + (VEC2).x; \
                                            (RES).y = (VEC1).y + (VEC2).y;}
#define sub_vec(RES, VEC1, VEC2)           {(RES).x = (VEC1).x - (VEC2).x; \
                                            (RES).y = (VEC1).y - (VEC2).y;}

#define add_mul_vec(RES, VEC1, VEC2, VAL1, VAL2)   {(RES).x = (VEC1).x * (VAL1) + (VEC2).x * (VAL2); \
		                                                (RES).y = (VEC1).y * (VAL1) + (VEC2).y * (VAL2);}

#define dot_product(VEC1, VEC2)            ((VEC1).x * (VEC2).x + (VEC1).y * (VEC2).y)
#define comp_product(RES, VEC1, VEC2)      {(RES).x = (VEC1).x * (VEC2).x; \
                                            (RES).y = (VEC1).y * (VEC2).y;}

#define dist_point(PNT1, PNT2)             (((PNT1).x - (PNT2).x) * ((PNT1).x - (PNT2).x) + \
                                            ((PNT1).y - (PNT2).y) * ((PNT1).y - (PNT2).y))
#define aver_point(RES, PNT1, PNT2)        {(RES).x = ((PNT1).x + (PNT2).x) / 2.0; \
																						(RES).y = ((PNT1).y + (PNT2).y) / 2.0;}


#define array(MAT,ROWS,COLS,ROW,COL)       MAT[(ROW)*(COLS)+(COL)]

#define min_val(VAL1, VAL2)                ((VAL1) < (VAL2) ? (VAL1) : (VAL2))
#define max_val(VAL1, VAL2)                ((VAL1) > (VAL2) ? (VAL1) : (VAL2))

#define abs_vec(VEC)                       {(VEC).x = fabs((VEC).x); (VEC).y = fabs((VEC).y);} 
#define max_vec(VEC)                       max_val((VEC).x, (VEC).y)
#define min_vec(VEC)                       min_val((VEC).x, (VEC).y)

#define check_shift_tmp(STATUS)            add_vec(total_shift_tmp, total_shift, shift_vec_tmp); \
	                                         abs_vec(total_shift_tmp);                             \
				                                   max_shift_tmp = max_vec(total_shift_tmp);             \
				                                   if(max_shift_tmp < max_shift){                        \
																						 max_shift = max_shift_tmp;                          \
																						 copy_vec(shift_vec, shift_vec_tmp);                 \
																						 corner_tmp = (STATUS);                              \
                                           }



#ifdef PARAMETRIC_ROUTINES
/* caution: shift, scale and expand macros do not modify ctrl points of ellipse */
#endif

/* caution: shift, scale and expand macros do not modify extreme points and diffs */

#define shift_ellipse(ELLIPSE, SHIFT_X, SHIFT_Y)   {(ELLIPSE).center.x += SHIFT_X; \
                                                    (ELLIPSE).center.y += SHIFT_Y;}

#define scale_ellipse(ELLIPSE, SCALE)      {(ELLIPSE).max *= SCALE;	\
                                            (ELLIPSE).min *= SCALE;}

#define expand_ellipse(ELLIPSE, EPS)       {(ELLIPSE).max += EPS;	\
                                            (ELLIPSE).min -= EPS;}


/* unique global numbering is introduced to allow for efficient mirroring;
	 since the mirroring is applied only between side curves (tile side, intersection of inclusions with tile side)
	 it is enough to use unique global numbering for curves;

	 1st concept:
	 inclusion_id | repetition | tile_id | edge     for intersection of inclusion with tile side
            	               | tile_id | edge     for rest of tile side

   edge: 1 - 4 : 1 decadic digit
   tile_id: 1 - (MAX_TILES - 1) : typically 2 (MAX_TILES == 100) or 3 (MAX_TILES == 1000) decadic digits
	 repetition: 0 - 1 : 1 decadic digit (repetition is reduced from 4 cases in 2D space to 2 cases on particular 1D side)
	 inclusion_id: 1 - 21474 (if MAX_TILES == 1000) for 32 bit signed long ==> insufficient !!!
                 1 - 214747 (if MAX_TILES == 100) for 32 bit signed long ==> may be sufficient if boundary inclusions are numbered first

	 note: I cannot used unsigned long as t3d does not support unsigned long ids
	       I cannot rely on 64 bit platform as long has 32 bits on LLP64 (Windows)
         ==> max signed long = 2147483648

	 2nd concept:
	 local_curve_id | tile_id | edge 

	 local_curve_id: 1 - 214748 (if MAX_TILES == 1000 and if NWT <= 364) for 32 bit signed long ==> reasonable
                   1 - 214747 (if MAX_TILES == 1000 and if NWT > 364) for 32 bit signed long ==> reasonable
									 1 - 2147483 (if MAX_TILES == 100 and if NWT <= 64) for 32 bit signed long ==> sufficient
									 1 - 2147482 (if MAX_TILES == 100 and if NWT > 64) for 32 bit signed long ==> sufficient

	 local_curve_id is separate for each side !!!
                   
	 note that unique curve numbering must be also applied to side curves bounding matrix;
	 number of such curves is number of inclusions intersecting/touching the side +/- 1;
	 this implies that tile side may be intersected/touched by up to (2147483 - 1) / 2 = 1073741 inclusions (including repeated inclusions);
	 note that this number should not be considered as max_side_intersections as it covers also touching;

	 there is no need to apply unique global curve numbering for hole outer curves duplicating tile side inner curves of inclusions with interface
	 because hole curve is not mirrored;
	 note, t3d currently does not recognize hole curve;
	 it is therefore necessary to use ordinary curve with suppressed output !
	 but there is still the need to prevent collision of the numbering of outer and inner curves;
	 there are 6 remaining digits for side position (5, 6, 7, 8, 9, 0 is not used by inner edge), which is enough for 4 sides;
	 digits 1, 2, 3, 4 (5, 6, 7, 8) are used for curves on inner (outer) edge;

   number of possible code combinations in a single direction: codes^2;
	 number of possible code combinations in a all directions: codes_x^2 * codes_y^2;
	 number of possible code combinations in a all directions for codes_x = codes_y = codes: codes^4;

	 for codes = 1: 1^4 = 1 ==> MAX_TILES = 10
	 for codes = 2: 2^4 = 16 ==> MAX_TILES = 100 ==> max_side_intersections = 2147483 ==> sufficient
	 for codes = 3: 3^4 = 81 ==> MAX_TILES = 100 ==> max_side_intersections = 2147482 ==> sufficient
	 for codes = 4: 4^4 = 256 ==> MAX_TILES = 1000 ==> max_side_intersections = 214748 ==> reasonable
#endif

   IMPORTANT: digits for side identification (edge) should be nonzero, because zero side specification is reserved (in t3d) for entities
	 not being on tile side !!!

	 contrary to 3D, where unique global numbering has been applied to patches which occur on tiles sides only,
	 in 2D there are many other curves not lying on tile side;
	 to identify these curves, their id must end by 0 or 9;

	 numbers of all other curves have the form:
	 global_curve_id | 0

	 global_curve_id: 1 - 214748364 for 32 bit signed long (without interface) ==> sufficient 
	 global_curve_id: 1 - 107374182 for 32 bit signed long (with interface) ==> sufficient */

#if 0
/* note that it is possible to number other curve as
	 global_curve_id | 0     for inner curves and virtual curves
	 global_curve_id | 9     for outer curves

	 in which case more curve are available if interface is used

	 global_curve_id (inner): 1 - 214748364 for 32 bit signed long ==> sufficient 
	 global_curve_id (outer): 1 - 214748363 for 32 bit signed long ==> sufficient 

	 this approach is however requires significant modifications to process_geometry
	 and is therefore not adopted */
#endif


/* IMPORTANT: numbering of inner and outer curves differ by 4;

	 inner (inclusion) curves;
	 outer (matrix) curves (irrelevant if interface == NO) */

#ifdef ARBITRARY_TILE_ID
/* IMPORTANT: macros get_x|y_side_curve_id are using fictitious tile id !!! */
#endif

#if 0
#define get_x_side_curve_id(IND) (++y_edge_curves[IND] * max_tiles + tile_id) * 10 + ((IND == 0) ? XN : XP + 1);
#define get_y_side_curve_id(IND) (++x_edge_curves[IND] * max_tiles + tile_id) * 10 + ((IND == 0) ? YN : YP + 1);
#else /* simplified */
#define get_x_side_curve_id(IND) (++y_edge_curves[IND] * max_tiles + tile_id) * 10 + 2 * X + (IND) + 1;
#define get_y_side_curve_id(IND) (++x_edge_curves[IND] * max_tiles + tile_id) * 10 + 2 * Y + (IND) + 1;
#endif


typedef enum logic logic;
typedef enum axes axes;
typedef enum edge edge;
typedef enum position position;

typedef enum vertex_status vertex_status;
typedef enum corner_status corner_status;

typedef struct vector_rec vector_rec;
typedef struct vector_rec point_rec;
typedef struct ellipse_rec ellipse_rec;

typedef struct tile_rec tile_rec;
typedef struct code_rec code_rec;
typedef struct inclusion_rec inclusion_rec;

typedef struct is_point_rec is_point_rec;
typedef struct is_pnt_rec is_pnt_rec;
typedef struct face_is_rec face_is_rec;
typedef struct edge_is_rec edge_is_rec;

typedef struct trans_matrix trans_matrix;

#ifdef CHECK_COMPATIBILITY
typedef struct side_inclusion side_inclusion;
#endif


enum logic{
	NO = 0,
	YES = 1
};


enum axes{
	X = 0,
	Y = 1
};


#ifdef MIN_MAX_HALVING

/* IMPORTANT: MIN and MAX are defined as continuation of axes */

#define MIN    2
#define MAX    3
#endif


enum edge{
	XN = 0,                             /* x negative */
	XP = 1,                             /* x positive */
	YN = 2,                             /* y negative */
	YP = 3                              /* y positive */
};


enum position{
	INDEF = 0,                          /* indefinite */
	POS = 1,                            /* positive */
	NEG = 2,                            /* negative */
	BOTH = 3                            /* positive and negative */
};


enum vertex_status{
	START = -1,                         /* starting */
	TOUCH = 0,                          /* touching */
	END = 1                             /* ending */
};


enum corner_status{
	IN = -1,                            /* inside */       /* CAUTION: INSIDE conflicts with Elixir */
	BND = 0,                            /* boundary */
	OUT = 1                             /* outside */
};



struct vector_rec{                    /* <==> point_rec */
	double x;
	double y;
};

/* IMPORTANT: since vertex id is valid when processing a single inclusion only (then all vertex ids corresponding to all
	 tile edge intersection points of that inclusion are zeroed again), additional array x|y_edge_vertex_id_array has been introduced;
	 to save some memory, struct is_pnt_rec was introduced to allow to store x|y_edge_vertex_data_array without vertex id */

struct is_pnt_rec{                    /* intersection point data */    
	double coord;                       /* coordinate along edge */
	vertex_status status;               /* status */
};
	

struct is_point_rec{                  /* intersection point */
	int id;                             /* vertex id - CAUTION: valid for processing of a single inclusion only */
	is_pnt_rec data;                    /* intersection point data */  
};


struct face_is_rec{                   /* parameters of ellipse - taken from 3D version */
	position side[2];                   /* position of ellipse center with respect to tile corner ordered as x, y */
	logic edge_half[2];                 /* flag that ellipse is (almost) halved by tile edge ordered as x, y */
	logic cross_half;                   /* flag that ellipse is (almost) halved by connection of adjacent intersection points */
	int cross_halving;                  /* halving (X or Y or MIN or MAX) if cross_half == YES */
	point_rec *extreme_pnt;             /* extreme points of intersection ellipse ordered as
                                         x-, x+, y-, y+ (2*X+0, 2*X+1, 2*Y+0, 2*Y+1)
														      			 needed really only if ellipse is intersecting or touching tile edge(s) */
	double weight;                      /* weight for arc between the 1st (2nd) and 4th (3rd) extreme point;
														      			 weight for arc between the 3rd (4th) and 1st (2nd) extreme point = sqrt(1.0 - weight * weight) */
};


struct edge_is_rec{                   /* intersection with tile edge */
	int count;                          /* number of intersection points */
	is_point_rec *point;                /* intersection points ordered increasingly with respect to coord */
	position side;                      /* position of intersection points with respect to tile corner */
};



struct ellipse_rec{
	point_rec center;                   /* ellipse center - may not fall outside reference tile  */
	vector_rec major, minor;            /* together with z+ righthanded Cartesian LCS */
	double max, min;                    /* size of half-axes (max < (size of tile - epsilon) / MAX_RADIUS_RATE, min > 10.0 * epsilon */
#ifdef PARAMETRIC_ROUTINES
	point_rec pnt0, pnt1, pnt2;         /* ctrl points in GCS for positive 1/4 (n LCS) of ellipse modelled as quadratic curve */
#endif
};


struct code_rec{
	long id;                            /* code number */
	long tile_id;                       /* tile id (fictitious) using code for the first time */
	long tile_id_ref;                   /* tile id (real) using code for the first time */
	edge side;                          /* tile side */
	long curves;                        /* number of side curves */

#ifdef CHECK_COMPATIBILITY
	long inclusions;                    /* number of inclusions interfering with reference side corresponding to code */
	side_inclusion *inclusion_array;    /* inclusions interfering with reference side corresponding to code */
#endif
};


struct tile_rec{
	long id;                            /* tile (real) id */
	point_rec size;                     /* size of reference tile */
	point_rec center;                   /* center of reference tile */
	point_rec corner[2];                /* opposite (min and max) corners of reference tile */
	double imsz, bmsz;                  /* internal and boundary default mesh size of reference tile */
	int mat;                            /* matrix material (0 < mat) */
	int code[4];                        /* side codes of particular tile  */
	long inclusions;                    /* number of inclusions in particular tile (on input) */
	long points;                        /* number of ctrl points in particular tile (on input) */
};


struct inclusion_rec{
	long id;                            /* inclusion id */
	ellipse_rec ellipse;                /* inclusion geometry */
	double imsz, bmsz;                  /* internal and boundary default mesh size of inclusion */
	int mat;                            /* inclusion material (0 <==> hole) used as inclusion property id */
	int ifmat;                          /* interface material (0 <==> no interface) used as interface property id */
	face_is_rec face_is;                /* parameters related to ellipse - in 3D handled as face intersection */
	edge_is_rec *edge_is[2];            /* intersection with tile edges (NULL <==> no intersection) */
	corner_status corner;               /* position of tile corner with respect to inclusion */
	logic active[4];                    /* active periodic repetitions */ 
	logic shift;                        /* flag whether inclusion was shifted ==> intersection check needed even if overlapping in reference tile is not allowed */
#ifdef ALLOW_OUTSIDE
	logic out[2];                       /* flag whether inclusion center is outside of reference tile */
#endif
};


struct trans_matrix{
	vector_rec x;
	vector_rec y;
};


#ifdef CHECK_COMPATIBILITY
struct side_inclusion{
	inclusion_rec *inclusion;           /* inclusion */
	short per_x, per_y;                 /* its particular periodic repetition */
};
#endif


static char *geom_file_name = NULL;
static char *t3d_in_file_name = NULL;

static FILE *geom_file = NULL;
static FILE *t3d_in_file = NULL;

static double epsilon = 0.0;
static double min_epsilon = 0.0;
static int max_subdivisions = 0;

static char line_buffer[BUFFER_SIZE];

static tile_rec tile;

static inclusion_rec *inclusion_array = NULL;
static code_rec *code_array = NULL;

static char output_spec[16];
static char mirror_spec_x[32], side_spec_x[8];         /* related to x-edge = y-side */        
static char mirror_spec_y[32], side_spec_y[8];         /* related to y-edge = x-side */

static double min_sz, max_sz;

/* static global variables for routines
	 get_inclusion_tile_intersection,
	 process_geometry, 
	 process_geometry_without_interface, 
	 process_geometry_with_interface, 
	 check_inclusion_outside_tile;

	 IMPORTANT: be aware of eventual variable shadowing !!! */

static long max_tiles = 1;                                                      /* maximum number of tiles */
static long max_side_intersections = 0;                                         /* maximum number of inclusions intersecting a single tile side */

static unsigned long max_tile_patches = 0;                                      /* maximum number of patches in tile */
static unsigned long max_edge_vertices[2] = {0, 0};                             /* maximum number of fixed vertices on tile edge */
static unsigned long max_edge_curves_1[2] = {0, 0};                             /* maximum number of segments correspodning to intersection of inclusion with tile edge */
static unsigned long max_edge_curves_2[2] = {1, 1};                             /* maximum number of segments between tile corners and intersections and 
																																									 touchings of inclusion with tile edge */

static long tile_patches;                                                       /* related to tile_patch_array */
static long x_edge_vertices[2], y_edge_vertices[2];                             /* related to x|y_edge_vertex_array */
static long x_edge_curves_1[2], y_edge_curves_1[2];                             /* related to x|y_edge_curve_1_array */
static long x_edge_curves_2[2], y_edge_curves_2[2];                             /* related to x|y_edge_curve_2_array */

static long x_edge_curves[2], y_edge_curves[2];                                 /* total number of edge curves needed to generate global unique curves id */

/* IMPORTANT: the following arrays contain the outer entities if interface == YES */

static long *tile_patch_array = NULL;
static long *x_edge_vertex_id_array[2] = {NULL, NULL};                          /* ordering y-, y+ */
static long *y_edge_vertex_id_array[2] = {NULL, NULL};                          /* ordering x-, x+ */
static is_pnt_rec **x_edge_vertex_data_array[2] = {NULL, NULL};                 /* ordering y-, y+ */
static is_pnt_rec **y_edge_vertex_data_array[2] = {NULL, NULL};                 /* ordering x-, x+ */
static long *x_edge_curve_1_array[2] = {NULL, NULL};                            /* ordering y-, y+ */
static long *y_edge_curve_1_array[2] = {NULL, NULL};                            /* ordering x-, x+ */
static long *x_edge_curve_2_array[2] = {NULL, NULL};                            /* ordering y-, y+ */
static long *y_edge_curve_2_array[2] = {NULL, NULL};                            /* ordering x-, x+ */


static long tile_id;

static long vertex_id = 0;
static long curve_id = 0;
static long patch_id = 0;
static long interface_id = 0;

static long origin_id, corner_id, cor_id;

static long x_curve_id, y_curve_id;

static int shifting_x[4] = {0, 1, 0, 1};
static int shifting_y[4] = {0, 0, 1, 1};

static int x_ind, y_ind;                             /* indices of tile sides {0;1} closer to inclusion center (actual position) */

static double shift_x, shift_y;                      /* shift of actual inclusion */

static face_is_rec *face_is = NULL;
static edge_is_rec *x_edge_is = NULL, *y_edge_is = NULL;

static logic no_shift = NO, shift_done = NO;

static int ref_position[2] = {POS, NEG}, sign[2] = {1, -1};

static inclusion_rec *inclusion = NULL;
static ellipse_rec *ellipse = NULL;
static code_rec *code = NULL;

static logic x_touching, y_touching;                 /* flags that ellipse is touching tile side */
static logic x_intersection, y_intersection;         /* flags that ellipse is intersecting tile side */
static logic interface;

static vector_rec dir_x = {1.0, 0.0};                /* GCS base vectors */
static vector_rec dir_y = {0.0, 1.0}; 

static vector_rec dir_1, dir_2;                      /* GCS base vectors in LCS of ellipsoid, then transformed to sphere space */
static vector_rec vec_x, vec_y;                      /* normals of tile sides in LCS of ellipsoid after being transformed to sphere space */

static double dist_x, dist_y;                        /* size of ellipse parallel to x and y edge at y and x edge intersection (closer) to corner */
static double diff_x, diff_y;                        /* differences of ellipse from tile sides */
static double x_diff, y_diff;                        /* differences of intersection with tile edges from tile corner */
static double x_coord[2], y_coord[2];                /* intersections with tile edges */

static int ind_x, ind_y;                             /* indices of tile sides {0;1} closer to ellipse center (reference position) */
static int x_count, y_count;                         /* number of intersection of elliposid with tile edges */

static point_rec ext_point_x[2], ext_point_y[2];     /* extreme points of ellipse in GCS */
static point_rec par_corner[2];                      /* opposite (min and max) corners of particular tile subjected to appropriate shift */


static trans_matrix trans_g2l, trans_l2g;            /* transformations between GCS and LCS of ellipse */

static vector_rec shift_vec;                         /* shifting vector */
static vector_rec total_shift;                       /* total shifting vector */

static logic corner_hole[4];                         /* flag indicating whether tile corner is inside hole */
static logic inside_hole = NO;                       /* flag indicating that there exist a hole inclusion (without interface)
																												or non-hole inclusion with interface with tile corner inside */

static logic compatibility_alarm = NO;               /* compatility alarm issued */

#ifdef CHECK_ISOLATED
static logic corner_isolated[4];                     /* flag indicating whether the tile will contain isolated vertex at the corner */
static logic boundary_hole = NO;                     /* flag indicating that there exist a hole inclusion (without interface)
																												or non-hole inclusion with interface with tile corner on boundary
																												while each tile edge is intersected twice */
#endif

#ifdef ELIXIR

#define FONT                  "-adobe-courier-bold-r-normal--*-120-*-*-m-*-iso8859-1"

#define TILE_LAYER            1
#define INCLUSION_LAYER       2

#ifdef DEBUG_DRAW
//#define DRAW_INCLUSION_CENTER
#define DRAW_EXTREME_POINT_ELLIPSE
#define DRAW_TOUCHING_POINT_ELLIPSE
#define DRAW_INTERSECTING_POINT
#endif

static EPixel green, red, yellow, blue, cyan, magenta, orange, black, white;
static EPixel bgcolor;
static Font font;
static logic draw = NO;

static void
draw_number(point_rec *pnt, int layer, EPixel color, int number, char *prefix);
static void
draw_point(point_rec *pnt, int layer, EPixel color);
static void
draw_line(point_rec *pnt1, point_rec *pnt2, int layer, EPixel color);
static void
draw_elliptic_arc(point_rec *pnt1, point_rec *pnt2, point_rec *pnt3, double w1, double w2, double w3, int layer, EPixel color);

static void
draw_elliptic_arc_segment(ellipse_rec *ellipse, int q, double ta, double tb, point_rec *pointa, point_rec *pointb, 
													trans_matrix *trans, int layer, EPixel color);
static void
draw_elliptic_arc_start_segment(ellipse_rec *ellipse, int q, double t, point_rec *point, trans_matrix *trans, int layer, EPixel color);
static void
draw_elliptic_arc_end_segment(ellipse_rec *ellipse, int q, double t, point_rec *point, trans_matrix *trans, int layer, EPixel color);

static void
draw_tile(tile_rec *tile, int layer, EPixel color);
static void
draw_inclusion(inclusion_rec *inclusion, int layer, EPixel color);
static void
draw_ellipse(ellipse_rec *ellipse, int layer, EPixel color);

#ifdef __cplusplus
#define CONST_CHAR_PTR           (char *)
#else
#define CONST_CHAR_PTR 
#endif

static void
augment_command_table(void);

static void
select_distance_points(void);
static void
distance_handler(Widget w, EView *v_p, XEvent *event);

#endif

static logic
get_inclusion_tile_intersection(void);

/* special routines called from get_inclusion_tile_intersection
	 and dependent on static global variables */

static void
resolve_ellipse_x_side_touching(void);
static void
resolve_ellipse_y_side_touching(void);

static void
get_ellipse_x_edge_intersection(logic fix);
static void
get_ellipse_y_edge_intersection(logic fix);

static double
get_ellipse_weight(ellipse_rec *ellipse, point_rec *point1, point_rec *point2, trans_matrix *trans_g2l);

static void
process_geometry(void);

static logic
check_inclusion_outside_tile(void);

static void
output_patch_boundary(long *edge_curves, long **edge_curve_array, int *edge_curve_sign, long side_patches, long *side_patch_array);

/* end of special routines */

#ifdef CHECK_OVERLAP
static void
check_active_inclusion_overlap(ellipse_rec *ellipse1, int per_x, int per_y, long i);
static void
check_inclusion_periodic_overlap(ellipse_rec *ellipse1, int per_x, int per_y, long i);
#endif

static logic
check_ellipse_ellipse_overlap(ellipse_rec *ellipse1, ellipse_rec *ellipse2, double eps);

static logic
check_point_inside_ellipse(point_rec *point, ellipse_rec *ellipse, double eps, double *value);

static int
get_ellipse_line_intersection(ellipse_rec *ellipse, point_rec *point, vector_rec *dir, point_rec *intersection);
static void
get_ellipse_line_center_intersection(ellipse_rec *ellipse, vector_rec *dir, point_rec *pnt1, point_rec *pnt2);

#ifdef PARAMETRIC_ROUTINES
static int
get_ellipse_ellipse_intersection(ellipse_rec *ellipse1, ellipse_rec *ellipse2, point_rec *intersection);

static void
get_ellipse_ellipse_interval_intersection(ellipse_rec *ellipse1, ellipse_rec *ellipse2, int q, double t, double val, double dt,
																					point_rec *point);
#endif

static void
project_point_to_ellipse(point_rec *point, ellipse_rec *ellipse, point_rec *pnt);
static void
get_ellipse_normal_point(ellipse_rec *ellipse, point_rec *point, vector_rec *normal);

static double
get_quadratic_elliptic_arc_weight(ellipse_rec *ellipse, point_rec *point1, point_rec *point2);

#ifdef PARAMETRIC_ROUTINES
static int
project_point_to_ellipse_t_quadrant(point_rec *point, ellipse_rec *ellipse, point_rec *pnt, double *t_par, logic spec);

static void
get_ellipse_point_t_quadrant(ellipse_rec *ellipse, int quadrant, double t, point_rec *pnt);
static double
get_ellipse_tangent_t_quadrant(ellipse_rec *ellipse, int quadrant, double t, vector_rec *grad);
static void
get_ellipse_normal_t_quadrant(ellipse_rec *ellipse, int quadrant, double t, vector_rec *normal);

static void
get_ellipse_point_t(ellipse_rec *ellipse, double t, point_rec *pnt);
static double
get_ellipse_tangent_t(ellipse_rec *ellipse, double t, vector_rec *grad);
static void
get_ellipse_normal_t(ellipse_rec *ellipse, double t, vector_rec *normal);

static logic
check_point_on_ellipse_interval(int q, double t, int q1, double t1, int q2, double t2);
#endif

static void
transform_vector_from_global_to_local(vector_rec *global, vector_rec *local, trans_matrix *trans_g2l);
static void
transform_vector_from_local_to_global(vector_rec *local, vector_rec *global, trans_matrix *trans_l2g);

static void
transform_point_from_global_to_local(point_rec *global, point_rec *local, point_rec *origin, trans_matrix *trans_g2l);
static void
transform_point_from_local_to_global(point_rec *local, point_rec *global, point_rec *origin, trans_matrix *trans_l2g);

static void
invert_rotation_transformation(trans_matrix *trans);
static void
get_inverted_rotation_transformation(trans_matrix *trans_in, trans_matrix *trans_out);

static void
transform_local_from_ellipse_to_circle(point_rec *elliptical, point_rec *circular, vector_rec *trans_e2c);
static void
transform_local_from_circle_to_ellipse(point_rec *circular, point_rec *elliptical, vector_rec *trans_c2e);

static void
invert_scale_transformation(vector_rec *trans);
static void
get_inverted_scale_transformation(vector_rec *trans_in, vector_rec *trans_out);

static char *
get_next_record_file(FILE *file, char *file_name, char *err_msg);
static char *
get_next_relevant_record_file(FILE *file, char *file_name, char *err_msg);

static void
write_record_file(FILE *file, char *file_name, char *format, ...);

static char *
get_next_record(char *err_msg);
static char *
get_next_relevant_record(char *err_msg);

static void
write_record(char *format, ...);

static void
error_message(int code, char *format, ...);
static void
error_exit(int exit_code);

static logic
solve_full_system(double *a, double *x, double *y, long n, long m);
static void
sort_id_coord(long n, long *id, is_pnt_rec **pnt);


void
print_geometry(void);
void
print_setting(void);


int
main(int argc, char** argv)
{
	logic relative, help = NO, halving, mirror;
	edge opposite[4] = {XP, XN, YP, YN};
	point_rec  point, *pnt_extreme = NULL, pnt_1, pnt_2, *p1 = NULL, *p2 = NULL;
	vector_rec vec, vec_tmp, normal1, normal2;
	double msz, size, val, per_shift_x, per_shift_y, offset, max_rad = 0.0, sz, product;
	double tile_shift_x, tile_shift_y, root_size, excess_size;
	long id, inclusions, codes, tiles, num_tiles_1D = 0;
	long ver_id, ver_id1, ver_id2, cur_id, inclusion_id;
	long i, j, k, m, n, kk;
	long *edge_curve_array[8], edge_curves[8]; 
	int power, incr = 0, per_x, per_y, per_xy, double_size, count, ind;
	int edge_curve_sign[8];
	char text[1024], suffix[][4] = {"-st", "-nd", "-rd", "-th"}, command[512];
	char facespec[][12] = {"negative x", "positive x", "negative y", "positive y"};
	char Facespec[][12] = {"Negative x", "Positive x", "Negative y", "Positive y"};
	vertex_status status;
	is_point_rec *is_point = NULL;

#ifdef REDUCED_ESTIMATE
	double a, b, sum;
	double area, x_length, y_length;
	double *min_area = NULL, *min_x_length = NULL, *min_y_length = NULL;
	double *sum_area_weight = NULL, *sum_x_length_weight = NULL, *sum_y_length_weight = NULL;
	double *sum_area = NULL, *sum_x_length = NULL, *sum_y_length = NULL;
	double *area_interval = NULL, *x_length_interval = NULL, *y_length_interval = NULL;
	long *area_count = NULL, *x_length_count = NULL, *y_length_count = NULL, x_touch_count = 0, y_touch_count = 0;
	logic corner_in = NO, corner_bnd = NO;
#endif
#ifdef CHECK_COMPATIBILITY
	long tmp_inclusions[4];
	side_inclusion *tmp_inclusion_array[4] = {NULL, NULL, NULL, NULL};
	side_inclusion *sd_inclusion = NULL, *dummy_inclusion_array = NULL;
	long dummy_inclusions;
	int dir;
	logic alarm;
#ifndef REDUCED_ESTIMATE
	long x_touch_count = 0, y_touch_count = 0;
#endif
#endif
#ifdef ARBITRARY_TILE_ID
	long *tile_array = NULL;
#else
	logic *tile_array = NULL;
#endif
#if SETUP == GRID
	int grid_size, row, column;
#endif
#ifdef ALLOW_OUTSIDE
	logic outside;
#endif
#ifdef ELIXIR
	EView *view = NULL;
	BOOLEAN success;
#endif
#if 0
	double min_msz;
	int bits;
#endif

	if(MAX_RADIUS_RATE < 2.0)error_message(GENERAL_ERROR, "Invalid MAX_RADIUS_RATE");
	if(MULTIPLE_EPS < 1.0)error_message(GENERAL_ERROR, "Invalid MULTIPLE_EPS");

#ifdef REDUCED_ESTIMATE
	if(INTERVAL_RATE_1D < 1.0)error_message(GENERAL_ERROR, "Invalid INTERVAL_RATE_1D");
	if(INTERVAL_RATE_2D < 1.0)error_message(GENERAL_ERROR, "Invalid INTERVAL_RATE_2D");
	if(INTERVALS_1D < 10)error_message(GENERAL_ERROR, "Invalid INTERVALS_1D");
	if(INTERVALS_2D < 10)error_message(GENERAL_ERROR, "Invalid INTERVALS_2D");
#endif

	if(argc > 1){
		if(strcmp(argv[1], "-h") == 0)help = YES;
	}
	
	if(help == NO){
		if(argc > 1){
			if(strcmp(argv[1], "-X") == 0){
				incr = 1;
				
#ifdef ELIXIR
				draw = YES;
#endif
				
				argc--;
			}
		}
	}
	
	fprintf(stdout, "\n");
	fprintf(stdout, "Wang2d4T3d - Preprocessor of 2D Wang Tiles for T3d\n");
	fprintf(stdout, "             (c) Daniel Rypl 2019 - Version %d.%d\n", VERSION_MAJOR, VERSION_MINOR);
	fprintf(stdout, "--------------------------------------------------\n");

	if(help == YES || argc < 3){
		fprintf(stderr, "\n"); 
#ifdef ELIXIR
		fprintf(stderr, "Usage: %s [-h] [-X] geom_file t3d_file [epsilon [subdivisions]]\n\n", argv[0]);
#else
		fprintf(stderr, "Usage: %s [-h] geom_file t3d_file [epsilon [subdivisions]]\n\n", argv[0]);
#endif
		fprintf(stderr, INDENT "geom_file: input geometry\n");
		fprintf(stderr, INDENT "t3d_file: output model for T3d4WT\n");
		fprintf(stderr, INDENT "epsilon: minimal distance of inclusion from tile sides\n");
		fprintf(stderr, INDENT "         minimal distance of inclusion intersection with tile edges from tile corners\n");
		fprintf(stderr, INDENT "    pos. value is considered as absolute epsilon\n");
		fprintf(stderr, INDENT "         (absolute epsilon <= min(dimension[xyz]) x MAX_REL_EPSILON)\n");
		fprintf(stderr, INDENT "    neg. value is considered as relative epsilon with respect to min(dimension[xyz])\n");
		fprintf(stderr, INDENT "         (relative epsilon <= MAX_REL_EPSILON)\n");
		fprintf(stderr, INDENT "    zero value is replaced by min(dimension[xyz]) x DEF_REL_EPSILON\n");
		fprintf(stderr, INDENT "    missing value is considered as zero value\n");
		fprintf(stderr, INDENT "subdivisions: number of times epsilon may be halved to satisfy proximity criteria \n");
		fprintf(stderr, INDENT "              for each particular inclusion\n");
		fprintf(stderr, INDENT "              missing value is considered as zero value\n");
		fprintf(stderr, INDENT "              use negative value to prevent inclusion shifting\n\n");
		if(help == YES){
			print_geometry();
			print_setting();
		}
		exit(1);
	}

#ifdef CHECK_COMPATIBILITY
#ifndef REDUCED_ESTIMATE
	error_message(WARNING, "Compatibility check may lead to significant memory consumptions if REDUCED_ESTIMATE is not defined");
#endif
#endif
	
	power = (int)log10(MAX_TILES - 0.5) + 1;
	while(power--)max_tiles *= 10;
	
#if 0
#if 1
	bits = sizeof(long) * CHAR_BIT;
#else /* LONGBITS is defined only in obsolete values.h */
	bits = LONGBITS;
#endif
	max_side_intersections = (long)(((1 << (bits - 1)) - 1) / (10 * max_tiles)) - 1;
#else
	max_side_intersections = (long)(LONG_MAX / (10 * max_tiles)) - 1;
#endif
	
	geom_file_name = argv[incr + 1];
	t3d_in_file_name = argv[incr + 2];
	
	if(argc > 3)epsilon = atof(argv[incr + 3]);
	if(argc > 4)max_subdivisions = atoi(argv[incr + 4]);
	
	if(max_subdivisions < 0){
		no_shift = YES;
		max_subdivisions = 0;
	}
	
	if((geom_file = fopen(geom_file_name, "r")) == NULL)
		error_message(FILE_OPEN_ERROR, "File %s opening error", geom_file_name);
	if((t3d_in_file = fopen(t3d_in_file_name, "w")) == NULL)
		error_message(FILE_OPEN_ERROR, "File %s opening error", t3d_in_file_name);
	
#ifdef ELIXIR
	if(draw == YES){
		if(ESIBuildInterface(ESI_GRAPHIC_EDITOR_MASK, 1, argv) != 0){
			fprintf(stderr, "Graphic interface error\n");
			exit(1);
		}
		augment_command_table();
		ESIPopup();
		view = ElixirNewView("Wang2D", "Wang2D", "midnightblue", "white", 400, 400);   /* use as bgcolor */
		EMAttachView(ESIModel(), view);
		EVSetViewOrientation(view, VIEW_ORIENT_TOP);
		EVShowAxes(view, NO);
		ESIHandleCmd("render ambient 0.1");
		EVSetRenderMode(view, NORMAL_RENDERING);
		EVSetLayerOnOff(view, TILE_LAYER, YES);
		EVSetLayerOnOff(view, INCLUSION_LAYER, YES);

		green = ColorGetPixelFromString("green", &success);
		red = ColorGetPixelFromString("red", &success);
		yellow = ColorGetPixelFromString("yellow", &success);
		blue = ColorGetPixelFromString("blue", &success);
		cyan = ColorGetPixelFromString("cyan", &success);
		magenta = ColorGetPixelFromString("magenta", &success);
		orange = ColorGetPixelFromString("orange", &success);
		black = ColorGetPixelFromString("black", &success);
		white = ColorGetPixelFromString("white", &success);
#if 1
		bgcolor = ColorGetPixelFromString("midnightblue", &success);
#else
		bgcolor = ColorGetPixelFromString("grey", &success);
#endif

		font = FontGetFontFromString(FONT, &success);
		if(success == NO)font = FontDefaultFont();
	}
#endif

	get_next_relevant_record("Missing problem identification");
	strcpy(text, line_buffer);

	get_next_relevant_record("Missing overall numbers");
	if(sscanf(line_buffer, "%ld %ld %ld", &tiles, &inclusions, &codes) != 3)
		error_message(INPUT_ERROR, "Invalid record - overall numbers");
	
	if(tiles <= 0 || tiles >= max_tiles)
		error_message(INPUT_ERROR, "Invalid number of Wang tiles: NWT = %ld (0 < NWT < %ld)", tiles, max_tiles);
	if(inclusions < 0)
		error_message(INPUT_ERROR, "Invalid number of reference inclusions: NRI = %ld (0 <= NRI)", inclusions);
	if(codes < 0)
		error_message(INPUT_ERROR, "Invalid number of side codes: NSC = %ld (0 <= NSC)", codes);
	
/* only array of usage of tile ids is allocated to save space */

/* IMPORTANT: I do rely on fact that calloc was used to allocate code_array */

#ifdef ARBITRARY_TILE_ID
	if((tile_array = calloc(tiles, sizeof(long))) == NULL)
		error_message(MEMORY_ERROR, "Memory allocation error");
#else
	if((tile_array = calloc(tiles, sizeof(logic))) == NULL)
		error_message(MEMORY_ERROR, "Memory allocation error");
#endif

	if(inclusions != 0){

/* array of complete data structures allocated */

		if((inclusion_array = calloc(inclusions, sizeof(inclusion_rec))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
	}

	if(codes != 0){

/* array of complete data structures allocated (small number of codes expected */

		if((code_array = calloc(codes, sizeof(code_rec))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
	}
	
/* reference tile must be read before inclusions to have epsilon available */

	get_next_relevant_record("Missing reference tile definition");
	if(sscanf(line_buffer, "%lf %lf %lf %lf %d %lf %lf", 
						&(tile.center.x), &(tile.center.y), &(tile.size.x), &(tile.size.y), 
						&(tile.mat), &(tile.imsz), &(tile.bmsz)) != 7)
		error_message(INPUT_ERROR, "Invalid record - definition of reference tile");

	if(tile.size.x <= 0.0)
		error_message(INPUT_ERROR, "Invalid tile dimension in x-direction: TDX = %e (0 < DX)", tile.size.x);
	if(tile.size.y <= 0.0)
		error_message(INPUT_ERROR, "Invalid tile dimension in y-direction: TDY = %e (0 < DY)", tile.size.y);
	if(tile.mat == 0)
		error_message(INPUT_ERROR, "Invalid matrix material: MMAT = %d (0 < MMAT)", tile.mat);
	if(tile.imsz < 0.0)
		error_message(INPUT_ERROR, "Invalid matrix internal mesh size: MIMSZ = %e (0 < MIMSZ)", tile.imsz);
	if(tile.bmsz < 0.0)
		error_message(INPUT_ERROR, "Invalid matrix boundary mesh size: MBMSZ = %e (0 < MBMSZ)", tile.bmsz);

	min_sz = max_sz = tile.size.x;
	if(tile.size.y < min_sz)min_sz = tile.size.y;
	if(tile.size.y > max_sz)max_sz = tile.size.y;

	if(tile.imsz == 0.0)
		tile.imsz = min_sz;
	else{
		if(tile.imsz > min_sz)tile.imsz = min_sz;
	}
	if(tile.bmsz == 0.0)
		tile.bmsz = min_sz;
	else{
		if(tile.bmsz > min_sz)tile.bmsz = min_sz;
	}

#if 0 /* not needed */
	min_msz = (tile.imsz < tile.bmsz) ? tile.imsz : tile.bmsz;
#endif

	if(epsilon == 0.0){
		epsilon = min_sz * DEF_REL_EPSILON;
		error_message(NOTIFICATION, "Epsilon set to %e", epsilon);
	}
	else if(epsilon < 0.0){
		if(-epsilon >= MAX_REL_EPSILON)error_message(GENERAL_ERROR, "Too large relative epsilon");
		epsilon *= -min_sz;
	}
	else{
		if(epsilon >= min_sz * MAX_REL_EPSILON)error_message(GENERAL_ERROR, "Too large absolute epsilon");
	}

	tile.corner[0].x = tile.center.x - tile.size.x / 2.0;
	tile.corner[0].y = tile.center.y - tile.size.y / 2.0;

	tile.corner[1].x = tile.center.x + tile.size.x / 2.0;
	tile.corner[1].y = tile.center.y + tile.size.y / 2.0;

#ifdef ELIXIR
	if(draw == YES){
		draw_tile(&tile, TILE_LAYER, magenta);
		EVFitAllIntoView(view);
			
#ifdef DEBUG_DRAW
		ESIEventLoop(YES, "Tile\n\n");
#endif
			
	}
#endif

	for(i = 0; i < inclusions; i++){
		k = (i <= 3) ? i : 3;

		get_next_relevant_record("Missing inclusion definition");
		if(sscanf(line_buffer, "%ld", &inclusion_id) != 1)
			error_message(INPUT_ERROR, "Invalid record - definition of the %ld%s inclusion", i + 1, suffix[k]);

		relative = NO;
		if(inclusion_id < 0){
			inclusion_id = -inclusion_id;
			relative = YES;
		}

		if(inclusion_id <= 0 || inclusion_id > inclusions)
			error_message(INPUT_ERROR, "Invalid id of the %ld%s inclusion: IID = %ld (0 < IID <= %ld)", i + 1, suffix[k], inclusion_id, inclusions);

		inclusion = &inclusion_array[inclusion_id - 1];
		if(inclusion -> id != 0)
			error_message(INPUT_ERROR, "Multiple definition of inclusion %ld encountered", inclusion_id);

		ellipse = &(inclusion -> ellipse);
		face_is = &(inclusion -> face_is);
		inclusion -> id = inclusion_id;

/* note: I do rely on fact that inclusion_array was allocated by calloc
	 thus there is no need to initialize face_is */

		if(sscanf(line_buffer, "%ld %lf %lf %lf %lf %lf %lf %lf %lf %d %d", &id,    /* IMPORTANT: do not overwrite inclusion -> id */
							&(ellipse -> center.x), &(ellipse -> center.y),
							&(ellipse -> major.x), &(ellipse -> major.y),
							&(ellipse -> max), &(ellipse -> min),
							&(inclusion -> imsz), &(inclusion -> bmsz),  
							&(inclusion -> mat), &(inclusion -> ifmat)) != 11){
			error_message(INPUT_ERROR, "Invalid record - definition of inclusion %ld", inclusion_id);
		}
		
#ifdef ALLOW_OUTSIDE

/* ellipse center of reference inclusion can be outside tile not more than by epsilon;
	 if outside in particular direction, it is shifted inside in that direction and flag for periodic repetition in that direction is inverted */

		if(ellipse -> center.x < tile.corner[0].x - epsilon || ellipse -> center.x > tile.corner[1].x + epsilon ||
			 ellipse -> center.y < tile.corner[0].y - epsilon || ellipse -> center.y > tile.corner[1].y + epsilon)
			error_message(INPUT_ERROR, "Invalid coordinates of center of inclusion %d", inclusion_id); 
		
		outside = NO;
		if(ellipse -> center.x < tile.corner[0].x || ellipse -> center.x > tile.corner[1].x){
			outside = inclusion -> out[X] = YES;
			if(ellipse -> center.x < tile.center.x)
				ellipse -> center.x += tile.size.x;
			else
				ellipse -> center.x -= tile.size.x;
		}
		if(ellipse -> center.y < tile.corner[0].y || ellipse -> center.y > tile.corner[1].y){
			outside = inclusion -> out[Y] = YES;
			if(ellipse -> center.y < tile.center.y)
				ellipse -> center.y += tile.size.y;
			else
				ellipse -> center.y -= tile.size.y;
		}

		if(outside == YES)error_message(WARNING, "Reference inclusion %ld subjected to periodic repetition (%d %d)", inclusion_id, 
																		(inclusion -> out[X] == YES) ? 1 : 0, (inclusion -> out[Y] == YES) ? 1 : 0);
#else

/* ellipse center of reference inclusion cannot be outside tile;
	 it can be outside only for periodically repeated inclusions which are processed as shifted reference inclusion */

		if(ellipse -> center.x < tile.corner[0].x || ellipse -> center.x > tile.corner[1].x ||
			 ellipse -> center.y < tile.corner[0].y || ellipse -> center.y > tile.corner[1].y)
			error_message(INPUT_ERROR, "Invalid coordinates of center of inclusion %d", inclusion_id); 
#endif
		
		if(ellipse -> max <= 0.0 || ellipse -> max > (min_sz - epsilon) / MAX_RADIUS_RATE)
			error_message(INPUT_ERROR, "Invalid size of major half-axis of inclusion %d: MAX = %e (10 * epsilon = %e <= MAX <= %e)", 
										inclusion_id, ellipse -> max, epsilon * 10.0, (min_sz - epsilon) / MAX_RADIUS_RATE);
		if(ellipse -> min <= epsilon * 10.0 || ellipse -> min > ellipse -> max)
			error_message(INPUT_ERROR, "Invalid size of minor half-axis of inclusion %d: MIN = %e (10 * epsilon = %e < MIN <= %e = MAX)", 
										inclusion_id, ellipse -> min, epsilon * 10.0, ellipse -> max);

		if(inclusion -> imsz < 0.0)
			error_message(INPUT_ERROR, "Invalid internal mesh size IIMSZ of inclusion %ld: IIMSZ = %e (0 <= IIMSZ)", inclusion_id, inclusion -> imsz);
		if(inclusion -> bmsz < 0.0)
			error_message(INPUT_ERROR, "Invalid boundary mesh size IBMSZ of inclusion %ld: IBMSZ = %e (0 <= IBMSZ)", inclusion_id, inclusion -> bmsz);
		if(inclusion -> mat < 0)
			error_message(INPUT_ERROR, "Invalid inclusion material: IMAT = %d (0 <= IMAT)", inclusion_id, inclusion -> mat);
		if(inclusion -> ifmat < 0)
			error_message(INPUT_ERROR, "Invalid interface material: IFMAT = %d (0 <= IFMAT)", inclusion_id, inclusion -> ifmat);

		if(inclusion -> mat == 0)inclusion -> ifmat = 0;

		if(inclusion -> imsz == 0.0)
			inclusion -> imsz = min_sz;
		else{
			if(inclusion -> imsz > min_sz)inclusion -> imsz = min_sz;
		}
		if(inclusion -> bmsz == 0.0)
			inclusion -> bmsz = min_sz;
		else{
			if(inclusion -> bmsz > min_sz)inclusion -> bmsz = min_sz;
		}

#if EXCESS_SIZE == MAX_PRESENT_SIZE
		if(max_rad < ellipse -> max)max_rad = ellipse -> max;
#endif

#if 0 /* not needed */
		if(min_msz > inclusion -> bmsz)min_msz = inclusion -> bmsz;
		if(ellipse -> mat != 0){
			if(min_msz > inclusion -> imsz)min_msz = inclusion -> imsz;
		}
#endif

		if(relative == NO)sub_vec(ellipse -> major, ellipse -> major, ellipse -> center);
		size = sqrt(size_vec(ellipse -> major));
		if(size < EPSILON)
			error_message(INPUT_ERROR, "Zero major axis of inclusion %ld", inclusion_id);
		div_vec(ellipse -> major, size);
		ellipse -> minor.x = -ellipse -> major.y;
		ellipse -> minor.y = ellipse -> major.x;

#ifdef PARAMETRIC_ROUTINES
		copy_mul_vec(vec_max, ellipse -> major, ellipse -> max);
		copy_mul_vec(vec_min, ellipse -> minor, ellipse -> min);

		add_vec(ellipse -> pnt0, ellipse -> center, vec_max);
		add_vec(ellipse -> pnt1, ellipse -> pnt0, vec_min);
		add_vec(ellipse -> pnt2, ellipse -> center, vec_min);
#endif

		inclusion -> corner = OUT;
	}

/* compute min_epsilon after min_sz, max_sz and max_rad have been determined and before inclusions are going to be processed
	 (namely before get_inclusion_tile_intersection is called);
	 therefore all inclusions are firstly read in to allow for determination of max_rad and after that they are processed (computing intersection);
	 this implies that I cannot consider real excess of present inclusions out of the tile, because the real excess can be computed in
	 get_inclusion_tile_intersection, which requires min_epsilon, which is dependent to the overall problem size, which is in turn dependent 
	 on the offset, which is finally dependent on the excess */

#if SETUP == GRID
	num_tiles_1D = grid_size = (int)(sqrt(tiles - 0.5)) + 1;
#endif
#if SETUP == DIAGONAL
	num_tiles_1D = tiles;
#endif

#if EXCESS_SIZE == MAX_ALLOWED_SIZE
	max_rad = (min_sz - epsilon) / MAX_RADIUS_RATE;
#endif

/* virtual curves may exceed out of the tile only up the half of the whole size of inclusion */

	excess_size = max_rad;

/* I can allow for min_sz based offset because all imsz and bmsz are guaranteed to be min_sz or smaller unless not modified manually !!! */

/* IMPORTANT: it is dangerous to compute offset from min_msz, because min_msz does not account for curvature based refinement !!! */

/*     o-------------o                o-------------o
       |             |                |             |
     __o             o--_           __o             o--_
    -  |             |   |         -  |             |   |
   |   |             |__-         |   |             |__-        e_sz = excess_size
    -__|             o             -__|             o           o_sz = OFFSET * min_sz
       o             |                o             |
       |             |                |             |
       o-------------o                o-------------o

    e_sz    max_sz    e_sz  o_sz  e_sz     max_sz    e_sz
   |---|-------------|---|--------|---|-------------|---|
   |   |                              |             |   |
   |   |------------------------------|             |   |
   |   |             offset                         |   |
   |   |                                            |   |
   |   |--------------------------------------------|   |
	 |          root_size - minimal (obsolete)            |
   |                                                    |
   |----------------------------------------------------|
	            root_size - maximal (obsolete) */

	offset = max_sz + OFFSET * min_sz + 2.0 * excess_size;

#ifdef TILE_OFFSET
	if(tiles != 1){
		if(max_sz - min_sz < max_sz * EPSILON)error_message(WARNING, "TILE_OFFSET applied to tile of not square shape");
	}

	if(offset < 2.0 * max_sz)offset = 2.0 * max_sz;
#else
/* it may be desirable to round up the offset to 4 or 5 valid digits */
#endif

	if(tiles == 1 || codes == 0){

/* if codes == 0 dispersion is irrelevant as tile sides are not mirrored;
	 otherwise root octant must be shifted so that tile is centered in the root octant !!!
	 this should prevent octree dispersion when opposite sides of tile are mirrored */

/* note: if tiles == 1 and minimal root_size = max_sz is used, t3d may report change of root origin and root size
	 because the computed domain extent may not exactly coincide (within tolerance < t3d_epsilon) with the prescribed root octant */

		root_size = num_tiles_1D * max_sz + (num_tiles_1D - 1) * (offset - max_sz);                       // minimal
		//root_size = num_tiles_1D * max_sz + (num_tiles_1D - 1) * (offset - max_sz) + 2.0 * excess_size;   // maximal

/* it may be desirable to round up the root_size to 4 or 5 valid digits */

	}
	else{

/* in order to eliminate dispersion of mesh size due to octree mirroring of tile sides it is necessary to ensure that
	 each tile is placed at the same relative position with respect to the smallest octant containing the whole tile !!!
	 this implies that size of root must be 2^k x offset, where 2^k >= num_tiles_1D;
	 although this leads to larger (smaller) root (min_epsilon) than it is really necessary, the elimination of dispersion is of the highest priority !!!
	 root octant must be shifted so that each tile is centered in the smallest octant containing the whole tile !!!
	 this should prevent octree dispersion when opposite sides of tile are mirrored */

		power = 1;
		while(power < num_tiles_1D)power *= 2;
		
		root_size = power * offset;
	}
	
/* compute the epsilon as t3d does but make it 10 times larger */

	min_epsilon = root_size;
	double_size = 2.0 * (sizeof(double) - 1) - 1;
	while(double_size--)min_epsilon /= 10.0;
	
/* make epsilon at least min_epsilon * 10 */

	if(epsilon < min_epsilon * 10.0){
		epsilon = min_epsilon * 10.0;
		error_message(NOTIFICATION, "Epsilon increased to min epsilon %e", min_epsilon * 10.0);
		if(max_subdivisions != 0){
			max_subdivisions = 0;
			error_message(NOTIFICATION, "Number of subdivisions reduced to 0");
		}
	}

#ifdef REDUCED_ESTIMATE

/* prepare histogram */

	if((min_area = (double *)calloc(INTERVALS_2D, sizeof(double))) == NULL)error_message(MEMORY_ERROR, "Memory allocation error");
	if((min_x_length = (double *)calloc(INTERVALS_1D, sizeof(double))) == NULL)error_message(MEMORY_ERROR, "Memory allocation error");
	if((min_y_length = (double *)calloc(INTERVALS_1D, sizeof(double))) == NULL)error_message(MEMORY_ERROR, "Memory allocation error");

	if((sum_area_weight = (double *)calloc(INTERVALS_2D, sizeof(double))) == NULL)error_message(MEMORY_ERROR, "Memory allocation error");
	if((sum_x_length_weight = (double *)calloc(INTERVALS_1D, sizeof(double))) == NULL)error_message(MEMORY_ERROR, "Memory allocation error");
	if((sum_y_length_weight = (double *)calloc(INTERVALS_1D, sizeof(double))) == NULL)error_message(MEMORY_ERROR, "Memory allocation error");

	if((sum_area = (double *)calloc(INTERVALS_2D, sizeof(double))) == NULL)error_message(MEMORY_ERROR, "Memory allocation error");
	if((sum_x_length = (double *)calloc(INTERVALS_1D, sizeof(double))) == NULL)error_message(MEMORY_ERROR, "Memory allocation error");
	if((sum_y_length = (double *)calloc(INTERVALS_1D, sizeof(double))) == NULL)error_message(MEMORY_ERROR, "Memory allocation error");

	if((area_interval = (double *)calloc(INTERVALS_2D, sizeof(double))) == NULL)error_message(MEMORY_ERROR, "Memory allocation error");
	if((x_length_interval = (double *)calloc(INTERVALS_1D, sizeof(double))) == NULL)error_message(MEMORY_ERROR, "Memory allocation error");
	if((y_length_interval = (double *)calloc(INTERVALS_1D, sizeof(double))) == NULL)error_message(MEMORY_ERROR, "Memory allocation error");

	if((area_count = (long *)calloc(INTERVALS_2D, sizeof(long))) == NULL)error_message(MEMORY_ERROR, "Memory allocation error");
	if((x_length_count = (long *)calloc(INTERVALS_1D, sizeof(long))) == NULL)error_message(MEMORY_ERROR, "Memory allocation error");
	if((y_length_count = (long *)calloc(INTERVALS_1D, sizeof(long))) == NULL)error_message(MEMORY_ERROR, "Memory allocation error");

/* I am storing upper bounds of individual intervals */

#if SERIES_2D == CONSTANT_SERIES
	b = 0.0;
	val = a = tile.size.x * tile.size.y / INTERVALS_2D;
	for(i = 0; i < INTERVALS_2D; i++){
		area_interval[i] = val;
		val += a;
	}
#endif
#if SERIES_2D == ARITHMETIC_SERIES
	val = a = tile.size.x * tile.size.y / (INTERVALS_2D * (1.0 + (INTERVAL_RATE_2D - 1.0) * 0.5));
	b = a * (INTERVAL_RATE_2D - 1.0) / (INTERVALS_2D - 1);
	for(i = 0; i < INTERVALS_2D; i++){
		area_interval[i] = val;
		a += b;
		val += a;
	}
#endif
#if SERIES_2D == GEOMETRIC_SERIES
	b = exp(log(INTERVAL_RATE_2D) / (INTERVALS_2D - 1));
	if((val = pow(b, INTERVALS_2D) - 1.0) < 1.0e-10){
		val = a = tile.size.x * tile.size.y / INTERVALS_2D;
		b = 1.0;
	}
	else
		val = a = tile.size.x * tile.size.y * (b - 1.0) / val;

	for(i = 0; i < INTERVALS_2D; i++){
		area_interval[i] = val;
		a *= b;
		val += a;
	}
#endif

#if SERIES_1D == CONSTANT_SERIES
	b = 0.0;
	val = a = tile.size.x / INTERVALS_1D;
	for(i = 0; i < INTERVALS_1D; i++){
		x_length_interval[i] = val;
		val += a;
	}
	val = a = tile.size.y / INTERVALS_1D;
	for(i = 0; i < INTERVALS_1D; i++){
		y_length_interval[i] = val;
		val += a;
	}
#endif
#if SERIES_1D == ARITHMETIC_SERIES
	val = a = tile.size.x / (INTERVALS_1D * (1.0 + (INTERVAL_RATE_1D - 1.0) * 0.5));
	b = a * (INTERVAL_RATE_1D - 1.0) / (INTERVALS_1D - 1);
	for(i = 0; i < INTERVALS_1D; i++){
		x_length_interval[i] = val;
		a += b;
		val += a;
	}
	val = a = tile.size.y / (INTERVALS_1D * (1.0 + (INTERVAL_RATE_1D - 1.0) * 0.5));
	b = a * (INTERVAL_RATE_1D - 1.0) / (INTERVALS_1D - 1);
	for(i = 0; i < INTERVALS_1D; i++){
		y_length_interval[i] = val;
		a += b;
		val += a;
	}
#endif
#if SERIES_1D == GEOMETRIC_SERIES
	b = exp(log(INTERVAL_RATE_1D) / (INTERVALS_1D - 1));
	if((val = pow(b, INTERVALS_1D) - 1.0) < 1.0e-10){
		val = a = tile.size.x / INTERVALS_1D;
		b = 1.0;
	}
	else
		val = a = tile.size.x * (b - 1.0) / val;
	for(i = 0; i < INTERVALS_1D; i++){
		x_length_interval[i] = val;
		a *= b;
		val += a;
	}
	b = exp(log(INTERVAL_RATE_1D) / (INTERVALS_1D - 1));
	if((val = pow(b, INTERVALS_1D) - 1.0) < 1.0e-10){
		val = a = tile.size.y / INTERVALS_1D;
		b = 1.0;
	}
	else
		val = a = tile.size.y * (b - 1.0) / val;
	for(i = 0; i < INTERVALS_1D; i++){
		y_length_interval[i] = val;
		a *= b;
		val += a;
	}
#endif

	for(i = 0; i < INTERVALS_2D; i++)min_area[i] = area_interval[i];
	for(i = 0; i < INTERVALS_1D; i++)min_x_length[i] = x_length_interval[i];
	for(i = 0; i < INTERVALS_1D; i++)min_y_length[i] = y_length_interval[i];
#endif

	inclusion = inclusion_array;
	for(i = 0; i < inclusions; i++, inclusion++){
		ellipse = &(inclusion -> ellipse);
		face_is = &(inclusion -> face_is);

#ifdef REDUCED_ESTIMATE
		area = PI_1 * ellipse -> max * ellipse -> min;
		if(area != 0.0){
			for(j = 0; j < INTERVALS_2D; j++){
				if(area < area_interval[j]){
					area_count[j]++;
					if(area < min_area[j])min_area[j] = area;
					sum_area[j] += area;
					sum_area_weight[j] += 1.0 / area;
					break;
				}
			}
		}
#endif

		if(get_inclusion_tile_intersection() == YES){
			if(inclusion -> mat == 0 || inclusion -> ifmat != 0){
				if(inclusion -> corner == IN)inside_hole = YES;
#ifdef CHECK_ISOLATED
				if(inclusion -> corner == BND){
					if(x_count == 2 && y_count == 2)boundary_hole = YES;
				}
#endif
			}

/* store ellipse extreme points */

			if((pnt_extreme = (point_rec *)calloc(4, sizeof(point_rec))) == NULL)
				error_message(MEMORY_ERROR, "Memory allocation error");
			face_is -> extreme_pnt = pnt_extreme;
			for(j = 0; j < 2; j++){
				copy_vec(pnt_extreme[2 * X + j], ext_point_x[j]);
				copy_vec(pnt_extreme[2 * Y + j], ext_point_y[j]);
			}
						
			face_is -> weight = get_ellipse_weight(ellipse, &pnt_extreme[2 * X + 0], &pnt_extreme[2 * Y + 1], &trans_g2l);

/* store edge intersection */

			if(x_count != 0){
				if((inclusion -> edge_is[X] = (edge_is_rec *)malloc(sizeof(edge_is_rec))) == NULL)
					error_message(MEMORY_ERROR, "Memory allocation error");
				inclusion -> edge_is[X] -> count = x_count;
				if((inclusion -> edge_is[X] -> point = (is_point_rec *)calloc(x_count, sizeof(is_point_rec))) == NULL)
					error_message(MEMORY_ERROR, "Memory allocation error");
				for(j = 0; j < x_count; j++){
					is_point = &(inclusion -> edge_is[X] -> point[j]);
					is_point -> data.coord = x_coord[j];
				}
			
/* resolve position of intersection points with respect to tile corner;
	 resolve status of intersection points;
	 resolve ellipse edge halving */

				if(x_count == 1){
					inclusion -> edge_is[X] -> point[0].data.status = TOUCH;
					
					if(inclusion -> corner == BND)
						inclusion -> edge_is[X] -> side = INDEF;
					else
						inclusion -> edge_is[X] -> side = (x_coord[0] > tile.corner[ind_x].x) ? POS : NEG;

#ifdef REDUCED_ESTIMATE
					x_touch_count++;
#else
#ifdef CHECK_COMPATIBILITY
					x_touch_count++;
#endif
#endif
				}
				else{
					inclusion -> edge_is[X] -> point[0].data.status = START;
					inclusion -> edge_is[X] -> point[1].data.status = END;
				
/* o----x-----x------------o   ==> POS
	 o---------------x----x--o   ==> NEG
   -----x---o----x----------   ==> BOTH

	 o - corner
	 x - edge_is */

					if(x_coord[0] > tile.corner[ind_x].x - min_epsilon * 10.0)
						inclusion -> edge_is[X] -> side = POS;
					else if(x_coord[1] < tile.corner[ind_x].x + min_epsilon * 10.0)
						inclusion -> edge_is[X] -> side = NEG;
					else{
#ifdef DEBUG
						if(inclusion -> corner != IN || y_count != 2)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
						inclusion -> edge_is[X] -> side = BOTH;

/* IMPORTANT: update coordinate of edge_is point outside tile to fall inside tile by appropriate periodic repetition;
	 this is necessary for coordinate-based ordering */

						inclusion -> edge_is[X] -> point[ind_x].data.coord += tile.size.x * sign[ind_x];
					}

					pnt_1.x = x_coord[0];
					pnt_2.x = x_coord[1];
					pnt_1.y = pnt_2.y = tile.corner[ind_y].y;

					get_ellipse_normal_point(ellipse, &pnt_1, &normal1);
					get_ellipse_normal_point(ellipse, &pnt_2, &normal2);

					if(-dot_product(normal1, normal2) > ZERO_ANGLE)face_is -> edge_half[X] = YES;

#ifdef REDUCED_ESTIMATE
					x_length = x_coord[1] - x_coord[0];
					for(j = 0; j < INTERVALS_1D; j++){
						if(x_length < x_length_interval[j]){
							x_length_count[j]++;
							if(x_length < min_x_length[j])min_x_length[j] = x_length;
							sum_x_length[j] += x_length;
							sum_x_length_weight[j] += 1.0 / x_length;
							break;
						}
					}
#endif
				}
			}
			if(y_count != 0){
				if((inclusion -> edge_is[Y] = (edge_is_rec *)malloc(sizeof(edge_is_rec))) == NULL)
					error_message(MEMORY_ERROR, "Memory allocation error");
				inclusion -> edge_is[Y] -> count = y_count;
				if((inclusion -> edge_is[Y] -> point = (is_point_rec *)calloc(y_count, sizeof(is_point_rec))) == NULL)
					error_message(MEMORY_ERROR, "Memory allocation error");
				for(j = 0; j < y_count; j++){
					is_point = &(inclusion -> edge_is[Y] -> point[j]);
					is_point -> data.coord = y_coord[j];
				}
				
/* resolve position of intersection points with respect to tile corner;
	 resolve status of intersection points;
	 resolve ellipse edge halving */

				if(y_count == 1){
					inclusion -> edge_is[Y] -> point[0].data.status = TOUCH;
					
					if(inclusion -> corner == BND)
						inclusion -> edge_is[Y] -> side = INDEF;
					else
						inclusion -> edge_is[Y] -> side = (y_coord[0] > tile.corner[ind_y].y) ? POS : NEG;

#ifdef REDUCED_ESTIMATE
					y_touch_count++;
#else
#ifdef CHECK_COMPATIBILITY
					y_touch_count++;
#endif
#endif
				}
				else{
					inclusion -> edge_is[Y] -> point[0].data.status = START;
					inclusion -> edge_is[Y] -> point[1].data.status = END;
				
					if(y_coord[0] > tile.corner[ind_y].y - min_epsilon * 10.0)
						inclusion -> edge_is[Y] -> side = POS;
					else if(y_coord[1] < tile.corner[ind_y].y + min_epsilon * 10.0)
						inclusion -> edge_is[Y] -> side = NEG;
					else{
#ifdef DEBUG
						if(inclusion -> corner != IN || x_count != 2)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
						inclusion -> edge_is[Y] -> side = BOTH;

/* IMPORTANT: update coordinate of edge_is point outside tile to fall inside tile by appropriate periodic repetition;
	 this is necessary for coordinate-based ordering */

						inclusion -> edge_is[Y] -> point[ind_y].data.coord += tile.size.y * sign[ind_y];
					}

					pnt_1.y = y_coord[0];
					pnt_2.y = y_coord[1];
					pnt_1.x = pnt_2.x = tile.corner[ind_x].x;
					
					get_ellipse_normal_point(ellipse, &pnt_1, &normal1);
					get_ellipse_normal_point(ellipse, &pnt_2, &normal2);
					
					if(-dot_product(normal1, normal2) > ZERO_ANGLE)face_is -> edge_half[Y] = YES;

#ifdef REDUCED_ESTIMATE
					y_length = y_coord[1] - y_coord[0];
					for(j = 0; j < INTERVALS_1D; j++){
						if(y_length < y_length_interval[j]){
							y_length_count[j]++;
							if(y_length < min_y_length[j])min_y_length[j] = y_length;
							sum_y_length[j] += y_length;
							sum_y_length_weight[j] += 1.0 / y_length;
							break;
						}
					}
#endif
				}
			}

/* resolve cross halving;
	 IMPORTANT: I do rely on fact that status (POS, NEG, BOTH) of intersection points is already resolved
	 and that inclusion -> corner (OUT, IN, BND) is already resolved */

/* note: intersection points checked against cross halving are related to relative position of center of intersection ellipse with respect to tile corner */

			if(x_count != 0 && y_count != 0){
				halving = NO;

				pnt_1.y = tile.corner[ind_y].y;
				pnt_2.x = tile.corner[ind_x].x;

				if(x_count == 2 && y_count == 2){
					if(inclusion -> corner == BND){
						pnt_1.x = x_coord[2 - inclusion -> edge_is[X] -> side];
						pnt_2.y = y_coord[2 - inclusion -> edge_is[Y] -> side];
					}
					else{
						if(ellipse -> center.x > tile.corner[ind_x].x)
							pnt_1.x = x_coord[1];
						else
							pnt_1.x = x_coord[0];
						if(ellipse -> center.y > tile.corner[ind_y].y)
							pnt_2.y = y_coord[1];
						else
							pnt_2.y = y_coord[0];
					}
					
					get_ellipse_normal_point(ellipse, &pnt_1, &normal1);
					get_ellipse_normal_point(ellipse, &pnt_2, &normal2);
					
					if(-dot_product(normal1, normal2) > ZERO_ANGLE)halving = YES;
				}
				else{
					if(x_count == 2 && y_count == 1){
						pnt_1.x = x_coord[2 - inclusion -> edge_is[X] -> side];
						pnt_2.y = y_coord[0];
						
						get_ellipse_normal_point(ellipse, &pnt_1, &normal1);
						get_ellipse_normal_point(ellipse, &pnt_2, &normal2);
						
						if(-dot_product(normal1, normal2) > ZERO_ANGLE)halving = YES;
					}
					else if(x_count == 1 && y_count == 2){
						pnt_1.x = x_coord[0];
						pnt_2.y = y_coord[2 - inclusion -> edge_is[Y] -> side];
						
						get_ellipse_normal_point(ellipse, &pnt_1, &normal1);
						get_ellipse_normal_point(ellipse, &pnt_2, &normal2);
						
						if(-dot_product(normal1, normal2) > ZERO_ANGLE)halving = YES;
					}
				}
				
				if((face_is -> cross_half = halving) == YES){

/* choose from directions extreme_x, extreme_y, min, max which is most perpendicular to connection pnt_1 and pnt_2 */

					val = 1.0;
					sub_vec(vec, pnt_2, pnt_1);
					sz = sqrt(size_vec(vec));
					div_vec(vec, sz);

#if 0
					p1 = &(face_is -> extreme_pnt[2 * X + 0]);
					p2 = &(face_is -> extreme_pnt[2 * X + 1]);
#else
					p1 = &(ext_point_x[0]);
					p2 = &(ext_point_x[1]);
#endif
					sub_vec(vec_tmp, *p2, *p1);
					sz = sqrt(size_vec(vec_tmp));
					div_vec(vec_tmp, sz);
					
					product = dot_product(vec, vec_tmp);
					product = fabs(product);
					if(product < val){
						ind = X;       /* halving with respect to x-extreme point */
						val = product;
					}
					
#if 0
					p1 = &(face_is -> extreme_pnt[2 * Y + 0]);
					p2 = &(face_is -> extreme_pnt[2 * Y + 1]);
#else
					p1 = &(ext_point_y[0]);
					p2 = &(ext_point_y[1]);
#endif
					sub_vec(vec_tmp, *p2, *p1);
					sz = sqrt(size_vec(vec_tmp));
					div_vec(vec_tmp, sz);

					product = dot_product(vec, vec_tmp);
					product = fabs(product);
					if(product < val){
						ind = Y;       /* halving with respect to y-extrme point */
						val = product;
					}
					
#ifdef MIN_MAX_HALVING
					product = dot_product(vec, ellipse -> minor);
					product = fabs(product);
					if(product < val){
						ind = MIN;       /* halving with respect to min diameter */
						val = product;
					}
					
					product = dot_product(vec, ellipse -> major);
					product = fabs(product);
					if(product < val){
						ind = MAX;       /* halving with respect to max diameter */
						val = product;
					}
#endif
					face_is -> cross_halving = ind;
				}
			}

#ifdef REDUCED_ESTIMATE
			if(x_intersection == YES && y_intersection == YES){
				if(inclusion -> corner == IN)
					corner_in = YES;
				else if(inclusion -> corner == BND)
					corner_bnd = YES;
			}
			else{
				if((x_intersection == YES && y_touching == YES && inclusion -> edge_is[X] == INDEF) || 
					 (y_intersection == YES && x_touching == YES && inclusion -> edge_is[Y] == INDEF))corner_bnd = YES;
			}
#else

/* WARNING: max_numbers can overflow to zero if too large total number of inclusions is processed
	 or if too large number of inclusions intersects or touches tiles sides */

/* number of potential instances of an inclusion "in" tile equals to 2^(number_of_side_intersections)
	 decreased by 1 if inclusion intersects all tile sides but tile corner is outside inclusion */

			count = 0;
			if(x_intersection == YES)count++;
			if(y_intersection == YES)count++;

			max_tile_patches += 1 << count;
			
			if(count == 2){
				if(inclusion -> corner == OUT)max_tile_patches -= 1;
			}

#if 0 /* alternative way how to compute max_edge_curves_1 */
			if(count != 0){

/* number of potential intersections of an inclusion with tile side equals to 2^(number_of_other_side_intersections)
	 decreased by 1 if inclusion intersects all tile sides but tile corner is outside inclusion */
		
				cnt = 1 << (count - 1);
				if(count == 2){
					if(inclusion -> corner == OUT)cnt--;
				}
			
				if(x_intersection == YES)max_edge_curves_1[X] += cnt;
				if(y_intersection == YES)max_edge_curves_1[Y] += cnt;
			}			
#endif

/* number of potential points on tile edge
	 periodic repetitions are naturally accounted for;
	 intersection points coinciding with tile corners should not be acounted for but this is NOT done !!! */

			max_edge_vertices[X] += x_count;
			max_edge_vertices[Y] += y_count;

/* number of potential inclusion segments on tile edge;
	 number of potential segments between inclusions on tile edge;
	 periodic repetition must be accounted for edge intersection side == BOTH */

			for(j = 0; j < 2; j++){
				if(inclusion -> edge_is[j] == NULL)continue;
				switch(inclusion -> edge_is[j] -> side){
				case POS:
				case NEG:
					if(inclusion -> edge_is[j] -> count == 2)max_edge_curves_1[j]++;
					
/* note: max_edge_curves_2 should not be incremented if one of 2 intersections with tile edge is passing through tile corner
	 but this is NOT done */

					max_edge_curves_2[j]++;
					break;
				case BOTH:
					max_edge_curves_1[j] += 2;
					break;
				default:
					break;
				}
			}
#endif
		}
#ifndef REDUCED_ESTIMATE
		else{

/* WARNING: max_number can overflow to zero if too large total number of inclusions is processed */

			max_tile_patches++;
		}
#endif

/* IMPORTANT: inclusion can be shifted even if there is no intersection with tile (eg. during overdetouch) */

		size = size_vec(total_shift);
		if(size != 0.0){
			inclusion -> shift = shift_done = YES;
			error_message(WARNING, "Inclusion %ld shifted by [%e %e]", inclusion -> id, total_shift.x, total_shift.y);

#ifdef DEBUG
			if(fabs(total_shift.x) > MULTIPLE_EPS * epsilon)error_message(WARNING, "Shift in x-direction violates maximal permissible shift"); 
			if(fabs(total_shift.y) > MULTIPLE_EPS * epsilon)error_message(WARNING, "Shift in y-direction violates maximal permissible shift"); 
#endif
		}

#ifdef ELIXIR
		if(draw == YES){
			draw_inclusion(inclusion, INCLUSION_LAYER, red);
			
#ifdef DEBUG_DRAW
			ESIEventLoop(YES, "Inclusion\n\n");
#endif
			
		}
#endif

	}

#ifdef REDUCED_ESTIMATE

/* IMPORTANT: max numbers are very precise estimated if reference inclusions do not mutually overlap;
	 if the reference inclusions are allowed to overlap (which is normally the case), max numbers are safe (large enough) but potentionally unrealistically large;
	 this leads to enlarged memory consumption, premature failure due to too large problem (although this may be not true)
	 or segmentation fault if too large problem is not detected;
	 problem of unrealistic estimate may be solved by calculating max numbers with respect to the measure of the geometrical entity 
	 (area of tile, length of tile edge) that will be occupied by inclusions;
	 for example it makes no sense to consider in max_tile_patches more inclusions that is the number of smallest inclusions sum of which
	 areas is equal to the area of the tile;
	 thus max number is estimated as number of ordered (from min to max) inclusion areas (intersection segment lengths) sum of which excceds
	 the tile area (tile edge length);
	 this would however require ordering of all areas of inclusions from mon;
	 this is solved by creating a histogram of nonzero areas/lengths of intesections on several intervals while storing simultaneously 
	 the smallest nonzero area/length, the sum of areas/lengths and optionally the sum of area/length weights for each interval;
	 max number can be then safely estimated as number of items in individual area/length intervals (ordered from min to max) needed to cover the
	 tile area/edge length plus number of detected zero area/length intersections;
	 for example let the tile edge of length 10 is divided into 10 equal size intervals (corresponding to the size of intersection with tile edge);
	 the histogram looks like this:
	 interval | count | min length | sum length
	 ------------------------------------------
	 0 - 1      10      0.2          5         
	 1 - 2      8       1.05         12
	 2 - 3      4       2.1          10
	 3 - 10     0                    0
	 to cover tile edge of length 10 it is needed all 10 intersection segments from interval 0-1 (of sum length 5) plus 5 intersection segments 
	 from interval 1-2 of min length 1.05 resulting in max number 10 instead of 10+8+4=22;
	 the more size intervals, the better estimate;
	 instead of using min length for the last relevant interval it is possible to use weighted length (with weight proportional to inverse of the length);
	 this will further reduce the estimate but it is not 100% safe;
	 although this approach accounts for overlapping of inclusions in an global sense, it does not account for overlapping in local sense;
	 for example, let's consider tile edge of length 10, which is intersected by 100 reference inclusions of the same size and at the same position
	 in the reference tile;
	 let the intersection of the inclusion with the tile edge has length 1;
	 if reduced estimate is not used, max number of intersections of the tile edge is estimated to 100 (all intersecting inclusions are considered);
	 if reduced estimate is used, max number of intersections of the tile edge is estimated to 10 (10 smallest intersection lengths covers length 10);
	 but it would be enough to estimate max number of intersections of the tile edge as 1 (because all intersecting inclusions are at the same location);
	 the overlapping in local sense can be accounted for if the reduced estimate is performed independently over several equal size space segments 
	 (corresponding to the location on the tile or tile edge);
	 let the 100 inclusions are exactly in the middle of the tile edge and let the tile edge is divided into 3 equal segments 
	 and each segments is divided into 10 intervals;
	 segment | interval | count | min length | sum length
	 ----------------------------------------------------
	 1         0 - 10     0                    0
   2         0 - 1      100     1            100
   2         1 - 10     0                    0
   3         0 - 10     0                    0
	 to cover the 1st space segments, 0 intersection segments are needed;
	 to cover the 2nd space segments, 3.333 intersection segments are needed;
	 to cover the 3rd space segments, 0 intersection segments are needed;
	 thus max number of tile edges intersections is estimated to 3.333 --> 4;
	 the more space segments the better estimate;
	 space segments are not implemented */
	 
	area = tile.size.x * tile.size.y;
	sum = 0.0;
	max_tile_patches = 0;
	for(i = 0; i < INTERVALS_2D; i++){
		if((sum += sum_area[i]) > area){
			sum -= sum_area[i];
			max_tile_patches += (long)((area - sum) / min_area[i]) + 1;
			break;
		}
		max_tile_patches += area_count[i];
	}

	x_length = tile.size.x;
	sum = 0.0;
	max_edge_curves_1[X] = 0;
	for(i = 0; i < INTERVALS_1D; i++){
		if((sum += sum_x_length[i]) > x_length){
			sum -= sum_x_length[i];
			max_edge_curves_1[X] += (long)((x_length - sum) / min_x_length[i]) + 1;
			break;
		}
		max_edge_curves_1[X] += x_length_count[i];
	}

	y_length = tile.size.y;
	sum = 0.0;
	max_edge_curves_1[Y] = 0;
	for(i = 0; i < INTERVALS_1D; i++){
		if((sum += sum_y_length[i]) > y_length){
			sum -= sum_y_length[i];
			max_edge_curves_1[Y] += (long)((y_length - sum) / min_y_length[i]) + 1;
			break;
		}
		max_edge_curves_1[Y] += y_length_count[i];
	}

/* IMPORTANT: account for periodic repetitions !!!;
	 note that I may account for repetition of an inclusion not actually included in max_number */

/* for nonoverlapping inclusions the estimates of max numbers should be exact (except one case) 
	 if all relevant inclusions are periodically repeated in all relevant directions (see estimate2d_[abc].geom);
	 the exception occurs when inclusion is passing through corner while touching one side (see estimate2d_d.geom);
	 then max_edge_curves_2 for that side is larger by 1;
	 consequently also max_edge_vertices for that side is larger by 1 */

	max_tile_patches += max_edge_curves_1[X] + max_edge_curves_1[Y];
	max_edge_curves_2[X] = max_edge_curves_1[X];
	max_edge_curves_2[Y] = max_edge_curves_1[Y];
	if(corner_in == YES){
		max_tile_patches++;
		max_edge_curves_1[X]++;
		max_edge_curves_1[Y]++;
	}
	else if(corner_bnd != YES){
		max_edge_curves_2[X]++;
		max_edge_curves_2[Y]++;
	}
		
	max_edge_curves_2[X] += x_touch_count * 2;
	max_edge_curves_2[Y] += y_touch_count * 2;

	max_edge_vertices[X] = max_edge_curves_1[X] + max_edge_curves_2[X] - 1;
	max_edge_vertices[Y] = max_edge_curves_1[Y] + max_edge_curves_2[Y] - 1;

#endif
	
/* IMPORTANT: allocate the array larger by 1 item so that there is no segmentation fault failure when max_number is once exceeded */

/* note: max_side_curves[X|Y] can be legally larger than max_side_intersections
	 because not all potential intersections will occur (not all inclusions must be used, not all periodic repetitions must be used) */

/* note: there may occur memory allocation failure despite the fact that the problem can fit in the memory;
	 this due to the fact that max_numbers are likely to be larger than the really needed dimension of the array
	 because not all inclusions must be used and not all periodic repetitions must be used */

/* note: even if max_number + 1 is negative (due to overflow), it does not mean that the problem does not fit in the memory;
	 this due to the fact that max_numbers are likely to be larger than the really needed dimension of the array
	 because not all inclusions must be used and not all periodic repetitions must be used;
	 in such a case it may be enough to allocate memory for LONG_MAX - 1 but this is likely to result in memory allocation failure;
	 this is how much memory would be required (on 32-bit machine) if LONG_MAX is used for all arrays:
	 
	 tile_patch_array = 4 x LONG_MAX

#REVISE

	 x_side_curve_array = 4 x LONG_MAX x 2
	 y_side_curve_array = 4 x LONG_MAX x 2

	 x_edge_vertex_id_array = 4 x LONG_MAX x 4
	 y_edge_vertex_id_array = 4 x LONG_MAX x 4

	 x_edge_vertex_data_array = (8 + 4) x LONG_MAX x 4
	 y_edge_vertex_data_array = (8 + 4) x LONG_MAX x 4

	 x_edge_curve_1_array = 4 x LONG_MAX x 4
	 y_edge_curve_1_array = 4 x LONG_MAX x 4

	 x_edge_curve_2_array = 4 x LONG_MAX x 4
	 y_edge_curve_2_array = 4 x LONG_MAX x 4

	 suma = LONG_MAX x 340 = 2147483647 x 340 = 7.3 x 10^11 = 680 GB */

/* check that (appropriatelly enlarged) max numbers are positive if represented as signed numbers */

	if((long)(max_tile_patches + 1) < 0)error_message(GENERAL_ERROR, "Too large problem");
	for(i = 0; i < 2; i++){
		if((long)(max_edge_vertices[i] + 2) < 0)error_message(GENERAL_ERROR, "Too large problem");
		if((long)(max_edge_curves_1[i] + 1) < 0)error_message(GENERAL_ERROR, "Too large problem");
		if((long)(max_edge_curves_2[i] + 1) < 0)error_message(GENERAL_ERROR, "Too large problem");
	}
	
	if((tile_patch_array = (long *)calloc(max_tile_patches + 1, sizeof(long))) == NULL)
		error_message(MEMORY_ERROR, "Memory allocation error");
	
/* IMPORTANT: allocate  x|y_edge_vertex_id_array and x|y_edge_vertex_data_array larger by 2 items 
	 so that there is no segmentation fault failure when max_number is twice exceeded;
	 note that when processing single side intersection, up to 2 intersection points on a single axis may be processed */

	for(i = 0; i < 2; i++){
		if((x_edge_vertex_id_array[i] = (long *)calloc(max_edge_vertices[X] + 2, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((y_edge_vertex_id_array[i] = (long *)calloc(max_edge_vertices[Y] + 2, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");

		if((x_edge_vertex_data_array[i] = (is_pnt_rec **)calloc(max_edge_vertices[X] + 2, sizeof(is_pnt_rec *))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((y_edge_vertex_data_array[i] = (is_pnt_rec **)calloc(max_edge_vertices[Y] + 2, sizeof(is_pnt_rec *))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");

		if((x_edge_curve_1_array[i] = (long *)calloc(max_edge_curves_1[X] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((y_edge_curve_1_array[i] = (long *)calloc(max_edge_curves_1[Y] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");

		if((x_edge_curve_2_array[i] = (long *)calloc(max_edge_curves_2[X] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((y_edge_curve_2_array[i] = (long *)calloc(max_edge_curves_2[Y] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
	}

#ifdef CHECK_COMPATIBILITY
	for(j = 0; j < 4; j++){
		switch(j){
		case 0:
		case 1:
			tmp_inclusion_array[j] = calloc(max_edge_curves_1[Y] + y_touch_count + 1, sizeof(side_inclusion));
			break;
		case 2:
		case 3:
			tmp_inclusion_array[j] = calloc(max_edge_curves_1[X] + x_touch_count + 1, sizeof(side_inclusion));
			break;
		}
		
		if(tmp_inclusion_array[j] == NULL)error_message(MEMORY_ERROR, "Memory allocation error");
	}
#endif

#ifdef CHECK_OVERLAP
#ifndef ALLOW_OVERLAP

/* check for overlap of all inclusions in reference tile */

	for(i = 1; i < inclusions; i++){

/* primary instance */

		inclusion = &inclusion_array[i];
		ellipse = &(inclusion -> ellipse);
		face_is = &(inclusion -> face_is);

		check_inclusion_periodic_overlap(ellipse, 0, 0, i);
		
/* secondary instances */

		if(inclusion -> edge_is[Y] == NULL)                   /* <==> inclusion -> face_is[X] == NULL */
			per_shift_x = 0.0;
		else{
			if(ellipse -> center.x < tile.center.x)             /* XN side */
				per_shift_x = tile.size.x;
			else                                                /* XP side */
				per_shift_x = -tile.size.x;

			shift_ellipse(*ellipse, per_shift_x, 0.0);
			check_inclusion_periodic_overlap(ellipse, 1, 0, i);
			shift_ellipse(*ellipse, -per_shift_x, 0.0);
		}
		if(inclusion -> edge_is[X] == NULL)                   /* <==> inclusion -> face_is[Y] == NULL */
			per_shift_y = 0.0;
		else{
			if(ellipse -> center.y < tile.center.y)             /* YN side */
				per_shift_y = tile.size.y;
			else                                                /* YP side */
				per_shift_y = -tile.size.y;

			shift_ellipse(*ellipse, 0.0, per_shift_y);
			check_inclusion_periodic_overlap(ellipse, 0, 1, i);
			shift_ellipse(*ellipse, 0.0, -per_shift_y);
		}
			
		if(per_shift_x != 0.0 && per_shift_y != 0.0){
			if(inclusion_array[i].corner == IN){
				shift_ellipse(*ellipse, per_shift_x, per_shift_y);
				check_inclusion_periodic_overlap(ellipse, 1, 1, i);
				shift_ellipse(*ellipse, -per_shift_x, -per_shift_y);
			}
		}
	}
#endif
#endif

	if(inside_hole == NO){
		for(j = 0; j < 4; j++)corner_hole[j] = NO;
	}
#ifdef CHECK_ISOLATED
	if(boundary_hole == NO){
		for(j = 0; j < 4; j++)corner_isolated[j] = NO;
	}
#endif

	if(strlen(text) + 2 > strlen(SEP_STRING)){
		write_record("#\n");
		write_record("# %s", text);
		write_record("#\n\n");
	}
	else{
		write_record(SEP_STRING "\n");
		write_record("# %s", text);
		write_record(SEP_STRING "\n\n");
	}

	sprintf(command, "t3d -d %g -t 0 -e %e", min_sz, min_epsilon);
	
	if(tiles == 1){     /* keep tile (PUC) in original position */
		if(codes == 0)
			write_record("# %s\n\n", command);
		else{

/* root octant is shifted so that tile is centered in root octant;
	 in order to reduce octree memory demands, tile is alligned with octree bottom */

			write_record("# %s -R "FMT1" -U "FMT3"\n\n", command, root_size, 
									 tile.center.x - root_size / 2.0, tile.center.y - root_size / 2.0, 0.0);
			write_record("# options -R and -U are relevant to prevent octree dispersion due to application of periodic BCs\n\n");
		}
		
		tile_shift_x = tile_shift_y = 0.0;
	}
	else{               /* shift x-y- corner of the 1st tile to origin of GCS */
		if(codes == 0)
			write_record("# %s\n\n", command);
		else{

/* root octant is shifted so that each tile is centered in the smallest octant containing the whole tile;
	 in order to reduce octree memory demands, tile is alligned with octree bottom */

			write_record("# %s -R "FMT1" -U "FMT3"\n\n", command, root_size,
									 -(offset - tile.size.x) / 2.0, -(offset - tile.size.y) / 2.0, 0.0);
			write_record("# options -R and -U are relevant to prevent octree dispersion due to application of generalized periodic BCs\n\n");
		}

		tile_shift_x = tile.size.x * 0.5 - tile.center.x;       
		tile_shift_y = tile.size.y * 0.5 - tile.center.y;

#if SETUP == GRID
		row = column = 0;
#endif
	}

	write_record("dimension 2\n");

/* CAUTION: tiles are processed according to their order (not id) in input file !!! */

	for(i = 0; i < tiles; i++){
		k = (i <= 3) ? i : 3;

		get_next_relevant_record("Missing inclusion definition");
		if(sscanf(line_buffer, "%ld", &(tile.id)) != 1)        /* caution: use tile.id */
			error_message(INPUT_ERROR, "Invalid record - definition of the %ld%s particular tile", i + 1, suffix[k]);
		
#ifdef ARBITRARY_TILE_ID
		if(tile.id < 0)
			error_message(INPUT_ERROR, "Invalid id of the %ld%s particular tile: TID = %ld (0 <= TID)", i + 1, suffix[k], tile.id);

		for(j = 0; j < i; j++){
			if(tile_array[j] == tile.id)
				error_message(INPUT_ERROR, "Multiple description of particular tile %ld encountered", tile.id);
		}
		tile_array[i] = tile.id;
		tile_id = i + 1;               /* IMPORTANT: I am using fictitious tile numbering to ensure tile_id < max_tiles */
#else
		if(tile.id <= 0 || tile.id >= tiles)
			error_message(INPUT_ERROR, "Invalid id of the %ld%s particular tile: TID = %ld (0 < TID <= %ld)", i + 1, suffix[k], tile.id, tiles);

		if(tile_array[tile.id - 1] != NO)
			error_message(INPUT_ERROR, "Multiple description of particular tile %ld encountered", tile.id);
		tile_array[tile.id - 1] = YES;
		tile_id = tile.id;
#endif

/* caution: when reporting error, use tile.id (not tile_id) which always contains real tile id no matter
	 whether aribitrary tile numbering is allowed or not */

		if(sscanf(line_buffer, "%ld %d %d %d %d %ld %ld", &(tile.id),        /* caution: use tile.id */
							&(tile.code[XN]), &(tile.code[XP]), &(tile.code[YN]), &(tile.code[YP]), 
							&(tile.inclusions), &(tile.points)) != 7)
			error_message(INPUT_ERROR, "Invalid record - definition of particular tile %ld", tile.id);        /* caution: use tile.id */
		
/* note: the codes can be generated automatically - NOT IMPLEMENTED;
	 this would require, however, that inclusions (including their periodic repetitions) intersecting or touching at least one side 
	 are remembered by particular tiles;
	 since I do not store particular tiles, there is an alternative, in which inclusions (including their periodic repetitions) intersecting or touching 
	 at least one side remember tiles they are used by;
	 to make the check more efficient each inclusion intersecting or touching at least one side will have 4 (for each possible periodic repetition) 
	 logical arrays of size equal to number of tiles;
	 then when processing a new particular tile, all previously processed particular tiles must be checked
	 whether the same set of inclusions intersecting a particular side is used;
	 since the code must be known in advance (to setup mirroring), it is necessary to process the check before the geometry of the tile is processed */

/* IMPORTANT: I do rely on fact that calloc was used to allocate code_array */

		for(j = 0; j < 4; j++){
			if(tile.code[j] < 0 || tile.code[j] > codes)
				error_message(INPUT_ERROR, "Invalid code for %s side of particular tile %ld: CODE = %d (0 < CODE < %ld)", 
											facespec[j], tile.id, tile.code[j], codes);        /* caution: use tile.id */
			
/* note: if both opposite tile sides have the same code, then the stored side corresponds to negative side;
	 this is consequence of the fact that codes are processed according to face ordering (in which negative side always preceeds opposite positive side);
	 I do rely on this fact when specifying mirroring between patches corresponding to periodically repeated inclusions */

			if(tile.code[j] == 0)continue;
			code = &code_array[tile.code[j] - 1];
			if(code -> tile_id != 0){
				if(code -> side != j && code -> side != opposite[j])
					error_message(INPUT_ERROR, "Code for %s side of particular tile %ld already used as code for %s side of particular tile %ld",
												facespec[j], tile.id, facespec[code -> side], code -> tile_id);        /* caution: use tile.id */
			}
			else{
				code -> id = tile.code[j];
				code -> tile_id = tile_id;          /* caution: store fictitious tile id if arbitrary tile numbering is allowed */
				code -> tile_id_ref = tile.id;      /* caution: store real tile id */
				code -> side = j;

#ifdef CHECK_COMPATIBILITY

/* alloc arrays increased by 1 to handle case when no inclusions interfere with particular side
	 and to prevent failure when max number is once exceeded */

				code -> inclusions = 0;
				switch(j){
				case 0:
				case 1:
					code -> inclusion_array = calloc(max_edge_curves_1[Y] + y_touch_count + 1, sizeof(side_inclusion));
					break;
				case 2:
				case 3:
					code -> inclusion_array = calloc(max_edge_curves_1[X] + x_touch_count + 1, sizeof(side_inclusion));
					break;
				}
					
				if(code -> inclusion_array == NULL)error_message(MEMORY_ERROR, "Memory allocation error");
#endif

			}
		}
					
		if(tile.inclusions < 0)
			error_message(INPUT_ERROR, "Invalid number of inclusions of particular tile %ld: NTI = %ld (0 <= NTI)", tile.id, tile.inclusions);        /* caution: use tile.id */
		if(tile.points < 0)
			error_message(INPUT_ERROR, "Invalid number of ctrl points of particular tile %ld: NCP = %ld (0 <= NCP)", tile.id, tile.points);        /* caution: use tile.id */
 
#ifdef DEBUG_PRINT
		fprintf(stderr, "\nProcessing tile %ld\n", tile.id);
		fprintf(stderr, "==================================================\n\n");
#endif

/* IMPORTANT: I cannot process geometry of inclusions here because the ordering of inclusions in particular tile is arbitrary
	 and I need to process inclusions according to their id;
	 therefore inclusions in particular tile are only marked as active and then are processed (see below) according to their id */

		for(j = 0; j < tile.inclusions; j++){
			kk = (j <= 3) ? j : 3;
			get_next_relevant_record("Missing inclusion id");
#ifdef SEPARATE_PER_REP
			if(sscanf(line_buffer, "%ld %d %d", &inclusion_id, &per_x, &per_y) != 3)
				error_message(INPUT_ERROR, "Invalid record - specification of the %ld%s inclusion of particular tile %ld", j + 1, suffix[kk], tile.id);        /* caution: use tile.id */
#else
			if(sscanf(line_buffer, "%ld %d", &inclusion_id, &per_xy) != 2)
				error_message(INPUT_ERROR, "Invalid record - specification of the %ld%s inclusion of particular tile %ld", j + 1, suffix[kk], tile.id);        /* caution: use tile.id */
#endif

			if(inclusion_id <= 0 || inclusion_id > inclusions)
				error_message(INPUT_ERROR, "Invalid id of the %ld%s inclusion of particular tile %ld: IID = %ld (0 < IID <= %ld)", 
											j + 1, suffix[kk], tile.id, inclusion_id, inclusions);        /* caution: use tile.id */

			inclusion = &inclusion_array[inclusion_id - 1];
			ellipse = &(inclusion -> ellipse);
			face_is = &(inclusion -> face_is);

#ifdef SEPARATE_PER_REP
			if(per_x != 0 && per_x != 1)
				error_message(INPUT_ERROR, "Invalid periodic repetition in x-direction of inclusion %ld of particular tile %ld: PRX = %d (PRX = {0;1})", 
											inclusion_id, tile.id, per_x);        /* caution: use tile.id */
			if(per_y != 0 && per_y != 1)
				error_message(INPUT_ERROR, "Invalid periodic repetition in y-direction of inclusion %ld of particular tile %ld: PRY = %d (PRY = {0;1})", 
											inclusion_id, tile.id, per_y);        /* caution: use tile.id */
#else
			if(per_xy < 0 || per_xy > 3)
				error_message(INPUT_ERROR, "Invalid periodic repetition iof inclusion %ld of particular tile %ld: PRXY = %d (PRXY = {0;1;2;3})", 
											inclusion_id, tile.id, per_xy);      /* caution: use tile.id */
			per_x = (per_xy & 1) >> 0;
			per_y = (per_xy & 2) >> 1;
#endif

#ifdef ALLOW_OUTSIDE
			if(inclusion -> out[X] == YES)per_x = 1 - per_x;
			if(inclusion -> out[Y] == YES)per_y = 1 - per_y;
#endif

			if(per_x != 0 || per_y != 0){

/* check that inclusion is intersecting or touching boundary with respect to periodic repetition;
	 note: although warning that inclusion is outside tile is relevant for all tiles, particular tile is specified
	 in order to eventually correct enumeration of inclusions for that particular tile */

				if(per_x == 1){
					if(inclusion -> edge_is[Y] == NULL){                   /* <==> inclusion -> face_is[X] == NULL */
						error_message(NOTIFICATION, "Inclusion %ld (%d %d) is outside tile %ld ==> ignored", inclusion_id, per_x, per_y, tile.id);
						continue;
					}
				}
				if(per_y == 1){
					if(inclusion -> edge_is[X] == NULL){                   /* <==> inclusion -> face_is[Y] == NULL */
						error_message(NOTIFICATION, "Inclusion %ld (%d %d) is outside tile %ld ==> ignored", inclusion_id, per_x, per_y, tile.id);
						continue;
					}
				}

/* ind_x|y defines portion of inclusion falling inside reference tile;
	 e.g. ind_y == 0 indicates that upper portion of inclusion above YN side falls inside reference tile
	 while ind_y == 1 indicates that lower portion of inclusion below YP side falls inside reference tile */
							
				if(ellipse -> center.x < tile.center.x)             /* XN side */
					ind_x = 0;
				else                                                /* XP side */
					ind_x = 1;
				if(ellipse -> center.y < tile.center.y)             /* YN side */
					ind_y = 0;
				else                                                /* YP side */
					ind_y = 1;

/* x|y_ind defines portion of inclusion (possibly periodically repeated) falling in particular tile;
	 e.g. y_ind == 0 indicates that upper portion of inclusion above YN side falls inside particular tile
	 while y_ind == 1 indicates that lower portion of inclusion below YP side falls inside particular tile */
	 
/* ind_y == y_ind <==> per_y = 0  .... primary instance of inclusion
	 ind_y != y_ind <==> per_y = 1  .... secondary instance of inclusion */
			
				if(per_x == 0)
					x_ind = ind_x;
				else{
					//x_ind = ind_x + sign[ind_x];
					x_ind = 1 - ind_x;
				}
				if(per_y == 0)
					y_ind = ind_y;
				else{
					//y_ind = ind_y + sign[ind_y];
					y_ind = 1 - ind_y;
				}
							
/* check that inclusion is intersecting or touching tile */

				x_count = y_count = 0;

				if(inclusion -> edge_is[X] != NULL)x_count = inclusion -> edge_is[X] -> count;
				if(inclusion -> edge_is[Y] != NULL)y_count = inclusion -> edge_is[Y] -> count;

/* CAUTION: inclusion touching tile from outside is not classified as outside !!! */

				if(check_inclusion_outside_tile() == YES){
					error_message(NOTIFICATION, "Inclusion %ld (%d %d) is outside tile %ld ==> ignored", inclusion_id, per_x, per_y, tile.id);
					continue;
				}
			}

			if(inclusion -> active[(per_x << 0) + (per_y << 1)] == YES){
				error_message(INPUT_ERROR, "Multiple use of inclusion %ld with periodic repetition %d %d in particular tile %ld", 
											inclusion_id, per_x, per_y, tile.id);        /* caution: use tile.id */
			}
			
			inclusion -> active[(per_x << 0) + (per_y << 1)] = YES;
		}
		
		tile_patches = 0;
		for(j = 0; j < 2; j++){
			x_edge_curves[j] = y_edge_curves[j] = 0;
			x_edge_vertices[j] = y_edge_vertices[j] = 0;
			x_edge_curves_1[j] = y_edge_curves_1[j] = 0;
			x_edge_curves_2[j] = y_edge_curves_2[j] = 1;
		}

		if(inside_hole == YES){
			for(j = 0; j < 4; j++)corner_hole[j] = NO;

/* check for tile corner inside hole;
	 note that this check is undesirable overhead because it is (indirectly) also performed when processing individual tile inclusions;
	 in the past I needed the information before tile corners were defined;
	 now I can suppress output at all corner vertices and after the whole tile is processed (and tile corners inside holes are checked)
	 I can process corner not inside the hole simply by fixing another vertex to the original corner vertex and enforcing output for the fixed vertex */

			count = 0;
			for(j = 0; j < inclusions; j++){
				inclusion = &inclusion_array[j];
				if(inclusion -> corner == IN){
					if(inclusion -> mat == 0 || inclusion -> ifmat != 0){
						for(k = 0; k < 4; k++){
							if(inclusion -> active[k] == YES){
								if(corner_hole[k] == NO){
									corner_hole[k] = YES;
									count++;
								}
							}
						}
						if(count == 4)break;
					}
				}
			}
		}

#ifdef CHECK_ISOLATED

/* CAUTION: there is a pathological case when tile corner is on boundary of hole inclusion and each of the tile edges is intersected twice;

          |  /h
	   t3   | /h   t2
          |/h
   -------o--------   hole inclusion is on side marked by h
         /|C
     t4 /h|    t1
       /h |

	 node at corner C needs to be generated even in tile t1 in order to allow for connection of t1 to t2 or t4;
	 an exception is when only a single tile is processed in which case the node is not needed;
	 note that this node in tile t1 will be isolated (not bounding any edge);
	 such node will cause problems in FE computation if t1 is the left-upper tile in an assembly, as there will be no stiffness assigned to the node;
	 isolated may be be therefore considered for removal from the assembled mesh !!! 
	 therefore a warning is printed;
	 note also that this problem is related not only to a single node by also to group of isolated elements;
	 for example:

          |     /h
	   t3   | t2 /h 
         C|   /h
   -------o--o-----   hole inclusion is on side marked by h
          | /h
     t4   |/h  t1
          oh
         /|
        /h|

	 I do not check for such a case;
	 note that it is very likely that periodic BCs are prescribed for the whole assembly;
	 then the isolated node or grou[ of elements is connected to the opposite side(s) of the tile and everything is ok */
							
		if(boundary_hole == YES){
			for(j = 0; j < 4; j++)corner_isolated[j] = NO;

			count = 0;
			for(j = 0; j < inclusions; j++){
				inclusion = &inclusion_array[j];
				if(inclusion -> corner == BND){
					if(inclusion -> mat == 0 || inclusion -> ifmat != 0){
						for(k = 0; k < 4; k++){
							if(inclusion -> active[k] == YES){
								if(corner_isolated[k] == NO){
									if(inclusion -> edge_is[X] -> count != 2)continue;
									if(inclusion -> edge_is[Y] -> count != 2)continue;

/* check that center of inclusion is inside of processed tile */

									if(inclusion -> ellipse.center.x  < tile.corner[0].x || inclusion -> ellipse.center.x  > tile.corner[1].x)continue;
									if(inclusion -> ellipse.center.y  < tile.corner[0].y || inclusion -> ellipse.center.y  > tile.corner[1].y)continue;

									corner_isolated[k] = YES;
									count++;
								}
							}
						}
						if(count == 4)break;
					}
				}
			}

			if(count != 0){
#if 0 /* not recommended */

/* warning: it is not desirable to suppress output of nodes at isolated vertices of a single tile
	 because it is very likely that there are periodic BCs !!! */

				if(tiles == 1){

/* in case when a single tile is processed vertices corresponding to isolated corners are not subjected to output */

					for(j = 0; j < 4; j++){
						if(corner_isolated[j] == YES)corner_hole[j] = YES;
					}
				}
				else{
#endif
					for(j = 0; j < 4; j++){
						if(corner_isolated[j] == YES)error_message(WARNING, "Isolated node at corner %d of tile %ld\n", j, tile.id);        /* caution: use tile.id */
					}
#if 0 /* not recommended */
				}
#endif
			}
		}
#endif

/* IMPORTANT: tile corner vertices must be ordered so that x|y-coordinates of latter from any two vertices 
	 on the same edge is larger or equal to coordinates of former from these two vertices;
	 thus vertices are ordered according to their periodic repetition;
	         id:  1  2  3  4
	 shifting_x:  0  1  0  1
	 shifting_y:  0  0  1  1 */

#ifdef ARBITRARY_TILE_ID
		write_record("\n# tile %ld (%ld)\n\n", tile_id, tile.id);
		write_record("tile %ld ref_id %ld", tile_id, tile.id);
#else
		write_record("\n# tile %ld\n\n", tile_id);
		write_record("tile %ld", tile_id);
#endif

/* caution: use shift from input position to output position */

		write_record(" out_shift "FMT3, -tile_shift_x, -tile_shift_y,  0.0);

#ifdef USE_CODE
		write_record(" code %d %d %d %d\n\n", tile.code[0], tile.code[1], tile.code[2], tile.code[3]); 
#else
		write_record("\n\n");
#endif 
		
		origin_id = vertex_id + 1;       /* remember id of the 1st tile vertex */

/* calculate coordinates of shifted particular tile */

		for(j = 0; j < 2; j++){
			par_corner[j].x = tile.corner[j].x + tile_shift_x;
			par_corner[j].y = tile.corner[j].y + tile_shift_y;
		}

		for(j = 0; j < 4; j++){
			if(corner_hole[j] == NO)
				write_record("vertex %ld xyz "FMT3" tile %ld\n", ++vertex_id, 
										 par_corner[shifting_x[j]].x, 
										 par_corner[shifting_y[j]].y, 
										 0.0, tile_id);
			else
				write_record("vertex %ld xyz "FMT3" output no tile %ld\n", ++vertex_id, 
										 par_corner[shifting_x[j]].x, 
										 par_corner[shifting_y[j]].y, 
										 0.0, tile_id);
		}

#ifdef CHECK_COMPATIBILITY
		tmp_inclusions[0] = tmp_inclusions[1] = tmp_inclusions[2] = tmp_inclusions[3] = 0; 
#endif

/* IMPORTANT: process (active) inclusions according to their id */

		for(j = 0; j < inclusions; j++){
			inclusion = &inclusion_array[j];
			ellipse = &(inclusion -> ellipse);
			face_is = &(inclusion -> face_is);

/* ind_x|y defines portion of inclusion falling inside reference tile;
	 e.g. ind_y == 0 indicates that upper portion of inclusion above YN side falls inside reference tile
	 while ind_y == 1 indicates that lower portion of inclusion below YP side falls inside reference tile */

			if(ellipse -> center.x < tile.center.x)           /* XN side */
				ind_x = 0;
			else                                              /* XP side */
				ind_x = 1;
			if(ellipse -> center.y < tile.center.y)           /* YN side */
				ind_y = 0;
			else                                              /* YP side */
				ind_y = 1;

/* x|y_ind defines portion of inclusion (possibly periodically repeated) falling in particular tile;
	 e.g. y_ind == 0 indicates that upper portion of inclusion above YN side falls inside particular tile
	 while y_ind == 1 indicates that lower portion of inclusion below YP side falls inside particular tile */
	 
/* ind_y == y_ind <==> per_y = 0  .... primary instance of inclusion
	 ind_y != y_ind <==> per_y = 1  .... secondary instance of inclusion */
			
			for(k = 0; k < 4; k++){
				if(inclusion -> active[k] == NO)continue;
				inclusion -> active[k] = NO;

				if((per_x = shifting_x[k]) == 0){
					per_shift_x = 0.0;
					x_ind = ind_x;
				}
				else{
					per_shift_x = sign[ind_x] * tile.size.x;
					//x_ind = ind_x + sign[ind_x];
					x_ind = 1 - ind_x;
				}
				
				if((per_y = shifting_y[k]) == 0){
					per_shift_y = 0.0;
					y_ind = ind_y;
				}
				else{
					per_shift_y = sign[ind_y] * tile.size.y;
					//y_ind = ind_y + sign[ind_y];
					y_ind = 1 - ind_y;
				}
				
				corner_id = origin_id + (y_ind << 1) + (x_ind << 0);
				cor_id = 0;

				write_record("\n# inclusion %ld (%d %d)\n\n", j + 1, per_x, per_y);

#ifdef CHECK_OVERLAP
#ifndef ALLOW_OVERLAP

/* apply the check also if ALLOW_OVERLAP is not defined (inclusions are not allowed to overlap in the reference tile) if inclusion shifting took place */

				if(shift_done == YES){
#endif

/* check for overlap of used inclusions in particular tile */

					shift_ellipse(*ellipse, per_shift_x, per_shift_y);
					check_active_inclusion_overlap(ellipse, per_x, per_y, j);
					shift_ellipse(*ellipse, -per_shift_x, -per_shift_y);
#ifndef ALLOW_OVERLAP
				}
#endif
#endif

				shift_vec.x = shift_x = tile_shift_x + per_shift_x;
				shift_vec.y = shift_y = tile_shift_y + per_shift_y;

				x_count = y_count = 0;

				if(inclusion -> edge_is[X] != NULL)x_count = inclusion -> edge_is[X] -> count;
				if(inclusion -> edge_is[Y] != NULL)y_count = inclusion -> edge_is[Y] -> count;

#ifdef DEBUG_PRINT
				fprintf(stderr, "\nProcessing inclusion %ld (%d %d)\n", j + 1, per_x, per_y);
				fprintf(stderr, "--------------------------------------------------\n\n");
				fprintf(stderr, "x|y_count: %d %d\n", x_count, y_count);
				fprintf(stderr, "corner: %s\n\n", (inclusion -> corner < 0) ? "in" : (inclusion -> corner > 0) ? "out" : "bnd");
#endif

/* resolve side geometry;
	 note: x|y_intersection, x|y_touching, and x|y_edge_is is set on process_geometry[_without|with_interface] */

				x_curve_id = y_curve_id = 0;

/* note that tile_id of inclusion patch (intersection of inclusion with tile side) is printed no matter whether hole or not;
	 similarly, tile_id of inclusion curve (intersection of inclusion with tile edge) is printed no matter whether outputed or not */

				if(inclusion -> mat == 0)
					sprintf(output_spec, "output no");
				else
					*output_spec = '\0';

				*mirror_spec_x = '\0';
				*mirror_spec_y = '\0';

#ifndef USE_SIDE
				*side_spec_x = '\0';
				*side_spec_y = '\0';
#endif

#ifdef USE_SIDE
				sprintf(side_spec_x, "side %d", 2 * Y + y_ind + 1);
				sprintf(side_spec_y, "side %d", 2 * X + x_ind + 1);
#else
				*side_spec_x = '\0';
				*side_spec_y = '\0';
#endif

				interface = (inclusion -> ifmat == 0) ? NO : YES;
				process_geometry();

#ifdef CHECK_COMPATIBILITY
				sd_inclusion = NULL;
				if(x_count != 0){
					if(tile.code[y_ind + 2] != 0){
						code = &code_array[tile.code[y_ind + 2] - 1];

/* CAUTION: if inclusions with INDEF edge_is are included, inclusions touching a tile from outside at the tile corner must be included
	 in enumeration of tile inclusions !!! */

						if(x_edge_is -> side == INDEF || x_edge_is -> side == BOTH)
							sd_inclusion = &(tmp_inclusion_array[y_ind + 2][tmp_inclusions[y_ind + 2]++]);
						else{
							if((x_edge_is -> side & ref_position[x_ind]) != 0){
								sd_inclusion = &(tmp_inclusion_array[y_ind + 2][tmp_inclusions[y_ind + 2]++]);
							}
						}
					}
				}						
				if(sd_inclusion != NULL){
					sd_inclusion -> inclusion = inclusion;
					sd_inclusion -> per_x = per_x;
					sd_inclusion -> per_y = per_y;

					if(tmp_inclusions[y_ind + 2] > max_edge_curves_1[X] + x_touch_count)error_message(GENERAL_ERROR, "Unexpected situation");
				}

				sd_inclusion = NULL;
				if(y_count != 0){
					if(tile.code[x_ind + 0] != 0){
						code = &code_array[tile.code[x_ind + 0] - 1];

/* CAUTION: if inclusions with INDEF edge_is are included, inclusions touching a tile from outside at the tile corner must be included
	 in enumeration of tile inclusions !!! */

						if(y_edge_is -> side == INDEF || y_edge_is -> side == BOTH)
							sd_inclusion = &(tmp_inclusion_array[x_ind + 0][tmp_inclusions[x_ind + 0]++]);
						else{
							if((y_edge_is -> side & ref_position[y_ind]) != 0){
								sd_inclusion = &(tmp_inclusion_array[x_ind + 0][tmp_inclusions[x_ind + 0]++]);
							}
						}
					}
				}
				if(sd_inclusion != NULL){
					sd_inclusion -> inclusion = inclusion;
					sd_inclusion -> per_x = per_x;
					sd_inclusion -> per_y = per_y;

					if(tmp_inclusions[x_ind + 0] > max_edge_curves_1[Y] + y_touch_count)error_message(GENERAL_ERROR, "Unexpected situation");
				}
#endif

#if 0   // shift including ctrl points - dodelat
#ifdef ELIXIR
				if(draw == YES){
					draw_inclusion(inclusion, INCLUSION_LAYER, red);
					
#ifdef DEBUG_DRAW
					ESIEventLoop(YES, "Inclusion\n\n");
#endif

				}
#endif
#endif

/* zero id of vertices at intersection points */

				if(x_count != 0){
					for(m = 0; m < x_count; m++)x_edge_is -> point[m].id = 0;
				}
				if(y_count != 0){
					for(m = 0; m < y_count; m++)y_edge_is -> point[m].id = 0;
				}
			}
		}

#ifdef CHECK_COMPATIBILITY

/* check for "compatibility" of sides with nonzero code with prototype side;
	 alarm is issued instead of error to see the geometry in t3d;
	 note that t3d is likely to not process the whole input because of the incompatibility */

		for(j = 0; j < 4; j++){
			if(tile.code[j] == 0)continue;

			alarm = NO;
			dir = j / 2;           /* X or Y */
			m = j - dir * 2;       /* 0 or 1 */

			code = &code_array[tile.code[j] - 1];
			if(code -> tile_id_ref == tile.id){         /* tile with the first appearance of the code */
				if(j == 0 || j == 2){                     /* processing negative side */
					swap(code -> inclusion_array, tmp_inclusion_array[j], dummy_inclusion_array);
					swap(code -> inclusions, tmp_inclusions[j], dummy_inclusions);
				}
				else{                                     /* processing positive side */
					if(tile.code[j] != tile.code[j - 1]){   /* opposite codes are different */
						swap(code -> inclusion_array, tmp_inclusion_array[j], dummy_inclusion_array);
						swap(code -> inclusions, tmp_inclusions[j], dummy_inclusions);
					}
					else{

/* perform check of opposite sides of the same tile */

/* check whether primary and secondary set are the same except for the periodic repetion in direction of checked side */

						if(code -> inclusions != tmp_inclusions[j])
							alarm = YES;
						else{
							for(k = 0; k < code -> inclusions; k++){
							
/* IMPORTANT: I do rely on the same ordering of inclusions and specific ordering of periodic repetitions !!! */

								if(code -> inclusion_array[k].inclusion != tmp_inclusion_array[j][k].inclusion)
									alarm = YES;
								else{
									switch(dir){
									case X:
										if(code -> inclusion_array[k].per_x == tmp_inclusion_array[j][k].per_x)alarm = YES;
										if(code -> inclusion_array[k].per_y != tmp_inclusion_array[j][k].per_y)alarm = YES;
										break;
									case Y:
										if(code -> inclusion_array[k].per_y == tmp_inclusion_array[j][k].per_y)alarm = YES;
										if(code -> inclusion_array[k].per_x != tmp_inclusion_array[j][k].per_x)alarm = YES;
										break;
									default:
										break;
									}
								}
								if(alarm == YES)break;

							}
						}

#if 0

/* alternatively, since I perform check within single tile, I could check whether all inclusions in primary set 
	 appear with both periodic repetitions in direction of checked side;
	 however I cannot use activity of inclusion as it was already canceled !!! */

						last_inclusion = NULL;
						for(k = 0; k < code -> inclusions; k++){
							sd_inclusion = &(code -> inclusion_array[k]);
							inclusion = sd_inclusion -> inclusion;
							if(inclusion != last_inclusion){      /* check when the first instance of particular inclusion is encountered */
								switch(dir){
								case X:
									if(inclusion -> active[0] != inclusion -> active[1] || inclusion -> active[2] != inclusion -> active[3])alarm = YES;
									break;
								case Y:
									if(inclusion -> active[0] != inclusion -> active[2] || inclusion -> active[1] != inclusion -> active[3])alarm = YES;
									break;
								default:
									break;
								}
								if(alarm == YES)break;
								last_inclusion = inclusion;
							}
						}
#endif

					}
				}
			}
			else{

/* check whether primary and secondary set are the same except for eventual periodic repetion in direction of checked side */

				if(code -> inclusions != tmp_inclusions[j])
					alarm = YES;
				else{
					for(k = 0; k < code -> inclusions; k++){
							
/* IMPORTANT: I do rely on the same ordering of inclusions and their periodic repetitions !!! */

						if(code -> inclusion_array[k].inclusion != tmp_inclusion_array[j][k].inclusion)
							alarm = YES;
						else{
							switch(dir){
							case X:
								if(code -> side == j){
									if(code -> inclusion_array[k].per_x != tmp_inclusion_array[j][k].per_x)alarm = YES;
								}
								else{
									if(code -> inclusion_array[k].per_x == tmp_inclusion_array[j][k].per_x)alarm = YES;
								}
								if(code -> inclusion_array[k].per_y != tmp_inclusion_array[j][k].per_y)alarm = YES;
								break;
							case Y:
								if(code -> side == j){
									if(code -> inclusion_array[k].per_y != tmp_inclusion_array[j][k].per_y)alarm = YES;
								}
								else{
									if(code -> inclusion_array[k].per_y == tmp_inclusion_array[j][k].per_y)alarm = YES;
								}
								if(code -> inclusion_array[k].per_x != tmp_inclusion_array[j][k].per_x)alarm = YES;
								break;
							default:
								break;
							}
						}
						if(alarm == YES)break;
					}
				}
			}

			if(alarm == YES){
				compatibility_alarm = YES;

#ifdef REPORT_MISSING_REDUNDANT
				fprintf(stderr, "\n");
#endif
				error_message(ALARM, "%s-side of tile %ld is not compatible with %s-side of tile %ld - code %ld",
											Facespec[2 * dir + m], tile.id, facespec[code -> side], code -> tile_id_ref, code -> id);
				
#ifdef REPORT_MISSING_REDUNDANT

/* make active all inclusion on just processed side corresponding to inclusions interfering with reference side */

				if(code -> side == j){
					for(k = 0; k < code -> inclusions; k++){
						sd_inclusion = &(code -> inclusion_array[k]);
						inclusion = sd_inclusion -> inclusion;
						inclusion -> active[((sd_inclusion -> per_x) << 0) + ((sd_inclusion -> per_y) << 1)] = YES;
					}
				}
				else{
					for(k = 0; k < code -> inclusions; k++){
						sd_inclusion = &(code -> inclusion_array[k]);
						inclusion = sd_inclusion -> inclusion;
						if(j == 0 || j == 1)
							inclusion -> active[((1 - sd_inclusion -> per_x) << 0) + ((sd_inclusion -> per_y) << 1)] = YES;
						else
							inclusion -> active[((sd_inclusion -> per_x) << 0) + ((1 - sd_inclusion -> per_y) << 1)] = YES;
					}
				}

/* check inclusions interfering with just processed side;
	 if not active, they are redundant */

				for(k = 0; k < tmp_inclusions[j]; k++){
					sd_inclusion = &(tmp_inclusion_array[j][k]);
					inclusion = sd_inclusion -> inclusion;
					if(inclusion -> active[((sd_inclusion -> per_x) << 0) + ((sd_inclusion -> per_y) << 1)] == NO){
#if 0
						error_message(WARNING, "Inclusion %ld (%d %d) redundant in tile %ld", inclusion -> id, sd_inclusion -> per_x, sd_inclusion -> per_y, tile.id);
#else /* account for potential contradiction */
						if(inclusion -> corner == IN || inclusion -> corner == BND) 
							error_message(WARNING, "Inclusion %ld (%d %d) redundant (?) in tile %ld", inclusion -> id, sd_inclusion -> per_x, sd_inclusion -> per_y, tile.id);
						else
							error_message(WARNING, "Inclusion %ld (%d %d) redundant in tile %ld", inclusion -> id, sd_inclusion -> per_x, sd_inclusion -> per_y, tile.id);
#endif
					}
				}

/* deactive all inclusions interfering with reference side */

				for(k = 0; k < code -> inclusions; k++){
					sd_inclusion = &(code -> inclusion_array[k]);
					inclusion = sd_inclusion -> inclusion;
					inclusion -> active[0] = inclusion -> active[1] = inclusion -> active[2] = inclusion -> active[3] = NO;
				}

/* make active all inclusions interfering with just processed side */

				for(k = 0; k < tmp_inclusions[j]; k++){
					sd_inclusion = &(tmp_inclusion_array[j][k]);
					inclusion = sd_inclusion -> inclusion;
					inclusion -> active[((sd_inclusion -> per_x) << 0) + ((sd_inclusion -> per_y) << 1)] = YES;
				}

/* check inclusions interfering with reference side;
	 if not active, they are missing;
	 account for opposite side with respect to reference side by applying periodic repetition indirection of checked side */

				if(code -> side == j){
					for(k = 0; k < code -> inclusions; k++){
						sd_inclusion = &(code -> inclusion_array[k]);
						inclusion = sd_inclusion -> inclusion;
						if(inclusion -> active[((sd_inclusion -> per_x) << 0) + ((sd_inclusion -> per_y) << 1)] == NO){
#if 0
							error_message(WARNING, "Inclusion %ld (%d %d) missing in tile %ld", inclusion -> id, sd_inclusion -> per_x, sd_inclusion -> per_y, tile.id);
#else /* account for potential contradiction */
							if(inclusion -> corner == IN || inclusion -> corner == BND)
								error_message(WARNING, "Inclusion %ld (%d %d) missing (?) in tile %ld", inclusion -> id, sd_inclusion -> per_x, sd_inclusion -> per_y, tile.id);
							else
								error_message(WARNING, "Inclusion %ld (%d %d) missing in tile %ld", inclusion -> id, sd_inclusion -> per_x, sd_inclusion -> per_y, tile.id);
#endif
						}
					}
				}
				else{
					for(k = 0; k < code -> inclusions; k++){
						sd_inclusion = &(code -> inclusion_array[k]);
						inclusion = sd_inclusion -> inclusion;
						if(j == 0 || j == 1){
							if(inclusion -> active[((1 - sd_inclusion -> per_x) << 0) + ((sd_inclusion -> per_y) << 1)] == NO){
#if 0
								error_message(WARNING, "Inclusion %ld (%d %d) missing in tile %ld", inclusion -> id, 1 - sd_inclusion -> per_x, sd_inclusion -> per_y, tile.id);
#else /* account for potential contradiction */
								if(inclusion -> corner == IN || inclusion -> corner == BND)
									error_message(WARNING, "Inclusion %ld (%d %d) missing (?) in tile %ld", inclusion -> id, 1 - sd_inclusion -> per_x, sd_inclusion -> per_y, tile.id);
								else
									error_message(WARNING, "Inclusion %ld (%d %d) missing in tile %ld", inclusion -> id, 1 - sd_inclusion -> per_x, sd_inclusion -> per_y, tile.id);
#endif
							}
						}
						else{
							if(inclusion -> active[((sd_inclusion -> per_x) << 0) + ((1 - sd_inclusion -> per_y) << 1)] == NO){
#if 0
								error_message(WARNING, "Inclusion %ld (%d %d) missing in tile %ld", inclusion -> id, sd_inclusion -> per_x, 1 - sd_inclusion -> per_y, tile.id);
#else /* account for potential contradiction */
								if(inclusion -> corner == IN || inclusion -> corner == BND)
									error_message(WARNING, "Inclusion %ld (%d %d) missing (?) in tile %ld", inclusion -> id, sd_inclusion -> per_x, 1 - sd_inclusion -> per_y, tile.id);
								else
									error_message(WARNING, "Inclusion %ld (%d %d) missing in tile %ld", inclusion -> id, sd_inclusion -> per_x, 1 - sd_inclusion -> per_y, tile.id);
#endif
							}
						}
					}
				}

/* deactive all inclusions interfering with just processed side */

				for(k = 0; k < tmp_inclusions[j]; k++){
					sd_inclusion = &(tmp_inclusion_array[j][k]);
					inclusion = sd_inclusion -> inclusion;
					inclusion -> active[0] = inclusion -> active[1] = inclusion -> active[2] = inclusion -> active[3] = NO;
				}
#endif

			}
		}
#endif

#ifndef CHECK_COMPATIBILITY  /* prevent duplicated alarms */

/* check for "compatibility" of sides with nonzero code with prototype side;
	 alarm is issued instead of error to see the geometry in t3d;
	 note that t3d is likely to not process the whole input because of the incompatibility */

		edge_curves[0] = y_edge_curves_1[0] + y_edge_curves_2[0];   /* x- side */
		edge_curves[1] = y_edge_curves_1[1] + y_edge_curves_2[1];   /* x+ side */
		
		for(m = 0; m < 2; m++){
			if(tile.code[2 * X + m] != 0){
				code = &code_array[tile.code[2 * X + m] - 1];
				if(code -> curves == 0)                /* not yet processed ==> store prototype counts */
					code -> curves = edge_curves[m];
				else{
					if(code -> curves != edge_curves[m]){
						error_message(ALARM, "%s-side of tile %ld is not compatible with %s-side of tile %ld - code %ld",
													Facespec[2 * X + m], tile.id, facespec[code -> side], code -> tile_id_ref, code -> id);
						compatibility_alarm = YES;
					}
				}
			}
		}					
		
		edge_curves[0] = x_edge_curves_1[0] + x_edge_curves_2[0];   /* y- side */
		edge_curves[1] = x_edge_curves_1[1] + x_edge_curves_2[1];   /* y+ side */
		
		for(m = 0; m < 2; m++){
			if(tile.code[2 * Y + m] != 0){
				code = &code_array[tile.code[2 * Y + m] - 1];
				if(code -> curves == 0)                /* not yet processed ==> store prototype counts */
					code -> curves = edge_curves[m];
				else{
					if(code -> curves != edge_curves[m]){
						error_message(ALARM, "%s-side of tile %ld is not compatible with %s-side of tile %ld - code %ld",
													Facespec[2 * Y + m], tile.id, facespec[code -> side], code -> tile_id_ref, code -> id);
						compatibility_alarm = YES;
					}
				}
			}
		}					
#endif		

/* process matrix */

		write_record("\n# matrix\n\n");

/* generate remaining segments on tile edges */

/* IMPORTANT: since the side codes are processed (when searching for the first occurrence of particular code) according to edge ordering, 
	 tile edge curves (at least the opposite ones) must be generated in the same order as is the edge numbering (XN, XP, YN, YP) !!! */

		for(m = 0; m < 2; m++){
			*mirror_spec_x = '\0';
			mirror = NO;
			if(tile.code[2 * Y + m] != 0){
				code = &code_array[tile.code[2 * Y + m] - 1];
				if(code -> tile_id != tile_id || (m != 0 && code -> side == YN))mirror = YES;
			}

#ifdef USE_SIDE
			sprintf(side_spec_x, "side %d", (int)(2 * Y + m + 1));
#endif

			if(x_edge_curves_2[m] > max_edge_curves_2[X])error_message(GENERAL_ERROR, "Unexpected situation");
			sort_id_coord(x_edge_vertices[m], x_edge_vertex_id_array[m], x_edge_vertex_data_array[m]);

			ver_id1 = origin_id + (1 << Y) * m;
			ver_id2 = ver_id1 + (1 << X);

			count = 0;
			status = END;
			for(n = 0; n < x_edge_vertices[m]; n++){
				ver_id = x_edge_vertex_id_array[m][n];
				if((status = x_edge_vertex_data_array[m][n] -> status) != END){
#ifdef DEBUG
					if(status == START){
						if(n != x_edge_vertices[m] - 1){
							if(x_edge_vertex_data_array[m][n + 1] -> status != END)error_message(GENERAL_ERROR, "Unexpected situation");
						}
					}
#endif
					if(mirror == YES){
						cur_id = ((x_edge_curves[m] + 1) * max_tiles + code -> tile_id) * 10 + code -> side + 1;    /* keep compatible with get_y_side_curve_id macro */
#ifdef USE_SIMPLE
						sprintf(mirror_spec_x, "mirror %ld simple", cur_id);
#else
						sprintf(mirror_spec_x, "mirror %ld", cur_id);
#endif
					}
					cur_id = get_y_side_curve_id(m);
					write_record("curve %ld vertex %ld %ld tile %ld %s %s\n", cur_id, ver_id1, ver_id, tile_id, mirror_spec_x, side_spec_x);
					x_edge_curve_2_array[m][count++] = cur_id;
						
					if(count > x_edge_curves_2[m])error_message(GENERAL_ERROR, "Unexpected situation");
				}
#ifdef DEBUG
				else{
					if(n != 0){
						if(x_edge_vertex_data_array[m][n - 1] -> status != START)error_message(GENERAL_ERROR, "Unexpected situation");
					}
				}
#endif
				ver_id1 = ver_id;
			}
			if(status != START){
				if(mirror == YES){
					cur_id = ((x_edge_curves[m] + 1) * max_tiles + code -> tile_id) * 10 + code -> side + 1;    /* keep compatible with get_y_side_curve_id macro */
#ifdef USE_SIMPLE
					sprintf(mirror_spec_x, "mirror %ld simple", cur_id);
#else
					sprintf(mirror_spec_x, "mirror %ld", cur_id);
#endif
				}
				cur_id = get_y_side_curve_id(m);
				write_record("curve %ld vertex %ld %ld tile %ld %s %s\n", cur_id, ver_id1, ver_id2, tile_id, mirror_spec_x, side_spec_x);
				x_edge_curve_2_array[m][count++] = cur_id;
			}
#ifdef DEBUG
			if(count != x_edge_curves_2[m])error_message(GENERAL_ERROR, "Unexpected situation");
#endif

			*mirror_spec_y = '\0';
			mirror = NO;
			if(tile.code[2 * X + m] != 0){
				code = &code_array[tile.code[2 * X + m] - 1];
				if(code -> tile_id != tile_id || (m != 0 && code -> side == XN))mirror = YES;
			}

#ifdef USE_SIDE
			sprintf(side_spec_y, "side %d", (int)(2 * X + m + 1));
#endif

			if(y_edge_curves_2[m] > max_edge_curves_2[Y])error_message(GENERAL_ERROR, "Unexpected situation");
			sort_id_coord(y_edge_vertices[m], y_edge_vertex_id_array[m], y_edge_vertex_data_array[m]);

			ver_id1 = origin_id + (1 << X) * m;
			ver_id2 = ver_id1 + (1 << Y);

			count = 0;
			status = END;
			for(n = 0; n < y_edge_vertices[m]; n++){
				ver_id = y_edge_vertex_id_array[m][n];
				if((status = y_edge_vertex_data_array[m][n] -> status) != END){
#ifdef DEBUG
					if(status == START){
						if(n != y_edge_vertices[m] - 1){
							if(y_edge_vertex_data_array[m][n + 1] -> status != END)error_message(GENERAL_ERROR, "Unexpected situation");
						}
					}
#endif
					if(mirror == YES){
						cur_id = ((y_edge_curves[m] + 1) * max_tiles + code -> tile_id) * 10 + code -> side + 1;    /* keep compatible with get_x_side_curve_id macro */
#ifdef USE_SIMPLE
						sprintf(mirror_spec_y, "mirror %ld simple", cur_id);
#else
						sprintf(mirror_spec_y, "mirror %ld", cur_id);
#endif
					}
					cur_id = get_x_side_curve_id(m);
					write_record("curve %ld vertex %ld %ld tile %ld %s %s\n", cur_id, ver_id1, ver_id, tile_id, mirror_spec_y, side_spec_y);
					y_edge_curve_2_array[m][count++] = cur_id;
						
					if(count > y_edge_curves_2[m])error_message(GENERAL_ERROR, "Unexpected situation");
				}
#ifdef DEBUG
				else{
					if(n != 0){
						if(y_edge_vertex_data_array[m][n - 1] -> status != START)error_message(GENERAL_ERROR, "Unexpected situation");
					}
				}
#endif
				ver_id1 = ver_id;
			}
			if(status != START){
				if(mirror == YES){
					cur_id = ((y_edge_curves[m] + 1) * max_tiles + code -> tile_id) * 10 + code -> side + 1;    /* keep compatible with get_x_side_curve_id macro */
#ifdef USE_SIMPLE
					sprintf(mirror_spec_y, "mirror %ld simple", cur_id);
#else
					sprintf(mirror_spec_y, "mirror %ld", cur_id);
#endif
				}
				cur_id = get_x_side_curve_id(m);
				write_record("curve %ld vertex %ld %ld tile %ld %s %s\n", cur_id, ver_id1, ver_id2, tile_id, mirror_spec_y, side_spec_y);
				y_edge_curve_2_array[m][count++] = cur_id;
			}
#ifdef DEBUG
			if(count != y_edge_curves_2[m])error_message(GENERAL_ERROR, "Unexpected situation");
#endif

		}

		write_record("\n");
		
		edge_curve_sign[0] = edge_curve_sign[4] = 1;
		edge_curve_sign[1] = edge_curve_sign[5] = 1;
		edge_curve_sign[2] = edge_curve_sign[6] = -1;
		edge_curve_sign[3] = edge_curve_sign[7] = -1;

/* caution: x|y_edge_curve_1|2_array must be stored in edge_curve_array to not violate edge_curve_sign (defined above);
	 note that x|y_edge_curve_2_array (corresponding to real boundary of tile side) is stored before x|y_edge_curve_1_array (corresponding to 
	 fictitious boundary of tile side) */

		write_record("patch %ld normal 0.0 0.0 1.0 boundary curve", ++patch_id);
		
		edge_curve_array[0] = x_edge_curve_2_array[0];
		edge_curve_array[1] = y_edge_curve_2_array[1];
		edge_curve_array[2] = x_edge_curve_2_array[1]; 
		edge_curve_array[3] = y_edge_curve_2_array[0];
		edge_curve_array[4] = x_edge_curve_1_array[0];
		edge_curve_array[5] = y_edge_curve_1_array[1];
		edge_curve_array[6] = x_edge_curve_1_array[1];
		edge_curve_array[7] = y_edge_curve_1_array[0];
		
		edge_curves[0] = x_edge_curves_2[0];
		edge_curves[1] = y_edge_curves_2[1];
		edge_curves[2] = x_edge_curves_2[1]; 
		edge_curves[3] = y_edge_curves_2[0];
		edge_curves[4] = x_edge_curves_1[0];
		edge_curves[5] = y_edge_curves_1[1];
		edge_curves[6] = x_edge_curves_1[1];
		edge_curves[7] = y_edge_curves_1[0];
		
		output_patch_boundary(edge_curves, edge_curve_array, edge_curve_sign, tile_patches, tile_patch_array);
		
		write_record(" size %g tile %ld property %d\n", tile.bmsz, tile_id, tile.mat);
		
		if(tile.points != 0)write_record("\n");
		
		for(j = 0; j < tile.points; j++){
			kk = (j <= 3) ? j : 3;
			get_next_relevant_record("Missing control point definition");
			
			if(sscanf(line_buffer, "%lf %lf %lf", &(point.x), &(point.y), &msz) != 3)
				error_message(INPUT_ERROR, "Invalid record - definition of the %ld%s ctrl point of particular tile %ld", j + 1, suffix[kk], tile.id);        /* caution: use tile.id */

			if(point.x < tile.corner[0].x - epsilon || point.x > tile.corner[1].x + epsilon ||
				 point.y < tile.corner[0].y - epsilon || point.y > tile.corner[1].y + epsilon)
				error_message(INPUT_ERROR, "Invalid coordinates of the %ld%s control point of particular tile %ld",  
											j + 1, suffix[kk], tile.id);        /* caution: use tile.id */
	
			if(msz < 0.0)error_message(INPUT_ERROR, "Invalid mesh size of the %ld%s ctrl point of particular tile %ld: MSZ = %e (0 < MSZ)", 
																 j + 1, suffix[k], tile.id, msz);        /* caution: use tile.id */
			
			write_record("vertex %ld virtual xyz "FMT3" size %g\n", ++vertex_id, point.x, point.y, 0.0, msz);
		}

		if(i != tiles - 1){
#if SETUP == GRID
			if(++row == grid_size){
				tile_shift_x = min_sz * 0.5 - tile.center.x;
				row = 0;
				if(++column == grid_size)
					error_message(GENERAL_ERROR, "Unexpected situation");
				else
					tile_shift_y += offset;
			}
			else
				tile_shift_x += offset;
#endif
#if SETUP == DIAGONAL
			tile_shift_x += offset;
			tile_shift_y += offset;
#endif
		}
	}

/* CAUTION: patches are not numbered continuously !!!
	 using the largest patch id is contraproductive as too large array would have to be allocated by t3d (see unique patch numbering)
	 resulting possibly in memory allocation failure */

	write_record("\n");
	write_record("# Optionally, add to command line -n %ld %ld %ld %ld %ld %ld %ld\n", vertex_id, 0, 0, patch_id, 0, 0, interface_id);
	write_record("# or uncomment the following records and move them to the start of this file\n");
	write_record("#\n");
	if(vertex_id != 0)write_record("#vertices %ld\n", vertex_id);
	if(patch_id != 0)write_record("#patches %ld\n", patch_id);
	if(interface_id != 0)write_record("#interfaces %ld\n", interface_id);

	if(tiles == 1)
		fprintf(stdout, "\n1 tile successfully processed\n\n");
	else
		fprintf(stdout, "\n%ld tiles successfully processed\n\n", tiles);

	if(compatibility_alarm == YES){
		error_message(WARNING, "Generated input file may not be processed by T3d due to compatibility alarm(s).\n         If processed, matching sides are likely to be not compatible\n         and undesirable refinement of the mesh may occur.\n");
	}
		
#ifdef ELIXIR
	if(draw == YES)ESIEventLoop(YES, "");
#endif

	fclose(t3d_in_file);
	fclose(geom_file);

	return(0);
}



/* YES is returned if inclusion is intersecting or touching the tile, otherwise NO is returned;
	 note that it is safe to return YES even if the no intersection or touching occurs */

static logic
get_inclusion_tile_intersection(void)
{
	point_rec pnt_1, pnt_2;
	vector_rec vec, vector, vec_tmp;
	vector_rec total_shift_tmp, shift_vec_tmp;
	double size, val, eps;
	double max_shift, max_shift_tmp;
	double a_diff_x, a_diff_y, a_x_diff, a_y_diff;
	double coord;
	int i, ind, subdivision;
	logic shifting, corner_x_between, corner_y_between;
	corner_status corner_tmp;
	
	x_intersection = y_intersection = NO;
	x_touching = y_touching = NO;

	x_count = y_count = 0;

	total_shift.x = total_shift.y = 0.0;

/* compute extreme points in GCS;

	 they are computed from the fact that gradient at extreme point is parallel to direction in which extreme is investigated;
	 the equality between the gradient and the direction is expressed in LCS;
	 for example when searching for extreme in y-direction:
	 {2x/max^2;2y/min^2}=dir_2*value;
	 this is equivalent to
	 {x/max^2;y/min^2}=dir_2*val;
	 this implies
	 x=dir_2.x*val*max^2;
	 y=dir_2.y*val*min^2;
	 assuming
	 vec_y.x=dir_2.x*max;
	 vec_y.y=dir_2.y*min;
	 one can write
	 x=vec_y.x*max*val
	 y=vec_y.y*min*val
	 val is computed from equation of the ellipse
	 x^2/max^2+y^2/min^2=1;
	 (vec_y.x)^2*val^2+(vec_y.y)^2*val^2=1;
	 val=sqrt(1.0/((vec_y.x)^2+(vec_y.y)^2) */

	copy_vec(trans_g2l.x, ellipse -> major);
	copy_vec(trans_g2l.y, ellipse -> minor);
	
	get_inverted_rotation_transformation(&trans_g2l, &trans_l2g);

	transform_vector_from_global_to_local(&dir_x, &dir_1, &trans_g2l);
	transform_vector_from_global_to_local(&dir_y, &dir_2, &trans_g2l);
	
	vec.x = ellipse -> max;
	vec.y = ellipse -> min;
	
	comp_product(vector, vec, vec);
	
	comp_product(vec_x, dir_1, vec);
	comp_product(vec_y, dir_2, vec);
	
	val = 1.0 / sqrt(size_vec(vec_x));
	copy_mul_vec(vec_tmp, vector, val);
	comp_product(pnt_1, dir_1, vec_tmp);
	
	transform_point_from_local_to_global(&pnt_1, &ext_point_x[1], &(ellipse -> center), &trans_l2g);
	
#if 1 /* more efficient */
	sub_vec(vec_tmp, ellipse -> center, ext_point_x[1]);
	add_vec(ext_point_x[0], ellipse -> center, vec_tmp);
#else
	copy_mul_vec(pnt_2, pnt_1, -1.0);
	transform_point_from_local_to_global(&pnt_2, &ext_point_x[0], &(ellipse -> center), &trans_l2g);
#endif
	
	val = 1.0 / sqrt(size_vec(vec_y));
	copy_mul_vec(vec_tmp, vector, val);
	comp_product(pnt_1, dir_2, vec_tmp);
	
	transform_point_from_local_to_global(&pnt_1, &ext_point_y[1], &(ellipse -> center), &trans_l2g);
	
#if 1 /* more efficient */
	sub_vec(vec_tmp, ellipse -> center, ext_point_y[1]);
	add_vec(ext_point_y[0], ellipse -> center, vec_tmp);
#else
	copy_mul_vec(pnt_2, pnt_1, -1.0);
	transform_point_from_local_to_global(&pnt_2, &ext_point_y[0], &(ellipse -> center), &trans_l2g);
#endif
	
#ifdef DRAW_EXTREME_POINT_ELLIPSE
	for(i = 0; i < 2; i++){
		draw_point(&ext_point_x[i], 0, cyan);
		draw_point(&ext_point_y[i], 0, green);
	}
#endif

/* note that ind_x|y should not change due to eventual ellipse shifting */

	if(ellipse -> center.x < tile.center.x)             /* XN side */
		ind_x = 0;
	else                                                /* XP side */
		ind_x = 1;
	if(ellipse -> center.y < tile.center.y)             /* YN side */
		ind_y = 0;
	else                                                /* YP side */
		ind_y = 1;
	
/* resolve side flags;
	 note: face_is -> side[*] is used for detection whether output of the ellipse should be performed
	 only if the ellipse does not intersect tile edge (x|y_count = 0);
	 this implies that center of such ellipse is always inside tile and it is therefore safe to use just inequalities;
	 if the center of the ellipse is on tile edge, the side should be set to BOTH but since side is not used in such a case, it is not done ! */

	if(ellipse -> center.x > tile.corner[ind_x].x)
		face_is -> side[X] = POS;
	else
		face_is -> side[X] = NEG;
	
	if(ellipse -> center.y > tile.corner[ind_y].y)
		face_is -> side[Y] = POS;
	else
		face_is -> side[Y] = NEG;

/* resolve intersections with tile sides */

	if(ext_point_x[0].x - tile.corner[0].x > epsilon && tile.corner[1].x - ext_point_x[1].x > epsilon &&
		 ext_point_y[0].y - tile.corner[0].y > epsilon && tile.corner[1].y - ext_point_y[1].y > epsilon)return(NO); /* inclusion is inside tile */
	
	if(ext_point_x[0].x < tile.corner[0].x + epsilon || ext_point_x[1].x > tile.corner[1].x - epsilon)x_intersection = YES;
	if(ext_point_y[0].y < tile.corner[0].y + epsilon || ext_point_y[1].y > tile.corner[1].y - epsilon)y_intersection = YES;
	
	shift_vec.x = shift_vec.y = 0.0;
	
	diff_x = diff_y = max_sz;    /* diffs of ellipse extreme points */
	
/* diff_x = x-coordinate of relevant tile side - x-coordinate of relevant extreme point;
	 diff_y = y-coordinate of relevant tile side - y-coordinate of relevant extreme point;

   diff_x|y are computed in resolve_ellipse_x|y_side_touching;
	 note that x|y_intersection is set to NO if touching is identified;
	 note that x|y_count is not set if touching is identified */

	if(x_intersection == YES)resolve_ellipse_x_side_touching();
	if(y_intersection == YES)resolve_ellipse_y_side_touching();
	
	size = size_vec(shift_vec);
	if(size != 0.0){

#ifdef DEBUG_PRINT
		fprintf(stderr, "Shifting inclusion %ld by [%e %e] due to side touching\n", inclusion -> id, shift_vec.x, shift_vec.y);
#endif		
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		for(i = 0; i < 2; i++){
			draw_point(&ext_point_x[i], 0, bgcolor);
			draw_point(&ext_point_y[i], 0, bgcolor);
		}
#endif
		
		copy_vec(total_shift, shift_vec);
		add_vec(ellipse -> center, ellipse -> center, shift_vec);
		for(i = 0; i < 2; i++){
			add_vec(ext_point_x[i], ext_point_x[i], shift_vec);
			add_vec(ext_point_y[i], ext_point_y[i], shift_vec);
		}
		diff_x -= shift_vec.x;
		diff_y -= shift_vec.y;
		
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		for(i = 0; i < 2; i++){
			draw_point(&ext_point_x[i], 0, cyan);
			draw_point(&ext_point_y[i], 0, green);
		}
#endif
	}
	
/* resolve intersections with tile edges */

	x_diff = y_diff = max_sz;     /* diffs of ellipse intersections with tile edges */

/* x_diff = x-coordinate of relevant tile side (corner) - x-coordinate of the closest ellipse intersection with relevant tile x-edge;
	 y_diff = y-coordinate of relevant tile side (corner) - y-coordinate of the closest ellipse intersection with relevant tile y-edge;

   note: diffs, counts and coords are computed in get_ellipse_x|y_edge_intersection */

	get_ellipse_x_edge_intersection(YES);
	get_ellipse_y_edge_intersection(YES);

	inclusion -> corner = OUT;

	if(x_count == 0 && y_count == 0)return(NO);
	if(x_count + y_count == 1)return(YES);
		
	a_diff_x = fabs(diff_x);
	a_diff_y = fabs(diff_y);

	a_x_diff = fabs(x_diff);
	a_y_diff = fabs(y_diff);

	eps = epsilon;
	subdivision = 0;

	while(YES){
		if(a_x_diff >= eps && a_y_diff >= eps){
			if(x_count == 2 && y_count == 2){

/* check that tile corner is between tile edges intersections
	 results should be same no matter which tile edge is considered */

				coord = tile.corner[ind_x].x;
				if(x_coord[0] < coord && x_coord[1] > coord){
					inclusion -> corner = IN;
					
#ifdef DEBUG
					coord = tile.corner[ind_y].y;
					if(y_coord[0] > coord || y_coord[1] < coord)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
				}
			}

			return(YES);
		}
		
		if(subdivision == 0){
			if((a_x_diff < min_epsilon || x_count == 0) && 
				 (a_y_diff < min_epsilon || y_count == 0)){
				shifting = NO;
			
/* shift relevant intersection to tile corner but do not shift the inclusion;
	 check at first that such shift will not cause that intersections of the same edge will get closer to each other than min_epsilon */

				corner_x_between = corner_y_between = NO;
				if(x_count == 2){
					if(x_coord[0] < tile.corner[ind_x].x && x_coord[1] > tile.corner[ind_x].x){
						corner_x_between = YES;
						if(x_coord[1] - x_coord[0] - a_x_diff < min_epsilon)shifting = YES;
					}
				}
				if(y_count == 2){
					if(y_coord[0] < tile.corner[ind_y].y && y_coord[1] > tile.corner[ind_y].y){
						corner_y_between = YES;
						if(y_coord[1] - y_coord[0] - a_y_diff < min_epsilon)shifting = YES;
					}
				}

				if(shifting == NO){
					inclusion -> corner = BND;
					
					if(x_count != 0){
						ind = 0;
						if(x_count == 2){
						
/* sign of diff determines intersection to be shifted */

							if(corner_x_between == YES){
								if(x_diff < 0.0)ind = 1;
							}
							else{
								if(x_diff > 0.0)ind = 1;
							}
						}
						x_coord[ind] += x_diff;
					}
					if(y_count != 0){
						ind = 0;
						if(y_count == 2){
							
/* sign of diff determines intersection to be shifted */

							if(corner_y_between == YES){
								if(y_diff < 0.0)ind = 1;
							}
							else{
								if(y_diff > 0.0)ind = 1;
							}
						}
						y_coord[ind] += y_diff;
					}

					return(YES);
				}
			}
		}
		
		if(no_shift == YES)error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);
		
/* check for various shifts and choose that which minimizes maximal shift while not exeeding MULTIPL_EPS * epsilon */

/* IMPORTANT: when shifting to corner check that length of intersection with tile edge > 2.0 * min_epsilon;
	 the length is precomputed in get_ellipse_x|y_edge_intersection */

		max_shift = MULTIPLE_EPS * epsilon + min_epsilon;

		if(x_count == 1 && y_count == 1){
			
/* firstly check for detouching in either direction */

			shift_vec_tmp.y = 0.0;
			if(x_diff > 0)
				shift_vec_tmp.x = -eps;
			else
				shift_vec_tmp.x = eps;
			
			check_shift_tmp(OUT);
			
			shift_vec_tmp.x = 0.0;
			if(y_diff > 0)
				shift_vec_tmp.y = -eps;
			else
				shift_vec_tmp.y = eps;
			
			check_shift_tmp(OUT);

/* then consider shift to corner */

			if(dist_y > 2.0 * min_epsilon){
				if(a_x_diff > eps){
					shift_vec_tmp.x = x_diff;
					shift_vec_tmp.y = 0.0;
					
					check_shift_tmp(BND);
				}
			}
			
			if(dist_x > 2.0 * min_epsilon){
				if(a_y_diff > eps){
					shift_vec_tmp.x = 0.0;
					shift_vec_tmp.y = y_diff;
					
					check_shift_tmp(BND);
				}
			}
		}
		else if(x_count == 1 && y_count == 2){

/* try detouch in y direction at first */

			shift_vec_tmp.x = 0.0;
			if(y_diff > 0.0)
				shift_vec_tmp.y = -eps;
			else
				shift_vec_tmp.y = eps;
			
			check_shift_tmp(OUT);
			
/* then consider x shift to pass through corner */

			if(dist_y > 2.0 * min_epsilon){
				shift_vec_tmp.x = x_diff;
				shift_vec_tmp.y = 0.0;
				
				check_shift_tmp(BND);
			}
			
/* then try enforce touching in x direction */

			shift_vec_tmp.x = diff_x;
			shift_vec_tmp.y = 0.0;
			
			check_shift_tmp(OUT);
		}
		else if(x_count == 2 && y_count == 1){
			
/* try detouch in x direction at first */

			shift_vec_tmp.y = 0.0;
			if(x_diff > 0.0)
				shift_vec_tmp.x = -eps;
			else
				shift_vec_tmp.x = eps;
			
			check_shift_tmp(OUT);

/* then consider y shift to pass through corner */

			if(dist_x > 2.0 * min_epsilon){
				shift_vec_tmp.x = 0.0;
				shift_vec_tmp.y = y_diff;
				
				check_shift_tmp(BND);
			}

/* then try enforce touching in y direction */

			shift_vec_tmp.x = 0.0;
			shift_vec_tmp.y = diff_y;
			
			check_shift_tmp(OUT);
		}
		else{ /* x_count == 2 && y_count == 2 */

/* first try shift in either direction to pass through corner */

			if(dist_y > 2.0 * min_epsilon){
				if(fabs(diff_x - x_diff) >= eps){
					shift_vec_tmp.x = x_diff;
					shift_vec_tmp.y = 0.0;
					
					check_shift_tmp(BND);
				}
			}
			if(dist_x > 2.0 * min_epsilon){
				if(fabs(diff_y - y_diff) >= eps){
					shift_vec_tmp.x = 0.0;
					shift_vec_tmp.y = y_diff;
					
					check_shift_tmp(BND);
				}
			}
			
/* then try enforce touching in either direction */

			if(fabs(diff_x - x_diff) >= eps){
				shift_vec_tmp.x = diff_x;
				shift_vec_tmp.y = 0.0;
				
				check_shift_tmp(OUT);
			}
			if(fabs(diff_y - y_diff) >= eps){
				shift_vec_tmp.x = 0.0;
				shift_vec_tmp.y = diff_y;
				
				check_shift_tmp(OUT);
			}
			
/* then try enforce touching in both direction */

			shift_vec_tmp.x = diff_x;
			shift_vec_tmp.y = diff_y;
			
			check_shift_tmp(OUT);
		}

		if(max_shift < MULTIPLE_EPS * epsilon)break;

		eps /= 2.0;
		subdivision++;
		if(subdivision == max_subdivisions || eps <= min_epsilon * 10.0)error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);
	}
	
#ifdef PARAMETRIC_ROUTINES
/* ctrl points are not shifted */
#endif

	if(shift_vec.x == 0.0){
		ellipse -> center.y += shift_vec.y;
		for(i = 0; i < 2; i++){
			ext_point_x[i].y += shift_vec.y;
			ext_point_y[i].y += shift_vec.y;
		}
		total_shift.y += shift_vec.y;
		if((diff_y -= shift_vec.y) < min_epsilon / 10.0){
			diff_y = 0.0;
			y_touching = YES;
			y_intersection = NO;
		}

		y_diff -= shift_vec.y;
		for(i = 0; i < y_count; i++)y_coord[i] += shift_vec.y;

		get_ellipse_x_edge_intersection(NO);
	}
	else if(shift_vec.y == 0.0){
		ellipse -> center.x += shift_vec.x;
		for(i = 0; i < 2; i++){
			ext_point_x[i].x += shift_vec.x;
			ext_point_y[i].x += shift_vec.x;
		}
		total_shift.x += shift_vec.x;
		if((diff_x -= shift_vec.x) < min_epsilon / 10.0){
			diff_x = 0.0;
			x_touching = YES;
			x_intersection = NO;
		}

		x_diff -= shift_vec.x;
		for(i = 0; i < x_count; i++)x_coord[i] += shift_vec.x;

		get_ellipse_y_edge_intersection(NO);
	}
	else{
		add_vec(ellipse -> center, ellipse -> center, shift_vec);
		for(i = 0; i < 2; i++){
			add_vec(ext_point_x[i], ext_point_x[i], shift_vec);
			add_vec(ext_point_y[i], ext_point_y[i], shift_vec);
		}
		add_vec(total_shift, total_shift, shift_vec);
		if((diff_x -= shift_vec.x) < min_epsilon / 10.0){
			diff_x = 0.0;
			x_touching = YES;
			x_intersection = NO;
		}
		if((diff_y -= shift_vec.y) < min_epsilon / 10.0){
			diff_y = 0.0;
			y_touching = YES;
			y_intersection = NO;
		}

		get_ellipse_x_edge_intersection(NO);
		get_ellipse_y_edge_intersection(NO);
	}

	inclusion -> corner = corner_tmp;

	if(eps != epsilon)error_message(WARNING, "Epsilon temporarily reduced to %e when processing inclusion %ld", inclusion -> id);

	return(YES);
}


/* compute distance of relevant ellipse extreme point in x direction from tile side perpendicular to x direction
	 and if closer to that side than epsilon while not touching it, shiftting yielding touching or minimal shifting
	 inside tile is computed;

	 IMPORTANT: I do rely on fact that inclusion shifting due to edge touching does not interfere
	 with shifting due to side touching (this is ensured by min > epsilon * 10);
	 this implies that I can safely use shift_vec.x|y = val instead of shift_vec.x|y += val */

static void
resolve_ellipse_x_side_touching(void)
{
	double a_diff_x;

	diff_x = tile.corner[ind_x].x - ext_point_x[ind_x].x;
	a_diff_x = fabs(diff_x);
	if(a_diff_x >= epsilon)return;

	if(no_shift == YES)error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);

	x_intersection = NO;

#ifdef OVERDETOUCH

/* note: also simple detouch is handled here */

	if(ind_x == 0)
		shift_vec.x = diff_x + epsilon;
	else
		shift_vec.x = diff_x - epsilon;

	if(fabs(shift_vec.x) <= MULTIPLE_EPS * epsilon)return;
	shift_vec.x = 0.0;

	error_message(WARNING, "Overdetouch of inclusion %ld in x-direction suppressed due to MULTIPLE_EPS (%f)", inclusion -> id, MULTIPLE_EPS);
#endif
	
#ifdef DETOUCH
	if(a_diff_x >= min_epsilon){
		if(ind_x == 0){
			if(diff_x < 0.0){
				shift_vec.x = diff_x + epsilon;
				return;
			}
		}
		else{
			if(diff_x > 0.0){
				shift_vec.x = diff_x - epsilon;
				return;
			}
		}
	}
#endif

	x_touching = YES;
	shift_vec.x = diff_x;
}


/* compute distance of relevant ellipse extreme point in y direction from tile side perpendicular to y direction
	 and if closer to that side than epsilon while not touching it, shiftting yielding touching or minimal shifting
	 inside tile is computed;

	 IMPORTANT: I do rely on fact that inclusion shifting due to edge touching does not interfere
	 with shifting due to side touching (this is ensured by min > epsilon * 10);
	 this implies that I can safely use shift_vec.x|y|z = val instead of shift_vec.x|y|z += val */

static void
resolve_ellipse_y_side_touching(void)
{
	double a_diff_y;

	diff_y = tile.corner[ind_y].y - ext_point_y[ind_y].y;
	a_diff_y = fabs(diff_y);
	if(a_diff_y >= epsilon)return;

	if(no_shift == YES)error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);

	y_intersection = NO;

#ifdef OVERDETOUCH

/* note: also simple detouch is handled here */

	if(ind_y == 0)
		shift_vec.y = diff_y + epsilon;
	else
		shift_vec.y = diff_y - epsilon;

	if(fabs(shift_vec.y) <= MULTIPLE_EPS * epsilon)return;
	shift_vec.y = 0.0;

	error_message(WARNING, "Overdetouch of inclusion %ld in y-direction suppressed due to MULTIPLE_EPS (%f)", inclusion -> id, MULTIPLE_EPS);
#endif
	
#ifdef DETOUCH
	if(a_diff_y >= min_epsilon){
		if(ind_y == 0){
			if(diff_y < 0.0){
				shift_vec.y = diff_y + epsilon;
				return;
			}
		}
		else{
			if(diff_y > 0.0){
				shift_vec.y = diff_y - epsilon;
				return;
			}
		}
	}
#endif

	y_touching = YES;
	shift_vec.y = diff_y;
}


/* compute intersection points and corresponding diff of ellipse with x edge;
	 if 2 intersections occur, diff is related to intersection point closer to corner;
	 sign of diffs determines the orientation of shift to enforce passing through corner */

static void
get_ellipse_x_edge_intersection(logic fix)
{
	point_rec xy_point[2], xy_pnt;
	double shift;
	int i;

	if(y_touching == YES){                                                  /* reuse ellipse extreme point */
		x_count = 1;
		x_coord[0] = ext_point_y[ind_y].x;

		x_diff = tile.corner[ind_x].x - x_coord[0];
	}
	else{
		if(y_intersection == NO)
			x_count = 0;
		else{
			if((x_count = get_ellipse_line_intersection(ellipse, &(tile.corner[ind_y]), &dir_x, xy_point)) != 2)error_message(GENERAL_ERROR, "Unexpected situation");

			if(fabs(xy_point[0].x - xy_point[1].x) <= 2.0 * min_epsilon){
				if(fix == NO || no_shift == YES)
					error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);
				else{

/* try to fix pathological case by shifting */

					if(diff_y > 0.0){
						if(shift_vec.y != 0.0)          /* detouch applied */
							shift = epsilon;              /* enforce touching (even if DETOUCH is defined) */
						else{
							if(diff_y <= MULTIPLE_EPS * epsilon)
								shift = diff_y;             /* enforce touching from distance up to MULTIPLE_EPS * epsilon */
							else{

/* apply max permissible shift while keeping inclusion center inside or on boundary of tile */

								if((shift = tile.corner[ind_y].y - ellipse -> center.y) < -MULTIPLE_EPS * epsilon)shift = -MULTIPLE_EPS * epsilon;
							}
						}
					}
					else{
						if(shift_vec.y != 0.0)          /* detouch applied */
							shift = epsilon;              /* enforce touching (even if DETOUCH is defined) */
						else{
							if(-diff_y <= MULTIPLE_EPS * epsilon)
								shift = diff_y;             /* enforce touching from distance up to MULTIPLE_EPS * epsilon */
							else{

/* apply max permissible shift while keeping inclusion center inside or on boundary of tile */

								if((shift = tile.corner[ind_y].y - ellipse -> center.y) > MULTIPLE_EPS * epsilon)shift = MULTIPLE_EPS * epsilon;
							}
						}
					}

#ifdef DEBUG_PRINT
					fprintf(stderr, "Shifting inclusion %ld by [%e %e] due to small side intersection\n", inclusion -> id, 0.0, shift);
#endif		
#ifdef DRAW_EXTREME_POINT_ELLIPSE
					for(i = 0; i < 2; i++){
						draw_point(&ext_point_x[i], 0, bgcolor);
						draw_point(&ext_point_y[i], 0, bgcolor);
					}
#endif
#ifdef PARAMETRIC_ROUTINES
/* ctrl points are not shifted */
#endif

					total_shift.y += shift;
					ellipse -> center.y += shift;
					for(i = 0; i < 2; i++){
						ext_point_x[i].y += shift;
						ext_point_y[i].y += shift;
					}
					shift_vec.y += shift;
					if((diff_y -= shift) < min_epsilon / 10.0){
						diff_y = 0.0;
						y_touching = YES;
						y_intersection = NO;
					}
	
#ifdef DRAW_EXTREME_POINT_ELLIPSE
					for(i = 0; i < 2; i++){
						draw_point(&ext_point_x[i], 0, cyan);
						draw_point(&ext_point_y[i], 0, green);
					}
#endif
					
					get_ellipse_x_edge_intersection(NO);         /* prevent fixing */
					return;
				}
			}
					
			if(xy_point[0].x < xy_point[1].x){
				x_coord[0] = xy_point[0].x;
				x_coord[1] = xy_point[1].x;
			}
			else{
				x_coord[0] = xy_point[1].x;
				x_coord[1] = xy_point[0].x;
			}
			
			if(x_coord[0] >= tile.corner[ind_x].x)
				x_diff = tile.corner[ind_x].x - x_coord[0];
			else if(x_coord[1] <= tile.corner[ind_x].x)
				x_diff = tile.corner[ind_x].x - x_coord[1];
			else if(x_coord[1] - tile.corner[ind_x].x <= tile.corner[ind_x].x - x_coord[0])
				x_diff = tile.corner[ind_x].x - x_coord[1];
			else
				x_diff = tile.corner[ind_x].x - x_coord[0];
		}
	}

	if(x_count != 0){

/* compute also distance between intersection points of ellipse with line paralell to y edge
	 and passing through the intersection for which the diff was computed */

		xy_pnt.x = tile.corner[ind_x].x - x_diff;
		xy_pnt.y = 0.0;

		dist_y = 0.0;
		if(get_ellipse_line_intersection(ellipse, &xy_pnt, &dir_y, xy_point) == 2){
			dist_y = dist_point(xy_point[0], xy_point[1]);
			dist_y = sqrt(dist_y);
		}

#ifdef DRAW_INTERSECTING_POINT
		xy_pnt.y = tile.corner[ind_y].y;
		for(i = 0; i < x_count; i++){
			xy_pnt.x = x_coord[i];
			draw_point(&xy_pnt, 0, green);
		}
#endif
	}
}


/* compute intersection points and corresponding diff of ellipse with y edge;
	 if 2 intersections occur, diff is related to intersection point closer to corner;
	 sign of diffs determines the orientation of shift to enforce passing through corner */

static void
get_ellipse_y_edge_intersection(logic fix)
{
	point_rec xy_point[2], xy_pnt;
	double shift;
	int i;

	if(x_touching == YES){                                                  /* reuse ellipse extreme point */
		y_count = 1;
		y_coord[0] = ext_point_x[ind_x].y;

		y_diff = tile.corner[ind_y].y - y_coord[0];
	}
	else{
		if(x_intersection == NO)
			y_count = 0;
		else{
			if((y_count = get_ellipse_line_intersection(ellipse, &(tile.corner[ind_x]), &dir_y, xy_point)) != 2)error_message(GENERAL_ERROR, "Unexpected situation");

			if(fabs(xy_point[0].y - xy_point[1].y) <= 2.0 * min_epsilon){
				if(fix == NO || no_shift == YES)
					error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);
				else{

/* try to fix pathological case by shifting */

					if(diff_x > 0.0){
						if(shift_vec.x != 0.0)          /* detouch applied */
							shift = epsilon;              /* enforce touching (even if DETOUCH is defined) */
						else{
							if(diff_x <= MULTIPLE_EPS * epsilon)
								shift = diff_x;             /* enforce touching from distance up to MULTIPLE_EPS * epsilon */
							else{

/* apply max permissible shift while keeping inclusion center inside or on boundary of tile */

								if((shift = tile.corner[ind_x].x - ellipse -> center.x) < -MULTIPLE_EPS * epsilon)shift = -MULTIPLE_EPS * epsilon;
							}
						}
					}
					else{
						if(shift_vec.x != 0.0)          /* detouch applied */
							shift = epsilon;              /* enforce touching (even if DETOUCH is defined) */
						else{
							if(-diff_x <= MULTIPLE_EPS * epsilon)
								shift = diff_x;             /* enforce touching from distance up to MULTIPLE_EPS * epsilon */
							else{

/* apply max permissible shift while keeping inclusion center inside or on boundary of tile */

								if((shift = tile.corner[ind_x].x - ellipse -> center.x) > MULTIPLE_EPS * epsilon)shift = MULTIPLE_EPS * epsilon;
							}
						}
					}

#ifdef DEBUG_PRINT
					fprintf(stderr, "Shifting inclusion %ld by [%e %e] due to small side intersection\n", inclusion -> id, shift, 0.0);
#endif		
#ifdef DRAW_EXTREME_POINT_ELLIPSE
					for(i = 0; i < 2; i++){
						draw_point(&ext_point_x[i], 0, bgcolor);
						draw_point(&ext_point_y[i], 0, bgcolor);
					}
#endif
#ifdef PARAMETRIC_ROUTINES
/* ctrl points are not shifted */
#endif

					total_shift.x += shift;
					ellipse -> center.x += shift;
					for(i = 0; i < 2; i++){
						ext_point_x[i].x += shift;
						ext_point_y[i].x += shift;
					}
					shift_vec.x += shift;
					if((diff_x -= shift) < min_epsilon / 10.0){
						diff_x = 0.0;
						x_touching = YES;
						x_intersection = NO;
					}
					
					get_ellipse_y_edge_intersection(NO);         /* prevent fixing */
					return;
				}
			}
			
			if(xy_point[0].y < xy_point[1].y){
				y_coord[0] = xy_point[0].y;
				y_coord[1] = xy_point[1].y;
			}
			else{
				y_coord[0] = xy_point[1].y;
				y_coord[1] = xy_point[0].y;
			}
			
			if(y_coord[0] >= tile.corner[ind_y].y)
				y_diff = tile.corner[ind_y].y - y_coord[0];
			else if(y_coord[1] <= tile.corner[ind_y].y)
				y_diff = tile.corner[ind_y].y - y_coord[1];
			else if(y_coord[1] - tile.corner[ind_y].y <= tile.corner[ind_y].y - y_coord[0])
				y_diff = tile.corner[ind_y].y - y_coord[1];
			else
				y_diff = tile.corner[ind_y].y - y_coord[0];
		}
	}

	if(y_count != 0){

/* compute also distance between intersection points of ellipse with line paralell to x edge
	 and passing through the intersection for which the diff was computed */

		xy_pnt.x = 0.0;
		xy_pnt.y = tile.corner[ind_y].y - y_diff;
		
		dist_x = 0.0;
		if(get_ellipse_line_intersection(ellipse, &xy_pnt, &dir_x, xy_point) == 2){
			dist_x = dist_point(xy_point[0], xy_point[1]);
			dist_x = sqrt(dist_x);
		}
	
#ifdef DRAW_INTERSECTING_POINT
		xy_pnt.x = tile.corner[ind_x].x;
		for(i = 0; i < y_count; i++){
			xy_pnt.y = y_coord[i];
			draw_point(&xy_pnt, 0, cyan);
		}
#endif
	}
}


/* points point1 and point2 are expected to lie on the ellipse in adjacent quadrants */

static double
get_ellipse_weight(ellipse_rec *ellipse, point_rec *point1, point_rec *point2, trans_matrix *trans_g2l)
{
	point_rec pnt1, pnt2;
	double weight, cosine, Radius, radius, x1, y1, x2, y2;

#ifdef DEBUG
	double val1, val2;
#endif

	Radius = ellipse -> max;
	radius = ellipse -> min;

	transform_point_from_global_to_local(point1, &pnt1, &(ellipse -> center), trans_g2l);
	transform_point_from_global_to_local(point2, &pnt2, &(ellipse -> center), trans_g2l);

	x1 = pnt1.x / Radius;
	y1 = pnt1.y / radius;

	x2 = pnt2.x / Radius;
	y2 = pnt2.y / radius;

#ifdef DEBUG
	val1 = x1 * x1 + y1 * y1;
	val2 = x2 * x2 + y2 * y2;

	if(fabs(1.0 - val1) > 1.0e-10 || fabs(1.0 - val2) > 1.0e-10)error_message(GENERAL_ERROR, "Unexpected situation");

	if(x1 * x2 < 0.0 && y1 * y2 < 0.0)error_message(GENERAL_ERROR, "Unexpected situation");
	if(x1 * x2 > 0.0 && y1 * y2 > 0.0)error_message(GENERAL_ERROR, "Unexpected situation");
#endif

#if 0
	if(x1 * x2 >= 0.0)
	  cosine = x1 * x2 - fabs(y1 * y2);
	else
	  cosine = y1 * y2 - fabs(x1 * x2);
#else

/* I do rely on fact that points are in adjacent quadrants;
	 this implies
	 if(x1 * x2 >= 0.0) ==> -fabs(y1 * y2) = y1 * y2
	 if(x1 * x2 <= 0.0) ==> -fabs(x1 * x2) = x1 * x2 */

	cosine = x1 * x2 + y1 * y2;
#endif

	weight = sqrt((1.0 + cosine) / 2.0);

	return(weight);
}


#ifdef CHECK_OVERLAP

static void
check_active_inclusion_overlap(ellipse_rec *ellipse1, int per_x, int per_y, long i)
{
	inclusion_rec *inclusion = NULL;
	ellipse_rec *ellipse2 = NULL;
	double per_shift_x, per_shift_y;
	int j, k, x_per, y_per;

#ifndef ALLOW_OVERLAP
	logic shift;

	shift = inclusion_array[i].shift;
#endif

	for(j = 0; j < i; j++){
		inclusion = &inclusion_array[j];
		ellipse2 = &(inclusion -> ellipse);

#ifndef ALLOW_OVERLAP

/* there is no need to check for intersection if neither of inclusions has been shifted and inclusions are not allowed to overlap in reference tile */

		if(shift == NO){
			if(inclusion -> shift == NO)continue;
		}
#endif

		for(k = 0; k < 4; k++){
			if(inclusion -> active[k] == NO)continue;
			
			if(shifting_x[k] == 0){
				per_shift_x = 0.0;
				x_per = 0;
			}
			else{
				x_per = 1;
				if(ellipse2 -> center.x < tile.center.x)            /* XN side */
					per_shift_x = tile.size.x;
				else                                                /* XP side */
					per_shift_x = -tile.size.x;
			}
			if(shifting_y[k] == 0){
				per_shift_y = 0.0;
				y_per = 0;
			}
			else{
				y_per = 1;
				if(ellipse2 -> center.y < tile.center.y)            /* YN side */
					per_shift_y = tile.size.y;
				else                                                /* YP side */
					per_shift_y = -tile.size.y;
			}
			
			shift_ellipse(*ellipse2, per_shift_x, per_shift_y);
			if(check_ellipse_ellipse_overlap(ellipse1, ellipse2, epsilon) == YES)
				error_message(GENERAL_ERROR, "Inclusion %ld (%d %d) intersects inclusion %ld (%d %d)", i + 1, per_x, per_y, j + 1, x_per, y_per);
			shift_ellipse(*ellipse2, -per_shift_x, -per_shift_y);
		}
	}
}
		

static void
check_inclusion_periodic_overlap(ellipse_rec *ellipse1, int per_x, int per_y, long i)
{
	inclusion_rec *inclusion = NULL;
	ellipse_rec *ellipse2 = NULL;
	double per_shift_x, per_shift_y;
	int j;

	for(j = 0; j < i; j++){
		inclusion = &inclusion_array[j];
		ellipse2 = &(inclusion -> ellipse);

/* primary instance */

		if(check_ellipse_ellipse_overlap(ellipse1, ellipse2, epsilon) == YES)
			error_message(GENERAL_ERROR, "Inclusion %ld (%d %d) intersects inclusion %ld (0 0)", i + 1, per_x, per_y, j + 1);

/* secondary instances */

		if(inclusion -> edge_is[Y] == NULL)                   /* <==> inclusion -> face_is[X] == NULL */
			per_shift_x = 0.0;
		else{
			if(ellipse2 -> center.x < tile.center.x)            /* XN side */
				per_shift_x = tile.size.x;
			else                                                /* XP side */
				per_shift_x = -tile.size.x;

			shift_ellipse(*ellipse2, per_shift_x, 0.0);
			if(check_ellipse_ellipse_overlap(ellipse1, ellipse2, epsilon) == YES)
				error_message(GENERAL_ERROR, "Inclusion %ld (%d %d) intersects inclusion %ld (1 0)", i + 1, per_x, per_y, j + 1);
			shift_ellipse(*ellipse2, -per_shift_x, 0.0);
		}
		if(inclusion -> edge_is[X] == NULL)                   /* <==> inclusion -> face_is[Y] == NULL */
			per_shift_y = 0.0;
		else{
			if(ellipse2 -> center.y < tile.center.y)            /* YN side */
				per_shift_y = tile.size.y;
			else                                                /* YP side */
				per_shift_y = -tile.size.y;

			shift_ellipse(*ellipse2, 0.0, per_shift_y);
			if(check_ellipse_ellipse_overlap(ellipse1, ellipse2, epsilon) == YES)
				error_message(GENERAL_ERROR, "Inclusion %ld (%d %d) intersects inclusion %ld (0 1)", i + 1, per_x, per_y, j + 1);
			shift_ellipse(*ellipse2, 0.0, -per_shift_y);
		}

		if(per_shift_x != 0.0 && per_shift_y != 0.0){
			if(inclusion_array[j].corner == IN){
				shift_ellipse(*ellipse2, per_shift_x, per_shift_y);
				if(check_ellipse_ellipse_overlap(ellipse1, ellipse2, epsilon) == YES)
					error_message(GENERAL_ERROR, "Inclusion %ld (%d %d) intersects inclusion %ld (1 1)", i + 1, per_x, per_y, j + 1);
				shift_ellipse(*ellipse2, -per_shift_x, -per_shift_y);
			}
		}
	}
}

#endif


/* ellipses greater/smaller by positive/negative eps */

static logic
check_ellipse_ellipse_overlap(ellipse_rec *ellipse1, ellipse_rec *ellipse2, double eps)
{
	point_rec pnt1, pnt2, p1, p2;
	double max, max1, max2;
	double dist, dist1, dist2, value;

#if 1
	vector_rec diff;
#else /* obsolete */
	double t1, tt1, ttt1, t2, tt2, ttt2, delta_t = EPSILON_T;
	int q1, qq1, qqq1, q2, qq2, qqq2;
#endif

	if(ellipse1 == ellipse2)return(NO);

	max1 = ellipse1 -> max + eps;
	max2 = ellipse2 -> max + eps;
	max = max1 + max2;

/* intersection of circumsribed circles */
	dist = dist_point(ellipse1 -> center, ellipse2 -> center);
	if(dist > max * max)return(NO);

/* center of ellipse2 inside ellipse1 */
	if(check_point_inside_ellipse(&(ellipse2 -> center), ellipse1, eps, &value) == YES)return(YES);    /* accounts for eps */

/* center of ellipse1 inside ellipse2 */
	if(check_point_inside_ellipse(&(ellipse1 -> center), ellipse2, eps, &value) == YES)return(YES);    /* accounts for eps */

/* intersection of circle circumscribed to ellipse1 with ellipse2 */
#if 1
	project_point_to_ellipse(&(ellipse1 -> center), ellipse2, &pnt2);                                  /* exact */
#else /* obsolete */
	q2 = project_point_to_ellipse_t_quadrant(&(ellipse1 -> center), ellipse2, &pnt2, &t2);             /* exact */
#endif
	dist1 = dist_point(ellipse1 -> center, pnt2);
	if(dist1 > max1 * max1)return(NO);                                                                 /* accounts for eps */

/* closest point on ellipse2 to center of ellipse1 inside ellipse1 */
	if(check_point_inside_ellipse(&pnt2, ellipse1, eps, &value) == YES)return(YES);                    /* accounts for eps */

/* intersection of circle circumscribed to ellipse2 with ellipse1 */
#if 1
	project_point_to_ellipse(&(ellipse2 -> center), ellipse1, &pnt1);                                  /* exact */
#else /* obsolete */
	q1 = project_point_to_ellipse_t_quadrant(&(ellipse2 -> center), ellipse1, &pnt1, &t1);             /* exact */
#endif
	dist2 = dist_point(ellipse2 -> center, pnt1);
	if(dist2 > max2 * max2)return(NO);                                                                 /* accounts for eps */
	
/* closest point on ellipse1 to center of ellipse2 inside ellipse2 */
	if(check_point_inside_ellipse(&pnt1, ellipse2, eps, &value) == YES)return(YES);                    /* accounts for eps */

#if 1
/* find closest points starting from pnt1 */
	while(YES){
		copy_vec(diff, pnt1);
		project_point_to_ellipse(&pnt1, ellipse2, &p2);                                                  /* exact */
		project_point_to_ellipse(&p2, ellipse1, &pnt1);                                                  /* exact */
		sub_vec(diff, diff, pnt1);
		if(size_vec(diff) < EPSILON * EPSILON)break;
	}

/* find closest points starting from pnt2 */
	while(YES){
		copy_vec(diff, pnt2);
		project_point_to_ellipse(&pnt2, ellipse1, &p1);                                                  /* exact */
		project_point_to_ellipse(&p1, ellipse2, &pnt2);                                                  /* exact */
		sub_vec(diff, diff, pnt2);
		if(size_vec(diff) < EPSILON * EPSILON)break;
	}

	aver_point(pnt1, pnt1, p1);
	aver_point(pnt2, pnt2, p2);
#else /* obsolete */

/* the iterations should not oscillate between quadrants because the circle tests have been already done */

/* find closest points starting from pnt1 */
	while(YES){
		qqq2 = project_point_to_ellipse_t_quadrant(&pnt1, ellipse2, &p2, &ttt2);                         /* exact */
		qq1 = project_point_to_ellipse_t_quadrant(&p2, ellipse1, &pnt1, &tt1);                           /* exact */
		if(fabs(t1 - tt1) < delta_t)break;
		t1 = tt1;
	}

/* find closest points starting from pnt2 */
	while(YES){
		qqq1 = project_point_to_ellipse_t_quadrant(&pnt2, ellipse1, &p1, &ttt1);                         /* exact */
		qq2 = project_point_to_ellipse_t_quadrant(&p1, ellipse2, &pnt2, &tt2);                           /* exact */
		if(fabs(t2 - tt2) < delta_t)break;
		t2 = tt2;
	}

/* there is a problem that intersections of ellipses may be in different quadrants;
	 in this case averaging of parameters makes no sense but it is clear that ellipses do intersect */

	if(qq1 != qqq1 || qq2 != qqq2)return(YES);

	t1 = (tt1 + ttt1) / 2.0;
	t2 = (tt2 + ttt2) / 2.0;
	get_ellipse_point_t_quadrant(ellipse1, qq1, t1, &pnt1);                                           /* exact */
	get_ellipse_point_t_quadrant(ellipse2, qq2, t2, &pnt2);                                           /* exact */
#endif

/* check pnt1 inside ellipse2 */
	if(check_point_inside_ellipse(&pnt1, ellipse2, eps, &value) == YES)return(YES);                   /* accounts for eps */
	
/* check pnt2 inside ellipse1 */
	if(check_point_inside_ellipse(&pnt2, ellipse1, eps, &value) == YES)return(YES);                   /* accounts for eps */

	return(NO);
} 



static logic
get_plane_line_intersection(point_rec *point, vector_rec *normal, point_rec *pnt, vector_rec *dir, point_rec *intersection)
{
	vector_rec vec;
	double product, t;
	
#ifdef DEBUG
	double size;

	size = sqrt(size_vec(*normal));
	if(fabs(size - 1.0) > EPSILON)
		error_message(GENERAL_ERROR, "Normal not normalized");
	size = sqrt(size_vec(*dir));
	if(fabs(size - 1.0) > EPSILON)
		error_message(GENERAL_ERROR, "Direction not normalized");
#endif

	product = dot_product(*normal, *dir);
	if(fabs(product) < EPSILON)return(NO);
	
	//t = (dot_product(*normal, *point) - dot_product(*normal, *pnt)) / product;
	sub_vec(vec, *point, *pnt);
	t = dot_product(*normal, vec) / product;
	
	intersection -> x = pnt -> x + t * dir -> x;
	intersection -> y = pnt -> y + t * dir -> y;

	return(YES);
}


static double
project_point_to_line(point_rec *pnt, point_rec *point, vector_rec *dir, point_rec *projection)
{
	vector_rec vec;
	double product, dist;

#ifdef DEBUG
	double size;

	size = sqrt(size_vec(*dir));
	if(fabs(size - 1.0) > EPSILON)
		error_message(GENERAL_ERROR, "Direction not normalized");
#endif

	sub_vec(vec, *pnt, *point);
	product = dot_product(*dir, vec);
	
	projection -> x = point -> x + product * dir -> x;
	projection -> y = point -> y + product * dir -> y;

	dist = dist_point(*pnt, *projection);
	return(sqrt(dist));
}



/* ellipse greater/smaller by positive/negative eps */

static logic
check_point_inside_ellipse(point_rec *point, ellipse_rec *ellipse, double eps, double *value)
{
	point_rec pnt;
	trans_matrix trans;
	double x_rate, y_rate;

	copy_vec(trans.x, ellipse -> major);
	copy_vec(trans.y, ellipse -> minor);

	transform_point_from_global_to_local(point, &pnt, &(ellipse -> center), &trans);

	x_rate = pnt.x / (ellipse -> max + eps);
	y_rate = pnt.y / (ellipse -> min + eps);

	*value = x_rate * x_rate + y_rate * y_rate - 1.0;
	if(*value <= 0.0)return(YES);

	return(NO);
}


static int
get_ellipse_line_intersection(ellipse_rec *ellipse, point_rec *point, vector_rec *dir, point_rec *intersection)
{
	point_rec pnt, pnt1, pnt2;
	vector_rec direction;
	trans_matrix trans;
	double a0, a1, a2, a0x, a0y, a1x, a1y, a2x, a2y;
	double D, max, min, t, t1, t2, norm;

	norm = size_vec(*dir);
	if(norm < EPSILON * EPSILON)error_message(GENERAL_ERROR, "Zero length direction");

	copy_vec(trans.x, ellipse -> major);
	copy_vec(trans.y, ellipse -> minor);

	transform_vector_from_global_to_local(dir, &direction, &trans);
	transform_point_from_global_to_local(point, &pnt, &(ellipse -> center), &trans);

	max = ellipse -> max;
	min = ellipse -> min;

	a2x = direction.x / max;
	a2y = direction.y / min;
	a2 = a2x * a2x + a2y * a2y;

	a0x = pnt.x / max;
	a0y = pnt.y / min;
	a0 = a0x * a0x + a0y *a0y - 1.0;

	a1x = a2x * a0x;
	a1y = a2y * a0y;
	a1 = a1x + a1y;

	D = a1 * a1 - a0 * a2;
	
	if(D < 0.0)return(0);

	invert_rotation_transformation(&trans);

	if(D == 0.0){
		t = -a1 / a2;

		pnt.x += t * direction.x;
		pnt.y += t * direction.y;

		transform_point_from_local_to_global(&pnt, &intersection[0], &(ellipse -> center), &trans);
		return(1);
	}

	D = sqrt(D);

	t1 = -(a1 - D) / a2;
	t2 = -(a1 + D) / a2;

	pnt1.x = pnt.x + t1 * direction.x;
	pnt1.y = pnt.y + t1 * direction.y;

	transform_point_from_local_to_global(&pnt1, &intersection[0], &(ellipse -> center), &trans);

	pnt2.x = pnt.x + t2 * direction.x;
	pnt2.y = pnt.y + t2 * direction.y;

	transform_point_from_local_to_global(&pnt2, &intersection[1], &(ellipse -> center), &trans);

	return(2);
}


static void
get_ellipse_line_center_intersection(ellipse_rec *ellipse, vector_rec *dir, point_rec *pnt1, point_rec *pnt2)
{
	vector_rec direction;
	trans_matrix trans;
	double a, b, t, norm;

	norm = size_vec(*dir);
	if(norm < EPSILON * EPSILON)error_message(GENERAL_ERROR, "Zero length direction");

	copy_vec(trans.x, ellipse -> major);
	copy_vec(trans.y, ellipse -> minor);

	transform_vector_from_global_to_local(dir, &direction, &trans);

	a = direction.x / (ellipse -> max);
	b = direction.y / (ellipse -> min);

	t = sqrt(1.0 / (a * a + b * b));
	
	invert_rotation_transformation(&trans);

	mul_vec(direction, t);
	transform_point_from_local_to_global(&direction, pnt1, &(ellipse -> center), &trans);

	mul_vec(direction, -1.0);
	transform_point_from_local_to_global(&direction, pnt2, &(ellipse -> center), &trans);
}


#ifdef PARAMETRIC_ROUTINES

static int
get_ellipse_ellipse_intersection(ellipse_rec *ellipse1, ellipse_rec *ellipse2, point_rec *intersection)
{
	point_rec point, pnt;
	trans_matrix trans;
	double product, t, tt, dt, ddt, delta_t = EPSILON_T;
	double prev_t, next_t, prev_val, next_val, tn, tp, tpp, tnn, expand = 1.2;
	double max1, max2, max, dist, val, value, value_p, value_n, value_pp, value_nn;
	int q, prev_q, next_q, count, qq, qp, qn, qpp, qnn;
	logic found, last;

	product = dot_product(ellipse1 -> normal, ellipse2 -> normal);
	if(fabs(fabs(product) - 1.0) > EPSILON)error_message(GENERAL_ERROR, "Ellipses are not coplanar"); 

	copy_vec(trans.x, ellipse1 -> major);
	copy_vec(trans.y, ellipse1 -> minor);
	copy_vec(trans.z, ellipse1 -> normal);

	transform_point_from_global_to_local(&(ellipse2 -> center), &point, &(ellipse1 -> center), &trans);
	if(fabs(point.z) > EPSILON)error_message(GENERAL_ERROR, "Ellipses are not in the same plane"); 

	max1 = ellipse1 -> max;
	max2 = ellipse2 -> max;
	max = max1 + max2;

/* intersection of circumsribed circles */
	dist = dist_point(ellipse1 -> center, ellipse2 -> center);
	if(dist > max * max)return(0);

/* ellipse1 and ellipse2 are affine */
	if(fabs(point.x) < EPSILON && fabs(point.y) < EPSILON){
		product = dot_product(ellipse1 -> major, ellipse2 -> minor);
		if(fabs(product) < EPSILON){
			if(fabs(max1 / ellipse1 -> min - max2 / ellipse2 -> min) < EPSILON){
				if(fabs(max1 / max2 - 1.0) < EPSILON)return(-1);   /* coinciding ellipses */
				return(0);
			}
		}
	}

/* intersection of circle circumscribed to ellipse2 with ellipse1 */
	if(check_point_inside_ellipse(&(ellipse2 -> center), ellipse1, &val) == YES){
		if(fabs(point.x) > EPSILON && fabs(point.y) > EPSILON){
			project_point_to_ellipse_t_quadrant(&(ellipse2 -> center), ellipse1, &pnt, &t, NO);
			dist = dist_point(ellipse2 -> center, pnt);
			if(dist > max2 * max2)return(0);
		}
	}

/* intersection of circle circumscribed to ellipse1 with ellipse2 */
	if(check_point_inside_ellipse(&(ellipse1 -> center), ellipse2, &val) == YES){
		copy_vec(trans.x, ellipse2 -> major);
		copy_vec(trans.y, ellipse2 -> minor);
		copy_vec(trans.z, ellipse2 -> normal);

		transform_point_from_global_to_local(&(ellipse1 -> center), &point, &(ellipse2 -> center), &trans);

		if(fabs(point.x) > EPSILON && fabs(point.y) > EPSILON){
			project_point_to_ellipse_t_quadrant(&(ellipse1 -> center), ellipse2, &pnt, &t, NO);
			dist = dist_point(ellipse1 -> center, pnt);
			if(dist > max1 * max1)return(0);
		}
	}

	dt = 1.0 / 10;

	prev_q = 4;
	prev_t = dt / 2.0;

	get_ellipse_point_t_quadrant(ellipse1, prev_q, prev_t, &pnt);
	check_point_inside_ellipse(&pnt, ellipse2, &prev_val);
	while(fabs(prev_val) < EPSILON){
		prev_t -= delta_t;
		get_ellipse_point_t_quadrant(ellipse1, prev_q, prev_t, &pnt);
		check_point_inside_ellipse(&pnt, ellipse2, &prev_val);
	}

	next_q = q = 1;
	next_t = t = dt / 2.0;

	get_ellipse_point_t_quadrant(ellipse1, q, t, &pnt);
	check_point_inside_ellipse(&pnt, ellipse2, &val);
	while(fabs(val) < EPSILON){
		t += delta_t;
		get_ellipse_point_t_quadrant(ellipse1, q, t, &pnt);
		check_point_inside_ellipse(&pnt, ellipse2, &val);
	}

	count = 0;
	last = NO;
	while(YES){
		if(next_q == 1 || next_q == 3){
			next_t += dt;
			if(next_t > 1.0){
				++next_q;
				next_t = 2.0 - next_t;
			}
		}
		else{
			next_t -= dt;
			if(next_t < 0.0){
				if(++next_q == 5){
					next_q = 1;
					last = YES;
				}
				next_t = -next_t;
			}
		}

		get_ellipse_point_t_quadrant(ellipse1, next_q, next_t, &pnt);
		check_point_inside_ellipse(&pnt, ellipse2, &next_val);
		while(fabs(next_val) < EPSILON){
			
/* note: in(de)crementing by delta_t I can never cross quadrant boundary because I did not
	 start at the beginning of quadrant and because there may be at most four intersections */

			if(next_q == 1 || next_q == 3)
				next_t += delta_t;
			else
				next_t -= delta_t;
			get_ellipse_point_t_quadrant(ellipse1, next_q, next_t, &pnt);
			check_point_inside_ellipse(&pnt, ellipse2, &next_val);
		}

		if(next_val * val < 0.0){

/* intersection in the last interval */

			get_ellipse_ellipse_interval_intersection(ellipse1, ellipse2, q, t, val, dt, &intersection[count++]);
		}
		else{

/* note: check the equality only in the prev interval */
			if((prev_val <= val && next_val < val) || (prev_val >= val && next_val > val)){
				if(prev_val * val > 0.0){
					if(val + 1.0 < expand && val + 1.0 > 1.0 / expand){
				
/* extreme in the last two intervals */

						tt = t;
						qq = q;
						value = val;

						tpp = prev_t;
						qpp = prev_q;

						tnn = next_t;
						qnn = next_q;

						value_pp = prev_val;
						value_nn = next_val;
						
						ddt = dt / 2;
						
						while(YES){
							tn = tp = tt;
							qn = qp = qq;

							if(qq == 1 || qq == 3){
								tn += ddt;
								if(tn > 1.0){
									++qn;
									tn = 2.0 - tn;
								}
							}
							else{
								tn -= ddt;
								if(tn < 0.0){
									if(++qn == 5)qn = 1;
									tn = -tn;
								}
							}
							
							if(q == 2 || q == 4){
								tp += ddt;
								if(tp > 1.0){
									--qp;
									tp = 2.0 - tp;
								}
							}
							else{
								tp -= ddt;
								if(tp < 0.0){
									if(--qp == 0)qp = 4;
									tp = -tp;
								}
							}
							
							get_ellipse_point_t_quadrant(ellipse1, qp, tp, &pnt);
							check_point_inside_ellipse(&pnt, ellipse2, &value_p);
							while(fabs(value_p) < EPSILON){
								if(qp == 1 || qp == 3){
									tp += delta_t;
									if(tp > 1.0){
										++qp;
										tp = 2.0 - tp;
									}
								}
								else{
									tp -= delta_t;
									if(tp < 0.0){
										if(++qp == 5)qp = 1;
										tp = -tp;
									}
								}
								ddt += 2.0 * delta_t;
								get_ellipse_point_t_quadrant(ellipse1, qp, tp, &pnt);
								check_point_inside_ellipse(&pnt, ellipse2, &value_p);
							}
							
							get_ellipse_point_t_quadrant(ellipse1, qn, tn, &pnt);
							check_point_inside_ellipse(&pnt, ellipse2, &value_n);
							while(fabs(value_n) < EPSILON){
								if(qn == 1 || qn == 3){
									tn += delta_t;
									if(tn > 1.0){
										++qn;
										tn = 2.0 - tn;
									}
								}
								else{
									tn -= delta_t;
									if(tn < 0.0){
										if(++qn == 5)qn = 1;
										tn = -tn;
									}
								}
								ddt += 2.0 * delta_t;
								get_ellipse_point_t_quadrant(ellipse1, qn, tn, &pnt);
								check_point_inside_ellipse(&pnt, ellipse2, &value_n);
							}
							
							found = NO;
							if(value_pp * value_p < 0.0){
								found = YES;
								get_ellipse_ellipse_interval_intersection(ellipse1, ellipse2, qpp, tpp, value_pp, ddt, &intersection[count++]);
							}
							if(value * value_p < 0.0){
								found = YES;
								get_ellipse_ellipse_interval_intersection(ellipse1, ellipse2, qp, tp, value_p, ddt, &intersection[count++]);
							}
							if(value * value_n < 0.0){
								found = YES;
								get_ellipse_ellipse_interval_intersection(ellipse1, ellipse2, qq, tt, value, ddt, &intersection[count++]);
							}
							if(value_nn * value_n < 0.0){
								found = YES;
								get_ellipse_ellipse_interval_intersection(ellipse1, ellipse2, qn, tn, value_n, ddt, &intersection[count++]);
							}
							
							if(found == YES)break;
							
							ddt /= 2.0;
							if(ddt < delta_t)break;

							if((value <= value_p && value_pp <= value_p) || (value >= value_p && value_pp >= value_p)){
								tnn = tt;
								qnn = qq;
								value_nn = value;
								tt = tp;
								qq = qp;
								value = value_p;
								continue;
							}
							if((value <= value_n && value_nn <= value_n) || (value >= value_n && value_nn >= value_n)){
								tpp = tt;
								qpp = qq;
								value_pp = value;
								tt = tn;
								qq = qn;
								value = value_n;
								continue;
							}
							tpp = tp;
							qpp = qp;
							value_pp = value_p;
							tnn = tn;
							qnn = qn;
							value_nn = value_n;
						}
					}				
				}
			}
		}

		if(last == YES)break;

		prev_q = q;
		prev_t = t;
		prev_val = val;

		q = next_q;
		t = next_t;
		val = next_val;
	}

	return(count);
}


static void
get_ellipse_ellipse_interval_intersection(ellipse_rec *ellipse1, ellipse_rec *ellipse2, int q, double t, double val, double dt,
																					point_rec *point)
{
	point_rec pnt;
	int qq, qqq;
	double tt, ttt, value, delta_t = EPSILON_T;

	qq = qqq = q;
	tt = ttt = t;

	dt /= 2.0;
	while(YES){
		if(qq == 1 || qq == 3){
			tt += dt;
			if(tt > 1.0){
				++qq;
				tt = 2.0 - tt;
			}
		}
		else{
			tt -= dt;
			if(tt < 0.0){
				if(++qq == 5)qq = 1;
				tt = -tt;
			}
		}
		
		get_ellipse_point_t_quadrant(ellipse1, qq, tt, &pnt);
		check_point_inside_ellipse(&pnt, ellipse2, &value);
		
		if(fabs(value) < EPSILON || dt < delta_t){
			if(fabs(value) > EPSILON){
				if(qq == qqq){
					tt = ttt + (tt - ttt) * val / (val - value);
					if(tt >= 0.0 && tt <= 1.0){
						get_ellipse_point_t_quadrant(ellipse1, qq, tt, &pnt);
						check_point_inside_ellipse(&pnt, ellipse2, &value);
					}
				}
			}
						
			copy_vec(*point, pnt);
			return;
		}
					
		if(value * val > 0.0){
			qqq = qq;
			ttt = tt;
			val = value;
		}
		else{
			qq = qqq;
			tt = ttt;
			dt /= 2.0;
		}
	}
}

#endif

static void
project_point_to_ellipse(point_rec *point, ellipse_rec *ellipse, point_rec *pnt)
{
	point_rec p;
	trans_matrix trans;
	double max, min, max2, min2, max4, min4, xp2, yp2;
	double val, rate, k, dk = 1.0, fk, dfk;
	double a, b, c, d, e;
	double px, py, gx, gy, dx, dy;
	logic skip = NO;
	int iter = 0;
	
	copy_vec(trans.x, ellipse -> major);
	copy_vec(trans.y, ellipse -> minor);

	transform_point_from_global_to_local(point, &p, &(ellipse -> center), &trans);

	xp2 = p.x * p.x;
	yp2 = p.y * p.y;

	max = ellipse -> max;
	min = ellipse -> min;

#ifdef DEBUG
	if(max < min)error_message(GENERAL_ERROR, "Unexpected situation");		
#endif

	max2 = max * max;
	min2 = min * min;

	max4 = max2 * max2;
	min4 = min2 * min2;

	a = 16.0;
	b = 16.0 * (max2 + min2);
	c = 4.0 * (4.0 * max2 * min2 + max4 + min4 - xp2 * max2 - yp2 * min2);
	d = 4.0 * max2 * min2 * (max2 + min2 - xp2 - yp2);
	e = max4 * min4 - max2 * min2 * (xp2 * min2 + yp2 * max2);

/* get radial projection */

	val = xp2 / max2 + yp2 / min2;

	if(val < 1.0e-10){       /* point is in center */
		rate = 0.0;
		k = -min2 / 2.0;
		skip = YES;
	}
	else{
		rate = sqrt(1.0 / val);
		if(rate <= 1.0){       /* point is outside ellipse */
			px = p.x * rate;
			py = p.y * rate;

			gx = 2.0 * px / max2;
			gy = 2.0 * py / min2;

			dx = px - p.x;
			dy = py - p.y;

			k = sqrt((dx * dx + dy * dy) / (gx * gx + gy * gy));
		}
		else{                  /* point is inside ellipse */
			k = 0.0;

			if(p.y == 0.0){      /* point is on major axis */
				skip = YES;
				if(fabs(p.x) <= max * (1.0 - min2 / max2))    /* point is in outer half of minor osculating circle */
					k = -min2 / 2.0;
				else               /* point is between ellipse center and center of minor osculating circle */
					k = -max * (max - fabs(p.x)) / 2.0; 
			}
			else{
				if(p.x == 0.0){    /* point on minor axis */
					skip = YES;
					k = -min * (min - fabs(p.y)) / 2.0; 
				}
			}
		}
	}

	if(skip == NO){
		dk = 1.0;
		while(fabs(dk) > 1.0e-10){
			iter++;
			fk = a * k * k * k * k + b * k * k * k + c * k * k + d * k + e;
			dfk = 4.0 * a * k * k * k + 3.0 * b * k * k + 2.0 * c * k + d;
			dk = fk / dfk;
			k -= dk;

/* this check can cause failure of get_ellipse_line_square_distance */

			if(iter == MAX_ITER)error_message(GENERAL_ERROR, "Projection to ellipse does not converge");
		}
	}

	if(rate <= 1.0){
		if(rate == 0.0){
			p.x = 0.0;
			p.y = ellipse -> min;
		}
		else{
			p.x *= max2 / (max2 + 2.0 * k);
			p.y *= min2 / (min2 + 2.0 * k);
		}
	}
	else{
		p.x *= max2 / (max2 + 2.0 * k);
		if(p.y != 0.0)
			p.y *= min2 / (min2 + 2.0 * k);
		else if(p.x > max)
			p.y = 0.0;
		else
			p.y = sqrt((1.0 - p.x * p.x / max2) * min2);
	}

	invert_rotation_transformation(&trans);
	transform_point_from_local_to_global(&p, pnt, &(ellipse -> center), &trans);
}


#define NORMALIZE_LOCALLY

static void
get_ellipse_normal_point(ellipse_rec *ellipse, point_rec *point, vector_rec *normal)
{
	point_rec pnt;
	vector_rec nor;
	trans_matrix trans;
	double size;

	copy_vec(trans.x, ellipse -> major);
	copy_vec(trans.y, ellipse -> minor);

	transform_point_from_global_to_local(point, &pnt, &(ellipse -> center), &trans);

	nor.x = 2.0 * pnt.x / (ellipse -> max * ellipse -> max);
	nor.y = 2.0 * pnt.y / (ellipse -> min * ellipse -> min);

#ifdef NORMALIZE_LOCALLY
	size = 1.0 / sqrt(nor.x * nor.x + nor.y * nor.y);
	nor.x *= size;
	nor.y *= size;
#endif

	invert_rotation_transformation(&trans);

	transform_vector_from_local_to_global(&nor, normal, &trans);

#ifndef NORMALIZE_LOCALLY
	size = sqrt(size_vec(*normal));
	div_vec(*normal, size);
#endif
}


/* points point1 and point2 are expected to lie on the ellipse;
	 I do not compute control point as it may be infinite if weight is zero !!! */

static double
get_quadratic_elliptic_arc_weight(ellipse_rec *ellipse, point_rec *point1, point_rec *point2)
{
	point_rec pnt1, pnt2;
	double x1, y1, x2, y2, val1, val2, sz1, sz2, scale_y_e2c, weight, cosine, Radius, radius;
	trans_matrix trans_g2l; 

	Radius = ellipse -> max;
	radius = ellipse -> min;

	copy_vec(trans_g2l.x, ellipse -> major);
	copy_vec(trans_g2l.y, ellipse -> minor);
				
	transform_point_from_global_to_local(point1, &pnt1, &(ellipse -> center), &trans_g2l);
	transform_point_from_global_to_local(point2, &pnt2, &(ellipse -> center), &trans_g2l);

	x1 = pnt1.x / Radius;
	y1 = pnt1.y / radius;

	x2 = pnt2.x / Radius;
	y2 = pnt2.y / radius;

	val1 = x1 * x1 + y1 * y1;
	val2 = x2 * x2 + y2 * y2;
	
	if(fabs(1.0 - val1) > 1.0e-10 || fabs(1.0 - val2) > 1.0e-10)error_message(GENERAL_ERROR, "Unexpected situation");

/* get radial projection of points pnt1 and pnt2 */

	val1 = sqrt(1.0 / val1);
	val2 = sqrt(1.0 / val2);

	scale_y_e2c = Radius / radius;

	pnt1.x *= val1;
	pnt2.x *= val2;

	pnt1.y *= val1 * scale_y_e2c;
	pnt2.y *= val2 * scale_y_e2c;

	sz1 = size_vec(pnt1);
	sz2 = size_vec(pnt2);

	cosine = dot_product(pnt1, pnt2) / sqrt(sz1 * sz2);

	if(cosine >= 1.0)return(1.0);
	if(cosine <= -1.0)return(0.0);

	weight = sqrt((1.0 + cosine) / 2.0);

	return(weight);
}
	
	
#ifdef PARAMETRIC_ROUTINES

/* parameter spec is used to allow for handling the case when the point is on major axis 
	 between centers of minor osculating circles;
	 y-coordinate of the point is slightly perturbated */

static int
project_point_to_ellipse_t_quadrant(point_rec *point, ellipse_rec *ellipse, point_rec *pnt, double *t_par, logic spec)
{
	point_rec point_spec, pnt_spec, p;
	vector_rec grad, vec;
	trans_matrix trans;
	double xx, yy, dist, norm, rate1 = 0.25, rate2 = 1.5, delta_t = EPSILON_T;
	double t = 0.5, dt, dtt, ddt, d_t = 0.0, max_dt = 0.25, rate;
	int sign = 0, sign_x = 0, sign_y = 0, iter = 0, quadrant = 0;

	copy_vec(trans.x, ellipse -> major);
	copy_vec(trans.y, ellipse -> minor);

/* move point to the first quadrant */
	transform_point_from_global_to_local(point, &p, &(ellipse -> center), &trans);

	xx = p.x;
	yy = p.y;

	if(xx == 0.0 && yy == 0.0){
		if(spec == NO)error_message(GENERAL_ERROR, "Point is in center of ellipse");
		yy = ellipse -> min / 1.0e+10;
	}
	if(ellipse -> max != ellipse -> min){
		if(yy == 0.0){
/* point on major axis between center of ellipse and center of minor osculating circle cannot be uniquelly projected */
			rate = ellipse -> min / ellipse -> max;
			if(fabs(xx) < ellipse -> max * (1.0 - rate * rate)){
				if(spec == NO)error_message(GENERAL_ERROR, "Point lies on major axis of ellipse");
				yy = ellipse -> min / 1.0e+10;
			}
		}
	}

	p.x = fabs(xx);
	p.y = fabs(yy);

	if(xx != 0.0)sign_x = (xx > 0.0) ? 1 : -1;
	if(yy != 0.0)sign_y = (yy > 0.0) ? 1 : -1;

	if(sign_x > 0){
		if(sign_y > 0)
			quadrant = 1;
		else
			quadrant = 4;
	}
	else{
		if(sign_y > 0)
			quadrant = 2;
		else
			quadrant = 3;
	}

	if(sign_y == 0)sign = sign_x;
	if(sign_x == 0)sign = sign_y;

	if(sign != 0){
		if(sign > 0)
			quadrant = 1;
		else
			quadrant = 3;
	}

	invert_rotation_transformation(&trans);
	transform_point_from_local_to_global(&p, &point_spec, &(ellipse -> center), &trans);

/* make projection (in the first quadrant) */
	while(YES){
		get_ellipse_point_t(ellipse, t, &pnt_spec);
		norm = sqrt(get_ellipse_tangent_t(ellipse, t, &grad));
		sub_vec(vec, point_spec, pnt_spec);
		dist = dot_product(vec, grad) / norm;

		dt = dist / norm;
		if(fabs(dt) < delta_t)break;

		if((ddt = fabs(dt)) > max_dt)dt /= ddt / max_dt;

		if(iter != 0){
			if(dt * d_t < 0.0){
				if(fabs(dt) > rate1 * fabs(d_t))dt = -d_t * rate1;
			}
			else{
				if(fabs(dt) > rate2 * fabs(d_t))dt = d_t * rate2;
			}
		}
		
		dtt = d_t = dt;
				
		if(dt < 0.0){
			if(t + dt < 0.0)dtt = (0.0 - t) / 2.0;
		}
		if(dt > 0.0){
			if(t + dt > 1.0)dtt = (1.0 - t) / 2.0;
		}

		iter++;
		if(iter == MAX_ITER)
			error_message(GENERAL_ERROR, "Projection to ellipse does not converge");

		t += (d_t = dtt);
	}

/* move projection to original quadrant */
	invert_rotation_transformation(&trans);
	transform_point_from_global_to_local(&pnt_spec, &p, &(ellipse -> center), &trans);
	p.x *= sign_x;
	p.y *= sign_y;
	invert_rotation_transformation(&trans);
	transform_point_from_local_to_global(&p, pnt, &(ellipse -> center), &trans);

/* note: t is continuous (but not monotonic)
         t = 0 on major axis
				 t = 1 on minor axis
	 to prevent confusion, quadrant is returned
         x+,y+   1
				 x-,y+   2
				 x-,y-   3
				 x+,y-   4
   zeros are included into quadrants 1 and 3 */

	*t_par = t;
	return(quadrant);
}


static void
get_ellipse_point_t_quadrant(ellipse_rec *ellipse, int quadrant, double t, point_rec *pnt)
{
	point_rec p;
	trans_matrix trans;

	get_ellipse_point_t(ellipse, t, pnt);

/* move point to quadrant */
	if(quadrant != 1){
		copy_vec(trans.x, ellipse -> major);
		copy_vec(trans.y, ellipse -> minor);

		transform_point_from_global_to_local(pnt, &p, &(ellipse -> center), &trans);
		if(quadrant == 2 || quadrant == 3)p.x *= -1.0;
		if(quadrant == 4 || quadrant == 3)p.y *= -1.0;

		invert_rotation_transformation(&trans);
		transform_point_from_local_to_global(&p, pnt, &(ellipse -> center), &trans);
	}
}


static double
get_ellipse_tangent_t_quadrant(ellipse_rec *ellipse, int quadrant, double t, vector_rec *grad)
{
	vector_rec g;
	trans_matrix trans;
	double norm;

	norm = get_ellipse_tangent_t(ellipse, t, grad);

/* move grad to quadrant */
	if(quadrant != 1){
		copy_vec(trans.x, ellipse -> major);
		copy_vec(trans.y, ellipse -> minor);

		transform_vector_from_global_to_local(grad, &g, &trans);
		if(quadrant == 2 || quadrant == 3)g.x *= -1.0;
		if(quadrant == 4 || quadrant == 3)g.y *= -1.0;

		invert_rotation_transformation(&trans);
		transform_vector_from_local_to_global(&g, grad, &trans);
	}

	return(norm);
}


static void
get_ellipse_normal_t_quadrant(ellipse_rec *ellipse, int quadrant, double t, vector_rec *normal)
{
	vector_rec n;
	trans_matrix trans;

	get_ellipse_normal_t(ellipse, t, normal);

/* move normal to quadrant */
	if(quadrant != 1){
		copy_vec(trans.x, ellipse -> major);
		copy_vec(trans.y, ellipse -> minor);

		transform_vector_from_global_to_local(normal, &n, &trans);
		if(quadrant == 2 || quadrant == 3)n.x *= -1.0;
		if(quadrant == 4 || quadrant == 3)n.y *= -1.0;

		invert_rotation_transformation(&trans);
		transform_vector_from_local_to_global(&n, normal, &trans);
	}
}


static void
get_ellipse_point_t(ellipse_rec *ellipse, double t, point_rec *pnt)
{
	double weight = SQRT1_2;
	double B0, B1, B2, Bx, By;
	double sum_B;

	B0 = B_0(t);
	B1 = B_1(t) * weight;
	B2 = B_2(t);

	sum_B = B0 + B1 + B2;

	Bx = B0 * ellipse -> pnt0.x + B1 * ellipse -> pnt1.x + B2 * ellipse -> pnt2.x;
	By = B0 * ellipse -> pnt0.y + B1 * ellipse -> pnt1.y + B2 * ellipse -> pnt2.y;
	
	pnt -> x = Bx / sum_B;
	pnt -> y = By / sum_B;
}


static double
get_ellipse_tangent_t(ellipse_rec *ellipse, double t, vector_rec *grad)
{
	double weight = SQRT1_2;
	double B0, B1, B2, dB0, dB1, dB2;
	double sum_B, sum_dB, sum_B2, Bx, By, dBx, dBy;

	B0 = B_0(t);
	B1 = B_1(t) * weight;
	B2 = B_2(t);

	dB0 = dB_0(t);
	dB1 = dB_1(t) * weight;
	dB2 = dB_2(t);

	sum_B = B0 + B1 + B2;
	sum_dB = dB0 + dB1 + dB2;

	sum_B2 = sum_B * sum_B;

	Bx = B0 * ellipse -> pnt0.x + B1 * ellipse -> pnt1.x + B2 * ellipse -> pnt2.x;
	By = B0 * ellipse -> pnt0.y + B1 * ellipse -> pnt1.y + B2 * ellipse -> pnt2.y;

	dBx = dB0 * ellipse -> pnt0.x + dB1 * ellipse -> pnt1.x + dB2 * ellipse -> pnt2.x;
	dBy = dB0 * ellipse -> pnt0.y + dB1 * ellipse -> pnt1.y + dB2 * ellipse -> pnt2.y;

	grad -> x = (dBx * sum_B - Bx * sum_dB) / sum_B2;
	grad -> y = (dBy * sum_B - By * sum_dB) / sum_B2;

	return(size_vec(*grad));
}


static void
get_ellipse_normal_t(ellipse_rec *ellipse, double t, vector_rec *normal)
{
	vector_rec grad;
	double norm;

	norm = get_ellipse_tangent_t(ellipse, t, &grad);
	norm = sqrt(norm);
	div_vec(grad, norm);

	cross_product(*normal, grad, ellipse -> normal);
}

#endif


/* delta_t */

static logic
check_point_on_ellipse_interval(int q, double t, int q1, double t1, int q2, double t2)
{
	double delta_t = EPSILON_T;

	if(q <= 0 || q > 4 || q1 <= 0 || q1 > 4 || q2 <= 0 || q2 > 4)error_message(GENERAL_ERROR, "Ivalid quadrant");

	if(q != q1 && q != q2){
		if(q1 == q2){
			if(q1 == 1 || q1 == 3){
				if(t1 < t2)return(NO);
			}
			else{
				if(t1 > t2)return(NO);
			}
		}
		else{
			if(q1 < q2){
				if(q < q1 || q > q2)return(NO);
			}
			else{
				if(q < q1){
					if(q > q2)return(NO);
				}
				if(q > q2){
					if(q < q1)return(NO);
				}
			}
		}

		return(YES);
	}

	if(q == q1){
		if(q == 1 || q == 3){
			if(t < t1 - delta_t)return(NO);
		}
		else{
			if(t > t1 + delta_t)return(NO);
		}
	}
	if(q == q2){
		if(q == 1 || q == 3){
			if(t > t2 + delta_t)return(NO);
		}
		else{
			if(t < t2 - delta_t)return(NO);
		}
	}

	return(YES);
}



static void
transform_vector_from_global_to_local(vector_rec *global, vector_rec *local, trans_matrix *trans_g2l)
{
	local -> x = dot_product(*global, trans_g2l -> x);
	local -> y = dot_product(*global, trans_g2l -> y);
}


static void
transform_vector_from_local_to_global(vector_rec *local, vector_rec *global, trans_matrix *trans_l2g)
{
	global -> x = dot_product(*local, trans_l2g -> x);
	global -> y = dot_product(*local, trans_l2g -> y);
}


static void
transform_point_from_global_to_local(point_rec *global, point_rec *local, point_rec *origin, trans_matrix *trans_g2l)
{
	vector_rec delta;

	delta.x = global -> x - origin -> x;
	delta.y = global -> y - origin -> y;

	local -> x = dot_product(delta, trans_g2l -> x);
	local -> y = dot_product(delta, trans_g2l -> y);
}


static void
transform_point_from_local_to_global(point_rec *local, point_rec *global, point_rec *origin, trans_matrix *trans_l2g)
{
	global -> x = dot_product(*local, trans_l2g -> x) + origin -> x;
	global -> y = dot_product(*local, trans_l2g -> y) + origin -> y;
}


static void
invert_rotation_transformation(trans_matrix *trans)
{
	double tmp;

	swap(trans -> x.y, trans -> y.x, tmp);
}

static void
get_inverted_rotation_transformation(trans_matrix *trans_in, trans_matrix *trans_out)
{
	trans_out -> x.x = trans_in -> x.x;
	trans_out -> x.y = trans_in -> y.x;
	trans_out -> y.x = trans_in -> x.y;
	trans_out -> y.y = trans_in -> y.y;
}


static void
transform_local_from_ellipse_to_circle(point_rec *elliptical, point_rec *circular, vector_rec *trans_e2c)
{
	circular -> x = elliptical -> x * trans_e2c -> x;
	circular -> y = elliptical -> y * trans_e2c -> y;
}


static void
transform_local_from_circle_to_ellipse(point_rec *circular, point_rec *elliptical, vector_rec *trans_c2e)
{
	elliptical -> x = circular -> x * trans_c2e -> x;
	elliptical -> y = circular -> y * trans_c2e -> y;
}


static void
invert_scale_transformation(vector_rec *trans)
{
	trans -> x = 1.0 / trans -> x;
	trans -> y = 1.0 / trans -> y;
}

static void
get_inverted_scale_transformation(vector_rec *trans_in, vector_rec *trans_out)
{
	trans_out -> x = 1.0 / trans_in -> x;
	trans_out -> y = 1.0 / trans_in -> y;
}	



static char *
get_next_record_file(FILE *file, char *file_name, char *err_msg)
{
	char *buf = NULL;

	if((buf = fgets(line_buffer, BUFFER_SIZE, file)) == NULL){
		if(!feof(file))error_message(FILE_READ_ERROR, "File %s reading error", file_name);
		if(err_msg == NULL)return(NULL);
		error_message(INPUT_ERROR, "%s", err_msg);
	}
	return(buf);
}


static char *
get_next_relevant_record_file(FILE *file, char *file_name, char *err_msg)
{
	char *buf = NULL, *character = NULL;
	
	do{
		if((buf = get_next_record_file(file, file_name, err_msg)) == NULL)return(NULL);
		character = line_buffer;
		while(*character == SPACE)character++;
	}while(*character == NEWLINE || *character == CR || *character == REMARK);
	return(buf);
}


static void
write_record_file(FILE *file, char *file_name, char *format, ...)
{
	char buffer[1024];
	va_list args;

	va_start(args, format);
	vsprintf(buffer, format, args);
	va_end(args);

	if(fprintf(file, "%s", buffer) < 0)
		error_message(FILE_WRITE_ERROR, "File %s writing error", file_name);
}


static char *
get_next_record(char *err_msg)
{
	char *buf = NULL;

	if((buf = fgets(line_buffer, BUFFER_SIZE, geom_file)) == NULL){
		if(!feof(geom_file))error_message(FILE_READ_ERROR, "File %s reading error", geom_file_name);
		if(err_msg == NULL)return(NULL);
		error_message(INPUT_ERROR, "%s", err_msg);
	}
	return(buf);
}


static char *
get_next_relevant_record(char *err_msg)
{
	char *buf = NULL, *character = NULL;
	
	do{
		if((buf = get_next_record(err_msg)) == NULL)return(NULL);
		character = line_buffer;
		while(*character == SPACE)character++;
	}while(*character == NEWLINE || *character == CR || *character == REMARK);
	return(buf);
}


static void
write_record(char *format, ...)
{
	char buffer[1024];
	va_list args;

	va_start(args, format);
	vsprintf(buffer, format, args);
	va_end(args);

	if(fprintf(t3d_in_file, "%s", buffer) < 0)
		error_message(FILE_WRITE_ERROR, "File %s writing error", t3d_in_file_name);
}


static void
error_message(int code, char *format, ...)
{
	char buffer[1024];
	va_list args;
	static logic first = YES;

	va_start(args, format);
	vsprintf(buffer, format, args);
	va_end(args);

	if(first == YES){
		fprintf(stderr, "\n");
		first = NO;
	}

	if(code == NOTIFICATION)
		fprintf(stderr, "Notification: %s\n", buffer);
	else if(code == WARNING)
		fprintf(stderr, "Warning: %s\n", buffer);
	else if(code == ALARM)
		fprintf(stderr, "Alarm: %s\n", buffer);
	else{
		fprintf(stderr, "\nError: %s\n\n", buffer);

#ifdef ELIXIR
		if(draw == YES){
			if(ElixirInitialized() == YES){
				strcat(buffer, "\n\n");
				ESIEventLoop(YES, buffer);
			}
		}
#endif

		if(t3d_in_file != NULL)fclose(t3d_in_file);
		if(geom_file != NULL)fclose(geom_file);

/* function is called rather then performing exit in order to allow for easy setting of break point (when exiting)
	 which is not dependent on particular line in the code */

		error_exit(code);
	}
}


static void
error_exit(int exit_code)
{
	exit(exit_code);
}


/* solution of A.X=Y;
	 matrix A is full;
	 A(n,n),X(n,m),Y(n,m) */

static logic
solve_full_system(double *a, double *x, double *y, long n, long m)
{
  long i, j, k, ac, acr, acc = 0, aca, aca1, acx, acy, acy1, aci, acj;
  long *av;
  double s, g;
  
  av = (long *)calloc(n, sizeof(long));
  
/* set vector of unknown ordering */

  for(i = 0; i < n; i++)av[i] = i;
  
  for(i = 0; i < n; i++){
    acr = i;  
		acc = i;

/* search for pivot */

		s = 0.0;
		for(j = i; j < n; j++){       /* loop over rows */
			aca = j * n + i;
			for(k = i; k < n; k++){     /* loop over columns */
				if(s < fabs(a[aca])){
					s = fabs(a[aca]);  
					acr = j;  
					acc = k;
				}
				aca++;
			}
		}
		if(s < 1.0e-10)return(NO);
    
/* swap rows */

    if(acr != i){
      aca = i * n + i;  
			aca1 = acr * n + i;
      for(j = i; j < n; j++){
				s = a[aca];
				a[aca] = a[aca1];
				a[aca1] = s;
				aca++;  
				aca1++;
      }
      acy = i * m;  
			acy1 = acr * m;
      for(j = 0; j < m; j++){
				s = y[acy];
				y[acy] = y[acy1];
				y[acy1] = s;
				acy++;  
				acy1++;
      }
    }

/* swap columns */

    if(acc != i){
      ac = av[i];
      av[i] = av[acc];
      av[acc] = ac;
      
      aca = i;  
			aca1 = acc;
      for(j = 0; j < n; j++){
				s = a[aca];
				a[aca] = a[aca1];
				a[aca1] = s;
				aca += n;  
				aca1 += n;
      }				
    }

/* elimination */

    for(j = i + 1; j < n; j++){
      acj = j * n + i;  
			aci= i * n + i;
      s = a[acj] / a[aci];
      for(k = i; k < n; k++){      /* modification of matrix */
				a[acj] -= s * a[aci];
				acj++;  
				aci++;
      }
      acj = j * m;  
			aci = i * m;
      for(k = 0; k < m; k++){      /* modification of rhs */
				y[acj] -= s * y[aci];
				acj++;  
				aci++;
      }
    }
  }
  
/* back reduction */

  for(i = n - 1; i > -1; i--){
    g = a[i * n + i];  
		acx = i * m;
    for(j = 0; j < m; j++){
      s = 0.0;  
			aca = i * n + i + 1;  
			acy = (i + 1) * m + j;
      for(k = i + 1; k < n; k++){
				s += a[aca] * x[acy];
				aca++;  
				acy += m;
      }
      x[acx] = (y[acx] - s) / g;
      acx++;
    }
  }

/* reordering to initial state */  

  for(i = 0; i < n; i++){
    if(av[i] != i){
      for(j = i; j < n; j++){
				if(av[j] == i){
					acc = j;  
					break;
				}
      }
      
      ac = av[i];
      av[i] = av[acc];
      av[acc] = ac;
      
      aca = i * m;  
			aca1 = acc * m;
      for(j = 0; j < m; j++){
				s = x[aca];
				x[aca] = x[aca1];
				x[aca1] = s;
				aca++;  
				aca1++;
      }
    }
  }
  
  free(av);

	return(YES);
}



static void
sort_id_coord(long n, long *id, is_pnt_rec **data)
{
	long i, j, d = 1, tmp_id;
	is_pnt_rec *tmp_data = NULL;

	if(n <= 1)return;

	i = (long)(log((double)n + 0.5) / LN_2);
	while(i--)d *= 2;
	d -= 1;

	while(d != 0){
		for(i = 0; i < n - d; i++){
#ifdef DEBUG
			if(data[i] -> coord == data[i + d] -> coord)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
			if(data[i] -> coord < data[i + d] -> coord)continue;
			tmp_data = data[i + d];
			data[i + d] = data[i];
			tmp_id = id[i + d];
			id[i + d] = id[i];
			if(i < d){
				data[i] = tmp_data;
				id[i] = tmp_id;
				continue;
			}
			for(j = i - d; j >= 0; j -= d){
#ifdef DEBUG
				if(tmp_data -> coord == data[j] -> coord)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
				if(tmp_data -> coord > data[j] -> coord)break;
				data[j + d] = data[j];
				id[j + d] = id[j];
			}
			data[j + d] = tmp_data;
			id[j + d] = tmp_id;
		}
		d /= 2;
	}
}


#ifdef ELIXIR

static void
draw_number(point_rec *pnt, int layer, EPixel color, int number, char *prefix)
{
	WCRec point;
	GraphicObj *gr_obj;
	unsigned long mask = 0L;
	char string[16];

	if(draw == NO)return;

	sprintf(string, "%s%d", prefix, number);

  mask |= LAYER_MASK;
  mask |= COLOR_MASK;
	mask |= FONT_MASK;

	EASValsSetLayer(layer);
	EASValsSetColor(color);
	EASValsSetFontId(font);

	copy_vec_ELIXIR(point, *pnt);

	if((gr_obj = CreateAnnText3D(&point, string)) == NULL){
		ERptErrMessage(ELIXIR_ERROR_CLASS, 2, "Failure of graphics creation", ERROR_GRADE);
		exit(ESISetExitCode(2));
	}

	EGWithMaskChangeAttributes(mask, gr_obj); 
	EMAddGraphicsToModel(ESIModel(), gr_obj);
}


static void
draw_point(point_rec *pnt, int layer, EPixel color)
{
	WCRec point[1];
	GraphicObj *gr_obj = NULL;
	unsigned long mask = 0L;
	
	if(draw == NO)return;

  mask |= LAYER_MASK;
  mask |= COLOR_MASK;
	mask |= MTYPE_MASK;
	mask |= MSIZE_MASK;

	EASValsSetLayer(layer);
	EASValsSetColor(color);
	EASValsSetMType(FILLED_CIRCLE_MARKER);
  EASValsSetMSize(6);	

	copy_vec_ELIXIR(point[0], *pnt);

	if((gr_obj = CreateMarker3D(point)) == NULL){
		ERptErrMessage(ELIXIR_ERROR_CLASS, 2, "Failure of graphics creation", ERROR_GRADE);
		exit(ESISetExitCode(2));
	}

	EGWithMaskChangeAttributes(mask, gr_obj);
	EMAddGraphicsToModel(ESIModel(), gr_obj);
}



static void
draw_line(point_rec *pnt1, point_rec *pnt2, int layer, EPixel color)
{
	WCRec point[2];
	GraphicObj *gr_obj = NULL;
	unsigned long mask = 0L;
	
	if(draw == NO)return;
	
  mask |= LAYER_MASK;
  mask |= COLOR_MASK;
	mask |= STYLE_MASK;
  mask |= SHRINK_MASK;

	EASValsSetLayer(layer);
	EASValsSetColor(color);
	EASValsSetLineStyle(SOLID_STYLE);
	EASValsSetShrink(1.0);

	copy_vec_ELIXIR(point[0], *pnt1);
	copy_vec_ELIXIR(point[1], *pnt2);

	if((gr_obj = CreateLine3D(point)) == NULL){
		ERptErrMessage(ELIXIR_ERROR_CLASS, 2, "Failure of graphics creation", ERROR_GRADE);
		exit(ESISetExitCode(2));
	}
		
	EGWithMaskChangeAttributes(mask, gr_obj);
	EMAddGraphicsToModel(ESIModel(), gr_obj);
}



static void
draw_elliptic_arc(point_rec *pnt1, point_rec *pnt2, point_rec *pnt3, double w1, double w2, double w3, int layer, EPixel color)
{
	WCRec point[3];
	float weight[3];
	GraphicObj *gr_obj = NULL;
	unsigned long mask = 0L;
	
	if(draw == NO)return;
	
  mask |= LAYER_MASK;
  mask |= COLOR_MASK;
	mask |= STYLE_MASK;
	mask |= WIDTH_MASK;
	mask |= SHOW_POLY_MASK;
	mask |= TESSEL_INTERVALS_MASK;

	EASValsSetLayer(layer);
	EASValsSetColor(color);
	EASValsSetLineStyle(SOLID_STYLE);
	EASValsSetLineWidth(0);
	EASValsSetShowPoly(NO);
	EASValsSetTesselIntervals(32);

	copy_vec_ELIXIR(point[0], *pnt1);
	copy_vec_ELIXIR(point[1], *pnt2);
	copy_vec_ELIXIR(point[2], *pnt3);

	weight[0] = w1;
	weight[1] = w2;
	weight[2] = w3;

	if((gr_obj = CreateRBezC3D(3, point, weight)) == NULL){
		ERptErrMessage(ELIXIR_ERROR_CLASS, 2, "Failure of graphics creation", ERROR_GRADE);
		exit(ESISetExitCode(2));
	}
		
	EGWithMaskChangeAttributes(mask, gr_obj);
	EMAddGraphicsToModel(ESIModel(), gr_obj);
}




static void
draw_tile(tile_rec *tile, int layer, EPixel color)
{
	point_rec pnt[4];
	int sign_x[4] = {1, -1, -1, 1};
	int sign_y[4] = {1, 1, -1, -1};
	int i;
	
	if(draw == NO)return;

	for(i = 0; i < 4; i++){
		pnt[i].x = tile -> center.x + sign_x[i] * tile -> size.x * 0.5;
		pnt[i].y = tile -> center.y + sign_y[i] * tile -> size.y * 0.5;
	}

	draw_line(&pnt[0], &pnt[1], layer, color);
	draw_line(&pnt[2], &pnt[3], layer, color);
	draw_line(&pnt[0], &pnt[3], layer, color);
	draw_line(&pnt[1], &pnt[2], layer, color);
}
	

static void
draw_inclusion(inclusion_rec *inclusion, int layer, EPixel color)
{
	if(draw == NO)return;

	draw_ellipse(&(inclusion -> ellipse), layer, color);

#ifdef DRAW_INCLUSION_CENTER
	draw_point(&(inclusion -> ellipse.center), layer, color);
	draw_number(&(inclusion -> ellipse.center), layer, color, inclusion -> id, " ");
#else
	draw_number(&(inclusion -> ellipse.center), layer, color, inclusion -> id, "");
#endif
}


static void
draw_ellipse(ellipse_rec *ellipse, int layer, EPixel color)
{
	point_rec point[3];
	vector_rec vec_max, vec_min;

	if(draw == NO)return;

	copy_mul_vec(vec_max, ellipse -> major, ellipse -> max);
	copy_mul_vec(vec_min, ellipse -> minor, ellipse -> min);

	add_vec(point[0], ellipse -> center, vec_max);
	add_vec(point[2], ellipse -> center, vec_min);
	add_vec(point[1], point[0], vec_min);

	draw_elliptic_arc(&point[0], &point[1], &point[2], 1.0, SQRT1_2, 1.0, layer, color);

	sub_vec(point[2], ellipse -> center, vec_min);
	sub_vec(point[1], point[0], vec_min);
	
	draw_elliptic_arc(&point[0], &point[1], &point[2], 1.0, SQRT1_2, 1.0, layer, color);

	sub_vec(point[0], ellipse -> center, vec_max);
	sub_vec(point[1], point[0], vec_min);

	draw_elliptic_arc(&point[0], &point[1], &point[2], 1.0, SQRT1_2, 1.0, layer, color);

	add_vec(point[2], ellipse -> center, vec_min);
	add_vec(point[1], point[0], vec_min);
	
	draw_elliptic_arc(&point[0], &point[1], &point[2], 1.0, SQRT1_2, 1.0, layer, color);
}


static void
draw_elliptic_arc_segment(ellipse_rec *ellipse, int q, double ta, double tb, point_rec *pointa, point_rec *pointb, 
													trans_matrix *trans, int layer, EPixel color)
{
	point_rec pnt_loc, pnt_glob, point;
	double w0, w1, w2, w01a, w12a, wa, w01b, w12b, wb;
	int sign_x[4] = {1, -1, -1, 1}, sign_y[4] = {1, 1, -1, -1};

	if(draw == NO)return;

	w0 = 1.0;
	w1 = SQRT1_2;
	w2 = 1.0;

	tb = (tb - ta) / (1.0 - ta);

	w01a = (1.0 - ta) * w0 + ta * w1;
	w12a = (1.0 - ta) * w1 + ta * w2;
	wa = (1.0 - ta) * w01a + ta * w12a;

	w01b = (1.0 - tb) * wa + tb * w12a;
	w12b = (1.0 - tb) * w12a + tb * w2;
	wb = (1.0 - tb) * w01b + tb * w12b;
	
	pnt_loc.x = ellipse -> max * sign_x[q - 1] * (1.0 - ta) * w1 / ((1.0 - ta) * w1 + ta * w2);
	pnt_loc.y = ellipse -> min * sign_y[q - 1];

	transform_point_from_local_to_global(&pnt_loc, &pnt_glob, &(ellipse -> center), trans);

	point.x = ((1.0 - tb) * wa * pointa -> x + tb * w12a * pnt_glob.x) / ((1.0 - tb) * wa + tb * w12a);
	point.y = ((1.0 - tb) * wa * pointa -> y + tb * w12a * pnt_glob.y) / ((1.0 - tb) * wa + tb * w12a);

	draw_elliptic_arc(pointa, &point, pointb, wa, w01b, wb, layer, color);
}



static void
draw_elliptic_arc_start_segment(ellipse_rec *ellipse, int q, double t, point_rec *point, trans_matrix *trans, int layer, EPixel color)
{
	point_rec pnt_loc, pnt_glob, point_loc, point_glob;
	double w0, w1, w2, w01, w12, w;
	int sign_x[4] = {1, -1, -1, 1}, sign_y[4] = {1, 1, -1, -1};

	if(draw == NO)return;

	w0 = 1.0;
	w1 = SQRT1_2;
	w2 = 1.0;

	w01 = (1.0 - t) * w0 + t * w1;
	w12 = (1.0 - t) * w1 + t * w2;
	w = (1.0 - t) * w01 + t * w12;

	point_loc.x = ellipse -> max * sign_x[q - 1];
	point_loc.y = 0.0;

	pnt_loc.x = ellipse -> max * sign_x[q - 1];
	pnt_loc.y = ellipse -> min * sign_y[q - 1] * t * w1 / ((1.0 - t) * w0 + t * w1);

	transform_point_from_local_to_global(&pnt_loc, &pnt_glob, &(ellipse -> center), trans);
	transform_point_from_local_to_global(&point_loc, &point_glob, &(ellipse -> center), trans);

	draw_elliptic_arc(&point_glob, &pnt_glob, point, w0, w01, w, layer, color);
}



static void
draw_elliptic_arc_end_segment(ellipse_rec *ellipse, int q, double t, point_rec *point, trans_matrix *trans, int layer, EPixel color)
{
	point_rec pnt_loc, pnt_glob, point_loc, point_glob;
	double w0, w1, w2, w01, w12, w;
	int sign_x[4] = {1, -1, -1, 1}, sign_y[4] = {1, 1, -1, -1};

	if(draw == NO)return;

	w0 = 1.0;
	w1 = SQRT1_2;
	w2 = 1.0;

	w01 = (1.0 - t) * w0 + t * w1;
	w12 = (1.0 - t) * w1 + t * w2;
	w = (1.0 - t) * w01 + t * w12;

	point_loc.x = 0.0;
	point_loc.y = ellipse -> min * sign_y[q - 1];
		
	pnt_loc.x = ellipse -> max * sign_x[q - 1] * (1.0 - t) * w1 / ((1.0 - t) * w1 + t * w2);
	pnt_loc.y = ellipse -> min * sign_y[q - 1];

	transform_point_from_local_to_global(&pnt_loc, &pnt_glob, &(ellipse -> center), trans);
	transform_point_from_local_to_global(&point_loc, &point_glob, &(ellipse -> center), trans);

	draw_elliptic_arc(point, &pnt_glob, &point_glob, w, w12, w2, layer, color);
}



static top_command table[] = {
	{CONST_CHAR_PTR "DISTANCE",   8, select_distance_points, CONST_CHAR_PTR "[# # #, # # #] \ncompute distance of two points"},
};



static void
augment_command_table(void)
{
  TypeInAugmentCommandTable(table, XtNumber(table));
}


void
ESICustomize(Widget parent_pane)
{
}


static void
select_distance_points(void)
{
	double dx, dy, dz, x1, y1, z1, x2, y2, z2, dist = 0.0;
	int i;
	char buffer[128];
	
	if(TypeInGetTokenType(2) == NO_TOKEN){
		EMInstallHandler(ESIModel(), distance_handler,
										 NULL, NULL,
										 NULL, NULL,
										 NULL, NULL,
										 NULL, NULL);
		EVSetSelectActionFunction(NULL, NULL, NULL);
		ERptPrompt(100, 100, "B1=Select first point|B3=Cancel");
	}
	else{
		for(i = 0; i < 6; i++){
			if(TypeInGetTokenType(i + 2) != NUMBER){
				TypeInBadSyntax(NULL);
				return;
			}
		}
		x1 = TypeInGetTokenAsDouble(2);
		y1 = TypeInGetTokenAsDouble(3);
		z1 = TypeInGetTokenAsDouble(4);
		x2 = TypeInGetTokenAsDouble(5);
		y2 = TypeInGetTokenAsDouble(6);
		z2 = TypeInGetTokenAsDouble(7);

		dx = x2 - x1;
		dy = y2 - y1;
		dz = z2 - z1;
		dist = dx * dx + dy * dy + dz * dz;
		sprintf(buffer, "Components [%.8e %.8e %.8e]\nDistance = %20.14e", dx, dy, dz, sqrt(dist));
		ERptStatusMessage(1, 1, buffer);
	}
}



static void
distance_handler(Widget w, EView *v_p, XEvent *event)
{
  static WCRec point1, point2, curr;
	static double total_dist = 0.0;
	static int pick = 0;
	static logic b2 = NO;
	double dist, dx, dy, dz;
	char buffer[128];
  
	switch(event->type){
  case ButtonPress:
    switch(event->xbutton.button){
    case Button1:
			pick++;
			if(pick == 1){
				EVInputPoint(v_p, event, &point1);
				ERptPrompt(100, 100, "B1=Select second point|B3=Cancel");
			}
			else{
				EVInputPoint(v_p, event, &point2);
				dx = point2.x - point1.x;
				dy = point2.y - point1.y;
				dz = point2.z - point1.z;
				dist = sqrt(dx * dx + dy * dy + dz * dz);
				total_dist += dist;
				sprintf(buffer, "Components [%.8e %.8e %.8e]\nDist = %20.14e    Total = %20.14e", 
								dx, dy, dz, dist, total_dist);
				ERptStatusMessage(1, 1, buffer);
				copy_vec(point1, point2);
				ERptPrompt(100, 100, "B1=Select next second point|B2=Repeat|B3=Cancel");
				pick = 1;
				b2 = YES;
			}
			break;
		case Button2:
			if(b2 == YES){
				ERptPrompt(100, 100, "B1=Select first point|B3=Cancel");
				if(pick == 0)total_dist = 0.0;
				sprintf(buffer, "\n%30s Total = %20.14e", "", total_dist);
				ERptStatusMessage(1, 1, buffer);
				pick = 0;
				b2 = NO;
			}
			break;
		case Button3:
			pick = 0;
			total_dist = 0.0;
			ERptStatusMessage(1, 1, "");
			EMUninstallHandler(EVGetModel(v_p));
			break;
		}
		break;
  case MotionNotify:
    EVInputPoint(v_p, event, &curr);
    ERptTrackPointer(&curr);
		break;
	case ClientMessage:
		if(pick == 0)ERptPrompt(100, 100, "B1=Select first point|B3=Cancel");
		if(pick == 1)ERptPrompt(100, 100, "B1=Select second point|B3=Cancel");
		break;
  }
}

#endif


static void
output_patch_boundary(long *edge_curves, long **edge_curve_array, int *edge_curve_sign, long side_patches, long *side_patch_array)
{
	long count, cnt, kk, lines;
	int j, ind;

	count = 0;
	for(j = 0; j < 8; j++)count += edge_curves[j];

	lines = count / ITEMS_PER_LINE;
	cnt = count - lines * ITEMS_PER_LINE;
	if(cnt != 0)
		lines++;
	else
		cnt = ITEMS_PER_LINE;

	ind = 0;
	count = edge_curves[ind];
	kk = 0;
	while(lines--){
		while(cnt){
			if(count <= cnt){
				cnt -= count;
				while(count--)write_record(" %ld", edge_curve_array[ind][kk++] * edge_curve_sign[ind]);
				count = edge_curves[++ind];
				kk = 0;
			}
			else{
				count -= cnt;
				while(cnt--)write_record(" %ld", edge_curve_array[ind][kk++] * edge_curve_sign[ind]);
				break;
			}
		}
		write_record(" \\\n  ");
		cnt = ITEMS_PER_LINE;
	}

	if((count = side_patches) != 0){
		write_record(" subpatch");
		
		lines = count / ITEMS_PER_LINE;
		cnt = count - lines * ITEMS_PER_LINE;
		if(cnt != 0)
			lines++;
		else
			cnt = ITEMS_PER_LINE;
		
		kk = 0;
		while(lines--){
			if(count == cnt){
				while(count--)write_record(" %ld", side_patch_array[kk++]);
			}
			else{
				count -= cnt;
				while(cnt--)write_record(" %ld", side_patch_array[kk++]);
			}
			write_record(" \\\n  ");
			cnt = ITEMS_PER_LINE;
		}
	}
}



void
print_geometry(void)
{
	fprintf(stderr, INDENT "Geometry description:\n");
	fprintf(stderr, INDENT "---------------------\n");
	fprintf(stderr, INDENT "problem identification\n");
	fprintf(stderr, INDENT "tiles inclusions codes\n");
	fprintf(stderr, INDENT "center[xy] dimension[xy] mat imsz bmsz (for reference tile)\n");
	fprintf(stderr, INDENT "id center[xy] point[xy] max min imsz bmsz mat ifmat (for each inclusion)\n");
	fprintf(stderr, INDENT "id codes[-+xy] inclusions ctrl_points (for each particular tile)\n");
#ifdef SEPARATE_PER_REP
	fprintf(stderr, INDENT "id per_x per_y (for each tile inclusion)\n");
#else
	fprintf(stderr, INDENT "id per_xy (for each tile inclusion)\n");
#endif
	fprintf(stderr, INDENT "coords[xy] msz (for each tile ctrl point)\n\n");
}


void
print_setting(void)
{
	fprintf(stderr, INDENT "Setting:\n");
	fprintf(stderr, INDENT "--------\n");
	fprintf(stderr, INDENT "DEF_REL_EPSILON = %e\n", DEF_REL_EPSILON);	
	fprintf(stderr, INDENT "MAX_REL_EPSILON = %e\n", MAX_REL_EPSILON);
	fprintf(stderr, INDENT "MAX_RADIUS_RATE = %f\n", MAX_RADIUS_RATE);
	fprintf(stderr, INDENT "MAX_TILES = %d\n", MAX_TILES);
	fprintf(stderr, INDENT "OFFSET = %f\n", (double)OFFSET);
	fprintf(stderr, INDENT "MULTIPLE_EPS = %f\n", (double)MULTIPLE_EPS);
	fprintf(stderr, INDENT "BUFFER_SIZE = %d\n", BUFFER_SIZE);
#if SETUP == GRID
	fprintf(stderr, INDENT "SETUP = GRID\n");
#endif
#if SETUP == DIAGONAL
	fprintf(stderr, INDENT "SETUP = DIAGONAL\n");
#endif
#if EXCESS_SIZE == MAX_ALLOWED_SIZE
		fprintf(stderr, INDENT "EXCESS_SIZE = MAX_ALLOWED_SIZE\n");
#endif
#if EXCESS_SIZE == MAX_PRESENT_SIZE
		fprintf(stderr, INDENT "EXCESS_SIZE = MAX_PRESENT_SIZE\n");
#endif
#if EXCESS_SIZE == MAX_PRESENT_EXCESS
		fprintf(stderr, INDENT "EXCESS_SIZE = MAX_PRESENT_EXCESS\n");
#endif
#ifdef ARBITRARY_TILE_ID
	fprintf(stderr, INDENT "ARBITRARY_TILE_ID . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "ARBITRARY_TILE_ID . . . . . not defined\n");
#endif
#ifdef SEPARATE_PER_REP
	fprintf(stderr, INDENT "SEPARATE_PER_REP  . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "SEPARATE_PER_REP  . . . . . not defined\n");
#endif
#ifdef DETOUCH
	fprintf(stderr, INDENT "DETOUCH . . . . . . . . . . . . defined\n");
#ifdef OVERDETOUCH
	fprintf(stderr, INDENT "OVERDETOUCH . . . . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "OVERDETOUCH . . . . . . . . not defined\n");
#endif
#else
	fprintf(stderr, INDENT "DETOUCH . . . . . . . . . . not defined\n");
#endif
#ifdef ALLOW_OVERLAP
	fprintf(stderr, INDENT "ALLOW_OVERLAP . . . . . . . . . defined\n");
#else	
	fprintf(stderr, INDENT "ALLOW_OVERLAP . . . . . . . not defined\n");
#endif
#ifdef CHECK_OVERLAP
	fprintf(stderr, INDENT "CHECK_OVERLAP . . . . . . . . . defined\n");
#else	
	fprintf(stderr, INDENT "CHECK_OVERLAP . . . . . . . not defined\n");
#endif
#ifdef ALLOW_OUTSIDE
	fprintf(stderr, INDENT "ALLOW_OUTSIDE . . . . . . . . . defined\n");
#else	
	fprintf(stderr, INDENT "ALLOW_OUTSIDE . . . . . . . not defined\n");
#endif
#ifdef CHECK_COMPATIBILITY
	fprintf(stderr, INDENT "CHECK_COMPATIBILITY . . . . . . defined\n");
#else	
	fprintf(stderr, INDENT "CHECK_COMPATIBILITY . . . . not defined\n");
#endif
#ifdef REDUCED_ESTIMATE
	fprintf(stderr, INDENT "REDUCED_ESTIMATE  . . . . . . . defined\n");
	fprintf(stderr, INDENT "INTERVALS_1D = %d\n", INTERVALS_1D);	
	fprintf(stderr, INDENT "INTERVALS_2D = %d\n", INTERVALS_2D);	
	fprintf(stderr, INDENT "INTERVAL_RATE_1D = %f\n", INTERVAL_RATE_1D);	
	fprintf(stderr, INDENT "INTERVAL_RATE_2D = %f\n", INTERVAL_RATE_2D);	
#if SERIES_1D == CONSTANT_SERIES
	fprintf(stderr, INDENT "SERIES_1D = CONSTANT_SERIES\n");
#endif
#if SERIES_1D == ARITHMETIC_SERIES
	fprintf(stderr, INDENT "SERIES_1D = ARITHMETIC_SERIES\n");
#endif
#if SERIES_1D == GEOMETRIC_SERIES
	fprintf(stderr, INDENT "SERIES_1D = GEOMETRIC_SERIES\n");
#endif
#if SERIES_2D == CONSTANT_SERIES
	fprintf(stderr, INDENT "SERIES_2D = CONSTANT_SERIES\n");
#endif
#if SERIES_2D == ARITHMETIC_SERIES
	fprintf(stderr, INDENT "SERIES_2D = ARITHMETIC_SERIES\n");
#endif
#if SERIES_2D == GEOMETRIC_SERIES
	fprintf(stderr, INDENT "SERIES_2D = GEOMETRIC_SERIES\n");
#endif
#else	
	fprintf(stderr, INDENT "REDUCED_ESTIMATE  . . . . . not defined\n");
#endif
#ifdef USE_TILE
	fprintf(stderr, INDENT "USE_TILE  . . . . . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "USE_TILE  . . . . . . . . . not defined\n");
#endif
#ifdef USE_SIDE
	fprintf(stderr, INDENT "USE_SIDE  . . . . . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "USE_SIDE  . . . . . . . . . not defined\n");
#endif
#ifdef USE_CODE
	fprintf(stderr, INDENT "USE_CODE  . . . . . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "USE_CODE  . . . . . . . . . not defined\n");
#endif
#ifdef USE_SIMPLE
	fprintf(stderr, INDENT "USE_SIMPLE  . . . . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "USE_SIMPLE  . . . . . . . . not defined\n");
#endif
#ifdef DEBUG
	fprintf(stderr, INDENT "DEBUG . . . . . . . . . . . . . defined\n");
#ifdef DEBUG_PRINT
	fprintf(stderr, INDENT "DEBUG_PRINT . . . . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "DEBUG_PRINT . . . . . . . . not defined\n");
#endif
#ifdef ELIXIR
	fprintf(stderr, INDENT "ELIXIR  . . . . . . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "ELIXIR  . . . . . . . . . . not defined\n");
#endif
#endif
	fprintf(stderr, "\n");
}


#include "geom2d.c"

